// $Id: BasicActor.cpp,v 1.1.1.1 2003/07/04 17:15:03 darseq Exp $
#include "BasicActor.h"
#include <X11/keysym.h>
#define STARTCYCLE 3
#include <sys/timeb.h>

#include "behaviour/behaviour.h"
#include "SituationModel.h"
#include "WorldModel.h"
#include "DataForm.h"

BasicActor::BasicActor(int type) : Controller(1 << type,
    (Rescue::ObjectPool *) new WorldModel(gSectorX,gSectorY)) {
    agent_type = type;
    m_logger = NULL;
}
//----------------------------------------------------------------------------

BasicActor::~BasicActor() {
    if (m_logger) {
        delete m_logger;
    }
}
//----------------------------------------------------------------------------

void BasicActor::EnableLogging(char * filename) {
    if (m_logger) {
        delete m_logger;
        m_logger = NULL;
    }
    if (filename) {
        m_logger = new Logger;
        m_logger->Open(filename);
    }
}
//----------------------------------------------------------------------------

void BasicActor::init() {
    fprintf(stderr,"Initiating %s\n",agentName());

    mActs = 0;
    m_prev_time = 1;
    mLastTree = NULL;
    mLastBehaviour = NULL;
    mView = mDataView = mInfoView = NULL;
    m_act_policy = ACT_ONCE_WHEN_IDLE;
    mInitialized = false;

    model()->analyseMap();

    m_last_hear = m_last_send = hear_msgs = 0;
    mBrowser = new AgentForm(this);
    mBrowser->Generate();
}
//----------------------------------------------------------------------------

void BasicActor::act() {
    Debug(spam_spam,"Acting [%s %ld@%ld]\n",agentName(),ID(),m_time);
    assert(m_prev_time!=m_time);

    mActs++;
    // log memory
    if (m_logger) {m_logger->Log(m_time,m_memory);}

//  if(m_time == STARTCYCLE) {
    if(!Initialized() && m_time >= STARTCYCLE) {
//        ASSERT(Initialized() == false);
        initialize();
        mInitialized = true;
    // initialize could use 1 cycle for communication....
    } else if(m_time > STARTCYCLE) {
        update();
        think();
    }
    m_prev_time=m_time;
}
//----------------------------------------------------------------------------

void BasicActor::sense_change(Object * object) {
    sense(object);
    // update time
    model()->SetSenseTime(object,m_time);

    Road * R;
    Building * B;
    Humanoid * H;
    if ((R=dynamic_cast<Road*>(object))) {
        Debug(spam_input,"Input:[Agent %ld Time %ld] Road R%ld "
            "blocked=%ld repaircost=%ld\n",m_id,m_time,R->id(),R->block(),
            R->repairCost());
    } else if ((B=dynamic_cast<Building*>(object))) {
        Debug(spam_input,"Input:[Agent %ld Time %ld] Building B%ld "
            "ignition=%ld fieryness=%ld brokenness=%ld\n",m_id,m_time,
            B->id(),B->ignition(),B->fieryness(),B->brokenness());
    } else if ((H=dynamic_cast<Humanoid*>(object))) {
        Debug(spam_input,"Input:[Agent %ld Time %ld] Humanoid H%ld "
            "hp=%ld dmg=%ld burriedness=%ld position=%ld\n",m_id,m_time,
            H->id(),H->hp(),H->damage(),H->buriedness(),H->position()->id());
    }
}
//----------------------------------------------------------------------------

void BasicActor::act_say(char * msg) {
    packet_start(m_output,AK_SAY);
    m_output.put(m_id);
    m_output.putString(msg);
    packet_end(m_output);
}
//----------------------------------------------------------------------------

void BasicActor::act_tell(char * msg) {
    packet_start(m_output,AK_TELL);
    m_output.put(m_id);
    m_output.putString(msg);
    packet_end(m_output);
}
//----------------------------------------------------------------------------

void BasicActor::Draw(ViewerWindow* aWindow) {
    PointObject *P1,*P2;
    Objects objs=m_memory->objects();

    /*
     * Create views just in time if needed
     */
     
    if (!mView)
    {
        // compute extreme coordinates
        mView=gViewerWindow->CreateView("Object O'Vision");
        mView->SetView(200,40,-1,-1);
        mView->SetSquare(true);

        long minx,miny,maxx,maxy;
        minx=miny=LONG_MAX;
        maxx=maxy=LONG_MIN;
        for(Objects::const_iterator I=objs.begin();I!=objs.end();I++)
        {
            if ((P1=dynamic_cast<Rescue::PointObject*>(*I)))
            {
                if (P1->x()<minx) minx=P1->x();
                if (P1->x()>maxx) maxx=P1->x();
                if (P1->y()<miny) miny=P1->y();
                if (P1->y()>maxy) maxy=P1->y();
            }

        }
        // add edge for buildings near edge
        maxx+=10000;
        minx-=10000;
        maxy+=10000;
        miny-=10000;
        mView->SetScroll(0.0,0.0);
        if (maxx-minx>maxy-miny)
        {
            mView->SetZoom(1.0/(maxx-minx));
        }
        else
        {
            mView->SetZoom(1.0/(maxy-miny));
        }
        mView->SetScroll(-minx,-miny);
        mView->SetMirror(false,true);
     }
    if (!mInfoView)
    {
        mInfoView=gViewerWindow->CreateView("Info Vision");
        mInfoView->SetView(0,0,-1,40);
    }
    if (!mDataView)
    {
        mDataView=gViewerWindow->CreateView("Data Vision");
        mDataView->SetView(0,40,200,-1);
    }
    
    // draw agent info
    char buffer[1024];
    aWindow->SelectView(mInfoView);
    aWindow->SetFill(true);
    aWindow->SetPenColor(0x0);
    aWindow->SetFillColor(0xFFFFFF);
    
/*
    sprintf(buffer,"Type: %s ID: %ld Time: %ld\nHP: %ld DMG: %ld STA: %ld BURRY: %ld",
        agentName(),ID(),m_time,
        Hp(),Damage(),Stamina(),Buriedness());
*/
    sprintf(buffer,"Type: %s ID: %ld Time: %ld",agentName(),m_id,m_time);
    aWindow->Text(0.0,0.0,buffer);
     
    // draw data browser
    // make browser of myself if I don't have one yet
    // base item is always opened
    if (!mBrowser->Opened())
    {
        mBrowser->Open();
    }
    mBrowser->sWorldModel=model();
    mBrowser->Refresh();
    if (aWindow->mViewFlags&view_browser)
    {
        aWindow->SelectView(mDataView);
        aWindow->SetFill(true);
        aWindow->SetPenColor(0x0);
        aWindow->SetFillColor(0xFFFFFF);
        const string& text=mBrowser->ToText("");
        aWindow->Text(0.0,0.0,text);
     }
       
    // draw objectpool
    aWindow->SelectView(mView);
    Object* obj;
    Road* Road;
    Building* Building;
    Humanoid* Humanoid;
    Node* N;

    if (aWindow->mViewFlags&view_roads)
    {
        for(Objects::const_iterator I=objs.begin();I!=objs.end();I++)
        {
            obj=*I;
            if ((Road=dynamic_cast<Rescue::Road*>(obj)))
            {

                int v=0xA0/(1+Road->repairCost());
                aWindow->SetPenColor((v<<16)+(v<<8)+v);
    //            aWindow->SetPenColor(0xA0A0A0);

                P1=(PointObject*)Road->head();
                P2=(PointObject*)Road->tail();
                aWindow->Line(P1->x(),P1->y(),P2->x(),P2->y(),Road->linesToHead()+Road->linesToTail());

                double x,y;
                GetObjectPosition(Road,x,y);
/*                
                x=(P1->x()+P2->x())/2;
                y=(P1->y()+P2->y())/2;
*/                if (Road->repairCost()>0)
                {
                    Memory::TraversalObjectPairFn fn;
                    if ((fn.consider(Road,Road->head())) && (fn.consider(Road,Road->tail())))
                    {   // passable block
                        aWindow->SetPenColor(0x808080);
                    }
                    else
                    {   // not passable block
                        aWindow->SetPenColor(0x0);
                        aWindow->Marker(x,y,'x');
                    }
                }
            }
        }
    }
    if (aWindow->mViewFlags&view_nodes)
    {
        for(Objects::const_iterator I=objs.begin();I!=objs.end();I++)
        {
            obj=*I;
            if ((N=dynamic_cast<Rescue::Node*>(obj)))
            {
                aWindow->SetPenColor(0xFFFF00);
                aWindow->Marker(N->x(),N->y(),'.');
            }
        }
    }
    if (aWindow->mViewFlags&view_buildings)
    {
        for(Objects::const_iterator I=objs.begin();I!=objs.end();I++)
        {
            obj=*I;
            if ((Building=dynamic_cast<Rescue::Building*>(obj)))
            {
                double* vertex=0;
                // Todo: fire is fill yellow to red
                // broken is edge white to black
                if (Building->fieryness()>0)
                {
                    int v=255-(255/(Building->fieryness()+1));
                    aWindow->SetFillColor(0xF00000+(v<<8));
                }
                else
                {
                    switch(Building->type())
                    {
                        case TYPE_FIRE_STATION:
                            aWindow->SetFillColor(0xFF0000);
                        break;
                        case TYPE_POLICE_OFFICE:
                            aWindow->SetFillColor(0x0000FF);
                        break;
                        case TYPE_AMBULANCE_CENTER:
                            aWindow->SetFillColor(0xFFFFFF);
                        break;
                        case TYPE_REFUGE:
                            aWindow->SetFillColor(0x00FF00);
                        break;
                        default:
                        {
                            int v;
                            if (Building->brokenness()!=0)
                            {
                                v=128-(Building->brokenness())/10;
                                if (v<10) v=10;
                            }
                            else
                            {
                                v=128;
                            }
                            aWindow->SetFillColor((v<<16)+(v<<8)+v);
                        }
                        break;
                    }
                }
                Bytes::const_iterator a=Building->buildingApexes().begin();
                int i=0,p,points;
                points=Building->buildingApexes().size()/8;
                vertex=new double[2*points];
                if (vertex)
                {
                    for(i=0,p=0;i<points*8;i+=8,p+=2)
                    {
                        vertex[p+0]=(double)((a[i+0]<<24) + (a[i+1]<<16) + (a[i+2]<<8) + (a[i+3]));
                        vertex[p+1]=(double)((a[i+4]<<24) + (a[i+5]<<16) + (a[i+6]<<8) + (a[i+7]));
                    }
                    aWindow->SetFill(true);
                    aWindow->SetPenColor(0x0);
                    aWindow->Polygon(points,vertex);
                    delete vertex;
                }
            }
        }
    }
    // draw sense times
    if (aWindow->mViewFlags&view_sense)
    {
        for(Objects::const_iterator I=objs.begin();I!=objs.end();I++)
        {
            int sensetime= model()->GetSenseTime(*I);
            if (sensetime>0)
            {
                int size=5-(m_time-sensetime);
                double x,y;
                GetObjectPosition(*I,x,y);
                aWindow->SetPenColor(0xF050FF);
                aWindow->SetFillColor(0xF050FF);
                aWindow->SetFill(true);
                if (size>1)
                {
                    aWindow->Marker(x,y,'o',size);
                }
                else
                {
                    aWindow->Plot(x,y);
                }
            }
        }
    }
    if (aWindow->mViewFlags&view_agents)
    {
        // draw agents
        for(Objects::const_iterator I=objs.begin();I!=objs.end();I++)
        {
            obj=*I;
            if ((Humanoid=dynamic_cast<Rescue::Humanoid*>(obj)))
            {
                switch(Humanoid->type())
                {
                    case TYPE_FIRE_BRIGADE:
                        aWindow->SetFillColor(0xFF0000);
                    break;
                    case TYPE_POLICE_FORCE:
                        aWindow->SetFillColor(0x0000FF);
                    break;
                    case TYPE_AMBULANCE_TEAM:
                        aWindow->SetFillColor(0xFFFFFF);
                    break;
                    case TYPE_CIVILIAN:
                        aWindow->SetFillColor(0x00FF00);
                    break;
                }
                aWindow->SetFill(true);
                Object* P=Humanoid->position();
                if (P)
                {
                    double x,y;
                    GetObjectPosition(P,x,y);

                    if (Humanoid==Self())
                    {
                        aWindow->SetPenColor(0xFFFFFF);
                        aWindow->SetFill(false);
                        aWindow->Marker(x,y,'#',5);
                    }
                    aWindow->SetPenColor(0x0);
                    aWindow->SetFill(true);
                    aWindow->Marker(x,y,'O',5);
                }
            }
        }
    }
    model()->draw(aWindow);

    if (aWindow->mViewFlags&view_cursor)
    {
        // draw selected browser objects
        DataTarget* browser_target=mBrowser->Target();
        if (browser_target)
        {
            browser_target->Draw(aWindow,0xFF);
            delete browser_target;
        }
        DataForm* cursorform=dynamic_cast<DataForm*>(mBrowser->GetAtCursor());
        if (cursorform)
        {
             DataTarget* cursor_target=cursorform->Target();
             if (cursor_target)
             {
                 cursor_target->Draw(aWindow,0xFFFFFF);
                 delete cursor_target;
             }
        }
    }

    if (aWindow->mViewFlags&view_behaviour)
    {
         // todo: draw current move, behaviour, path to destination and behaviour
         if (mLastBehaviour)
         {
             mLastBehaviour->Draw(aWindow);
         }
    }
    if (aWindow->mViewFlags&view_tree)
    {

        if (mLastTree)
        {
            mLastTree->Draw(aWindow);
        }
    }
}


void BasicActor::MouseClick(int mx,int my,int b)
{
    if ((mView) && (mView->Inside(mx,my)))
    {
       Rescue::Object *obj,*nearest=0;
       Objects objs=m_memory->objects();
       double dist,bestdist=3000.0*3000.0;   // minimal distance is 3000.0 millimeter ^2
       double x,y,ox,oy;
       x=mView->ConvertViewToX(mx);
       y=mView->ConvertViewToY(my);
              
       for(Objects::const_iterator I=objs.begin();I!=objs.end();I++)
       {
           obj=*I;
                                   
           GetObjectPosition(obj,ox,oy);
           dist=(ox-x)*(ox-x) + (oy-y)*(oy-y);
           if (dist<bestdist)
           {
               bestdist=dist;
               nearest=obj;
           }
        }       
        if (nearest)
        {
            if (mBrowser)
            {
                delete mBrowser;
                mBrowser=0;
            }
            if (dynamic_cast<Humanoid*>(nearest))
            {
                // TODO select model of target agent as known by this agent
                // instead of real agent
                for(vector<BasicActor*>::iterator AI=gActors.begin();AI!=gActors.end();AI++)
                {
                    if ((*AI)->Self()==nearest)
                    {
                        mBrowser=new AgentForm(*AI);
                        break;
                    }
                }
                if (!mBrowser)
                {
                    mBrowser=new AgentForm(this);
                }
            }
            else if (dynamic_cast<Node*>(nearest))
            {
                mBrowser=new NodeForm((Rescue::Node*)nearest);
            }
            else if (dynamic_cast<Road*>(nearest))
            {
                mBrowser=new RoadForm((Rescue::Road*)nearest);
            }
            else if (dynamic_cast<Building*>(nearest))
            {
                mBrowser=new BuildingForm((Rescue::Building*)nearest);
            }
            else
            {
                mBrowser=new UnknownForm(nearest);
            }
            mBrowser->Generate();
        }
    }
}

void BasicActor::Key(int aKey)
{
    switch (aKey)
    {
        case XK_Up:
            mBrowser->CursorPrev();
        break;
        case XK_Down:
            mBrowser->CursorNext();
        break;
        case XK_Left:
            mBrowser->CursorClose();
        break;
        case XK_Right:
            mBrowser->CursorOpen();
        break;
        case XK_Home:
            delete mBrowser;
            mBrowser=new AgentForm(this);
            mBrowser->Generate();
        break;
    }
}

const char* BasicActor::agentName() {
    char* agent_names[AGENTTYPE_MAZ] = {
        // order of definition is the same as definitions in basic.hxx
        "Civilian","FireAgent","FireCenter","AmbulanceAgent",
        "AmbulanceCenter","PoliceAgent","PoliceCenter"
    };
    return agent_names[agent_type - AGENTTYPE_CIVILIAN];
}
//----------------------------------------------------------------------------

