// $Id: BasicCenter.cpp,v 1.2 2003/07/04 18:52:41 darseq Exp $
// copyright, s.post
#include "BasicCenter.h"
#include "WorldModel.h"
#include "Summary.h"

BasicCenter::BasicCenter(int type) : BasicActor(type) {
}
//----------------------------------------------------------------------------

void BasicCenter::init() {
    BasicActor::init();
    switch(agent_type) {
        case AGENTTYPE_AMBULANCE_CENTER:
            msgs_a_c = MSGS_A_C * model()->a_agents.size();
        break;
        case AGENTTYPE_FIRE_STATION:
            msgs_a_c = MSGS_A_C * model()->f_agents.size();
        break;
        case AGENTTYPE_POLICE_OFFICE:
            msgs_a_c = MSGS_A_C * model()->p_agents.size();
        break;
        default: Debug(spam_startup,"Invalid center type\n"); break;
    }

    // prepare to leave CC_EXCHANGE, can not use CA_READY then
    // we assume this message has just been send, without
    // initialising the center_msgs counters to 1
    phase = CC_EXCHANGE;
    c_ready = 0;
    // assume center_msgs & center_ready
    // are set to 0 & false during insertion
    exchange = false;
    ready2hear = true;
//    newTeams();
}
//----------------------------------------------------------------------------

void BasicCenter::update() {
    communicateWorld();
}
//----------------------------------------------------------------------------

void BasicCenter::initialize() {
//    communicateTeams();
    model()->GetSummary()->centerReset();
}
//----------------------------------------------------------------------------

/*
void BasicCenter::newTeams() {
}
//----------------------------------------------------------------------------

void BasicCenter::communicateTeams() {
}
//----------------------------------------------------------------------------
*/

bool BasicCenter::ready2send() {
    // cycle    m_time      : SEND / KA_HEAR
    // sense -> m_time += 1 : KA_SENSE
    // cycle    m_time      : SEND / KA_HEAR
    if(m_time == m_last_send) {
        Debug(spam_agent_communication,"SEND: waiting for kernel cycle to "
            "complete\n");
        return false;
    }

    switch(phase) {
        case AC_EXCHANGE:
            if(exchange) {
                exchange = false;
                return true;
            } else if(m_time == m_last_hear) {
                Debug(spam_agent_communication,"CC_READY: waiting for "
                    "kernel cycle to complete\n");
            } else {
                hear_msgs = 0;
                phase = CC_READY;
//                Debug(spam_agent_communication,"%ld:phase CC_READY\n",
//                    m_id);
                return true;
            }
        break;
        case CC_EXCHANGE:
            if(exchange) {
                exchange = false;
                return true;
            } else if(ready2hear) {
                if(m_time == m_last_hear) {
                    Debug(spam_agent_communication,"CA_READY: waiting for "
                        "kernel cycle to complete\n");
                } else {
                    ready2hear = false;
                    hear_msgs = 0;
                    phase = CA_READY;
//                    Debug(spam_agent_communication,"%ld:phase CA_READY\n",
//                        m_id);
                    return true;
                }
            }
        break;
        default: break;
    }
    return false; // CA_READY - CC_READY
}
//----------------------------------------------------------------------------

void BasicCenter::hear(Object * sender,const char * msg) {
//    Debug(spam_agent_communication,"CENTER [%ld],t=%d Hear: %ld,\"%s\"\n"
//        ,m_id,m_time,sender->id(),msg,c_ready);
    int centerQ=model()->Centers().size();

    // assume that centers cannot receive say messages
    switch(phase) {
        case AC_EXCHANGE:
            if(anAgent(sender)) {
                Debug(spam_agent_communication,"1SYNC FAILED!\n");
            } else if(aCenter(sender)) {
                if((center_msgs[sender] =
                    (center_msgs[sender] + 1) % MSGS_C_CA) == 0 &&
                    ++c_ready > centerQ) {
                    Debug(spam_agent_communication,"2SYNC FAILED!\n");
                }
            } else {Debug(spam_agent_communication,"Invalid sender!\n");}
        break;
        case CC_EXCHANGE:
            if(ready2hear) {
                if(anAgent(sender)) {
                    Debug(spam_agent_communication,"3SYNC FAILED!\n");
                } else if(aCenter(sender)) {
                    if((center_msgs[sender] =
                        (center_msgs[sender] + 1) % MSGS_C_CA) == 0 &&
                        ++c_ready > centerQ) {
                        Debug(spam_agent_communication,"4SYNC FAILED!\n");
                    }
                } else {Debug(spam_agent_communication,"Invalid sender!\n");}
            } else if(anAgent(sender)) {
                Debug(spam_agent_communication,"5SYNC FAILED!\n");
            } else if(aCenter(sender)) {
                // limits: listen to, but do not interpret your own messages.
                // if a center has sent all CC_EXCHANGE msgs, stop interpret.
                if(sender != Self()) {
                    if(center_ready[sender] == false) {
                        if(++hear_msgs > MSGS_C_C * (centerQ-1)) {
                            Debug(spam_agent_communication,"EXCHANGE FAILED!\n");
                        } else {
                            cExchange(sender,msg);
                            m_last_hear = m_time;
                        }
                    } else {
                        Debug(spam_agent_communication,"ignoring non CC msg\n");
                    }
                }
                if((center_msgs[sender] =
                    (center_msgs[sender] + 1) % MSGS_C_C) == 0) {
                    center_ready[sender] = true;
                    if(++c_ready == centerQ) {
                        ready2hear = true;
                        obj2int_it it = center_ready.begin();
                        for(;it != center_ready.end();it++,c_ready--) {
                            center_ready[it->first] = false;
                        }
                        if(c_ready) {
                            Debug(spam_agent_communication,"6SYNC FAILED!\n");
                        }
                    }
                }
            } else {Debug(spam_agent_communication,"Invalid sender!\n");}
        break;
        case CC_READY:
            if(anAgent(sender)) {
                Debug(spam_agent_communication,"7SYNC FAILED!\n");
            } else if(aCenter(sender)) {
                if((center_msgs[sender] =
                    (center_msgs[sender] + 1) % MSGS_C_CA) == 0 &&
                    ++c_ready > centerQ) {
                    Debug(spam_agent_communication,"8SYNC FAILED!\n");
                }
            } else {Debug(spam_agent_communication,"Invalid sender!\n");}
            if(c_ready == centerQ) {
                Debug(spam_agent_communication,"----%ld:CC EXCHANGE----\n",
                    m_id);
                c_ready = 0;
                exchange = true;
                phase = CC_EXCHANGE;
//                Debug(spam_agent_communication,"%ld:phase CC_EXCHANGE\n",
//                    m_id);
            }
        break;
        case CA_READY:
            if(anAgent(sender)) {
                if(++hear_msgs > msgs_a_c) {
                    Debug(spam_agent_communication,"EXCHANGE FAILED!\n");
                } else {
                    aExchange(sender,msg);
                    m_last_hear = m_time;
                }
                if(hear_msgs == msgs_a_c) {
                    exchange = true;
                    phase = AC_EXCHANGE;
//                    Debug(spam_agent_communication,
//                        "%ld:phase AC_EXCHANGE\n",m_id);
                }
            } else if(aCenter(sender)) {
                if((center_msgs[sender] =
                    (center_msgs[sender] + 1) % MSGS_C_CA) == 0 &&
                    ++c_ready > centerQ) {
                    Debug(spam_agent_communication,"9SYNC FAILED!\n");
                }
            } else {Debug(spam_agent_communication,"Invalid sender!\n");}
        break;
        default: break;
    }
}
//----------------------------------------------------------------------------

void BasicCenter::communicateWorld() {
    if(ready2send() == false) {return;}

    switch(phase) {
        case AC_EXCHANGE:
            // TODO: store received knowledge in centers summary and fill the buffers
            // with a timed value, so A->C messages can fuse with it
            act_tell((char *) (model()->GetSummary()->sectorA));
            act_tell((char *) (model()->GetSummary()->sectorB));
            act_tell((char *) (model()->GetSummary()->agentA));
            act_tell((char *) (model()->GetSummary()->highwayA));
            m_last_send = m_time;
        break;
        case CC_EXCHANGE:
            act_tell((char *) (model()->GetSummary()->sectorA));
            act_tell((char *) (model()->GetSummary()->sectorB));
            act_tell((char *) (model()->GetSummary()->agentA));
            act_tell((char *) (model()->GetSummary()->highwayA));
            m_last_send = m_time;
        break;
        case CC_READY: {
            char message[MSG_SIZE];
            snprintf(message,MSG_SIZE,"CC_READY");
            act_tell(message);
            m_last_send = m_time;
        }
        break;
        case CA_READY: {
            char message[MSG_SIZE];
            snprintf(message,MSG_SIZE,"CA_READY");
            act_tell(message);
            m_last_send = m_time;
        }
        break;
        default: break;
    }
}
//----------------------------------------------------------------------------

void BasicCenter::aExchange(Object * sender,const char * msg) {
    ac_message += msg;
    Debug(spam_agent_communication,
        "CENTER AC[%ld],t=%d Hear: %ld,\"%s\"\n",m_id,m_time,
            sender->id(),ac_message.c_str());
    if(hear_msgs % MSGS_A_C == 0) {
        model()->GetSummary()->centerReceive(ac_message);
        ac_message = "";
    }
}
//----------------------------------------------------------------------------

void BasicCenter::cExchange(Object * sender,const char * msg) {
    Debug(spam_agent_communication,
        "CENTER CC[%ld],t=%d Hear: %ld,\"%s\"\n",m_id,m_time,
            sender->id(),msg);

    switch(hear_msgs % MSGS_C_C) {
        case 1: model()->GetSummary()->crSectorA((const byte*)msg); break;
        case 2: model()->GetSummary()->crSectorB((const byte*)msg); break;
        case 3: model()->GetSummary()->crAgentA((const byte*)msg); break;
        case 0: model()->GetSummary()->crHighwayA((const byte*)msg); break;
    }
}
//----------------------------------------------------------------------------

bool BasicCenter::anAgent(Object * object) {
    // can also be yourself but not a civilian
    return dynamic_cast<Humanoid *>(object) &&
            (dynamic_cast<Civilian *>(object) == NULL);
}
//----------------------------------------------------------------------------

bool BasicCenter::aCenter(Object * object) {
    return dynamic_cast<Building *>(object);
}
//----------------------------------------------------------------------------

void BasicCenter::act_say(char * msg) {
    BasicActor::act_say(msg);
    Debug(spam_agent_communication,"CENTER [%ld] Say: \"%s\"\n",m_id,msg);
}
//----------------------------------------------------------------------------

void BasicCenter::act_tell(char * msg) {
    BasicActor::act_tell(msg);
    Debug(spam_agent_communication,"CENTER [%ld] Tell: \"%s\"\n",m_id,msg);
}
//----------------------------------------------------------------------------

