#include <adk.hxx>
#include "DataForm.h"
//#include "BasicActor.h"
#include "BasicActor.h"
#include "behaviour/behaviour.h"
#include "SituationModel.h"
#include "Team.h"
#include "ImaginaryObject.h"
#include "WorldModel.h"
#include "path.h"
#include "ViewerWindow.h"

WorldModel* DataObject::sWorldModel=0;

    void DataLink::Refresh(void* aData)
    {
        bool opened=false;
        if ((mLink) && (mLink->Data()!=aData))
        {   // data changed. clearing old stupid form
            opened=mLink->Opened();
            delete mLink;
            mLink=0;
        }
        if (!mLink)
        {   // new link or changed link
            if (aData)
            {
                mLink=CreateLink(aData);
            }
            if (!mLink)
            {
                mLink=new UnknownForm(aData);
            }
            mLink->Generate();
            if (opened)
            {
                mLink->Open();
            }
        }
        if ((mLink) && (Opened()))
        {   // if viewed: refresh what I'm pointing to also
            mLink->Refresh();
        }
    }


    DataForm* DataAgentLink::CreateLink(void *aData)
    {
        BasicActor * A = (BasicActor*)aData;
        DataForm* out;
        out=new AgentForm(A);
        if (out)
        {
            out->SetTitle(A->agentName()+IntToString(A->ID()));
        }
        return out;
    }

    DataForm* DataObjectLink::CreateLink(void *aData)
    {
        Rescue::Object *O=(Rescue::Object*)aData;
        DataForm* out=0;
        string name;        
        if (dynamic_cast<Rescue::Node*>(O))
        {
            name="N"+IntToString(O->id());
            out=new NodeForm((Rescue::Node*)O);
        }
        else if (dynamic_cast<Rescue::Road*>(O))
        {
            name="R"+IntToString(O->id());
            out=new RoadForm((Rescue::Road*)O);
        }
        else if (dynamic_cast<Rescue::Building*>(O))
        {
            name="B"+IntToString(O->id());
            out=new BuildingForm((Rescue::Building*)O);
        }
        else
        {
            name="Unknown Object";
        }
        if (out)
        {
            out->SetTitle(name);
        }
        return out;
    }

    DataForm* DataBehaviourLink::CreateLink(void *aData)
    {
        Behaviour* B=(Behaviour*)aData;
        DataForm* out=new BehaviourForm(B);
        if (out)
        {
            out->SetTitle(B->ToString());
        }
        return out;
    }

    DataForm* DataSituationLink::CreateLink(void *aData)
    {
        SituationModel* M=(SituationModel*)aData;
        DataForm* out=new SituationForm(M);
        if (out)
        {
            out->SetTitle(M->ToString());
        }
        return out;
    }    

    DataForm* DataSectorLink::CreateLink(void *aData)
    {
        Sector* S=(Sector*)aData;
        DataForm* out=new SectorForm(S);
        if (out)
        {
            out->SetTitle(S->ToString());
        }
        return out;
    }

DataObject* DataArray::GetAtCursor()
{
    if (Opened())
    {
        if (mCursor==mItems.end())
        {
            return this;
        }
        else
        {
            return (*mCursor)->GetAtCursor();
        }
    }
    else
    {
        return this;
    }        

}

void DataArray::Refresh(const Objects& objs)
{
    DataLink* L;
    if ((int)objs.size()!=Items())
    {
        Clear();
        for(Objects::const_iterator I=objs.begin();I!=objs.end();I++)
        {
            AddItem(L=new DataObjectLink(""));
            L->Refresh(*I);
        }
    }
    else
    {
        int i=0;
        for(Objects::const_iterator I=objs.begin();I!=objs.end();I++)
        {
            L=(DataLink*)Item(i++);
            L->Refresh(*I);
        }
    }
}

void DataArray::Refresh(const vector<Rescue::Object*>& objs)
{
    DataLink* L;
    if ((int)objs.size()!=Items())
    {
        Clear();
        for(vector<Rescue::Object*>::const_iterator I=objs.begin();I!=objs.end();I++)
        {
            AddItem(L=new DataObjectLink(""));
            L->Refresh(*I);
        }
    }
    else
    {
        int i=0;
        for(vector<Rescue::Object*>::const_iterator I=objs.begin();I!=objs.end();I++)
        {
            L=(DataLink*)Item(i++);
            L->Refresh(*I);
        }
    }
}
void DataArray::Refresh(Rescue::Object** objs)
{
Objects o;
    if (objs)
    {
        for(int i=0;objs[i];i++)
        {
            o.push_back(objs[i]);
        }
        Refresh(o);
    }
    else
    {
        Clear();
    }    
}

void DataArray::Refresh(const sectors_t& secs)
{
    DataLink* L;
    if ((int)secs.size()!=Items())
    {
        Clear();
        for(sectors_t::const_iterator I=secs.begin();I!=secs.end();I++)
        {
            AddItem(L=new DataSectorLink(""));
            L->Refresh(*I);
        }
    }
    else
    {
        int i=0;
        for(sectors_t::const_iterator I=secs.begin();I!=secs.end();I++)
        {
            L=(DataLink*)Item(i++);
            L->Refresh(*I);
        }
    }
}

/*
 * Agent form
 */

void AgentForm::Generate()
{
    AddItem(new DataField("Type"));
    AddItem(new DataField("ID"));
    AddItem(new DataObjectLink("Pos"));
    AddItem(new DataObjectLink("Obj"));
    AddItem(new DataBehaviourLink("Behave"));

    AddItem(new DataSituationLink("Mind"));
    
    mCursor=mItems.begin();
}

void AgentForm::Refresh()
{
    BasicActor * A = (BasicActor*) mSourceData;
            
    ((DataField*)GetField("Type"))->SetDataString(A->agentName());
    ((DataField*)GetField("ID"))->SetDataID('A',A->ID());
    ((DataObjectLink*)GetField("Pos"))->Refresh(A->Position());
    ((DataObjectLink*)GetField("Obj"))->Refresh(A->Self());
    ((DataLink*)GetField("Behave"))->Refresh(A->mLastBehaviour);
    ((DataLink*)GetField("Mind"))->Refresh(A->mLastTree);
    // TODO last situationmodel
}
            

DataTarget* AgentForm::Target()
{
    return new DataObjectTarget(((BasicActor*)mSourceData)->Self());
}

void RoadForm::Generate()
{
    AddItem(new DataField("ID"));
    AddItem(new DataObjectLink("Head"));
    AddItem(new DataObjectLink("Tail"));
    AddItem(new DataArray("Moving"));
    AddItem(new DataField("Length"));
    AddItem(new DataField("Kind"));
    AddItem(new DataField("Width"));
    AddItem(new DataField("Block"));
    AddItem(new DataField("Repair"));
    AddItem(new DataField("Passable"));
    AddItem(new DataField("WalkerW"));
    AddItem(new DataField("Lines2H"));
    AddItem(new DataField("Lines2T"));
/*
    // not very relevant and probably impossible to keep up to date
    AddItem(new DataField("Median"));
    AddItem(new DataField("Cars2H"));
    AddItem(new DataField("Cars2T"));
    AddItem(new DataField("Humans2H));
    AddItem(new DataField("Humans2T"));
    AddItem(new DataField("Block"));
    AddItem(new DataField("Repair"));
*/
    AddItem(new DataArray("Sectors"));

    mCursor=mItems.begin();
}

void RoadForm::Refresh()
{
    Rescue::Road* R=(Rescue::Road*)mSourceData;
  
    ((DataField*)GetField("ID"))->SetDataID('R',R->id());
    ((DataLink*)GetField("Head"))->Refresh(R->head());
    ((DataLink*)GetField("Tail"))->Refresh(R->tail());
    // movers
    ((DataField*)GetField("Kind"))->SetDataInteger(R->roadKind());
    ((DataField*)GetField("Length"))->SetDataInteger(R->length());
    ((DataField*)GetField("Width"))->SetDataInteger(R->width());
    ((DataField*)GetField("Block"))->SetDataInteger(R->block());
    ((DataField*)GetField("Repair"))->SetDataInteger(R->repairCost());
    Memory::TraversalObjectPairFn fn;
    if ((fn.consider(R,R->head())) && (fn.consider(R,R->tail())))
    {
        ((DataField*)GetField("Passable"))->SetDataString("Yes");
    }
    else
    {
        ((DataField*)GetField("Passable"))->SetDataString("No");
    }
    ((DataField*)GetField("WalkerW"))->SetDataInteger(R->widthForWalkers());
    ((DataField*)GetField("Lines2H"))->SetDataInteger(R->linesToHead());
    ((DataField*)GetField("Lines2T"))->SetDataInteger(R->linesToTail());

    ((DataArray*)GetField("Sectors"))->Refresh(sWorldModel->findSectors(R));

}


DataTarget* RoadForm::Target()
{
    return new DataObjectTarget((Rescue::Road*)mSourceData);
}

/*
 * Node form
 */
 
void NodeForm::Generate()
{
    AddItem(new DataField("ID"));
    AddItem(new DataField("X"));
    AddItem(new DataField("Y"));
    AddItem(new DataArray("Neighbr"));
    AddItem(new DataArray("Sectors"));
    mCursor=mItems.begin();
}


void NodeForm::Refresh()
{
    Rescue::Node* N=(Rescue::Node*)mSourceData;
    
    ((DataField*)GetField("ID"))->SetDataID('N',N->id());
    ((DataField*)GetField("X"))->SetDataInteger(N->x());
    ((DataField*)GetField("Y"))->SetDataInteger(N->y());
    ((DataArray*)GetField("Neighbr"))->Refresh(N->neighbors());
    ((DataArray*)GetField("Sectors"))->Refresh(sWorldModel->findSectors(N));
    // todo: moving
}

DataTarget* NodeForm::Target()
{
    return new DataObjectTarget((Rescue::Node*)mSourceData);
}

/*
 * Building form
 */
void BuildingForm::Generate()
{
    AddItem(new DataField("Type"));
    AddItem(new DataField("ID"));
    AddItem(new DataField("X"));
    AddItem(new DataField("Y"));
    AddItem(new DataArray("Exits"));
    AddItem(new DataArray("Moving"));
    AddItem(new DataField("Ignite"));
    AddItem(new DataField("Fiery"));
    AddItem(new DataField("Broken"));
    AddItem(new DataField("Floors"));
    AddItem(new DataField("Area"));
    AddItem(new DataField("TArea"));
    AddItem(new DataField("Code"));
    AddItem(new DataArray("Sectors"));

    mCursor=mItems.begin();
}

void BuildingForm::Refresh()
{
    Rescue::Building* B=(Rescue::Building*)mSourceData;
    // todo: type is building, refuge, stations
    ((DataField*)GetField("ID"))->SetDataID('B',B->id());
    ((DataField*)GetField("X"))->SetDataInteger(B->x());
    ((DataField*)GetField("Y"))->SetDataInteger(B->x());
    // exits
    ((DataArray*)GetField("Exits"))->Refresh(B->entrances());

    // todo: moving
    ((DataField*)GetField("Ignite"))->SetDataInteger(B->ignition());
    ((DataField*)GetField("Fiery"))->SetDataInteger(B->fieryness());
    ((DataField*)GetField("Broken"))->SetDataInteger(B->brokenness());
    ((DataField*)GetField("Floors"))->SetDataInteger(B->floors());
    ((DataField*)GetField("Area"))->SetDataInteger(B->buildingAreaGround());
    ((DataField*)GetField("TArea"))->SetDataInteger(B->buildingAreaTotal());
    ((DataField*)GetField("Code"))->SetDataInteger(B->buildingCode());

    ((DataArray*)GetField("Sectors"))->Refresh(sWorldModel->findSectors(B));
        
}

DataTarget* BuildingForm::Target()
{
    return new DataObjectTarget((Rescue::Building*)mSourceData);
}


/*
 * Behaviour Form
 */

void BehaviourForm::Generate()
{
    AddItem(new DataField("Desc"));
    AddItem(new DataField("Priority"));
    AddItem(new DataField("Attain"));
    AddItem(new DataField("Operate"));
    AddItem(new DataObjectLink("Target"));
    AddItem(new DataObjectLink("Position"));
    mCursor=mItems.begin();

}

void BehaviourForm::Refresh()
{
Behaviour *B=(Behaviour*)mSourceData;
    ((DataField*)GetField("Desc"))->SetDataString(B->ToString());
    ((DataLink*)GetField("Target"))->Refresh(B->mTarget);
    ((DataLink*)GetField("Position"))->Refresh(B->mPosition);

}

DataTarget *BehaviourForm::Target()
{
Behaviour *B=(Behaviour*)mSourceData;
    return new DataBehaviourTarget(B);
}


/*
 * Situation Form
 */

void SituationForm::Generate()
{
    AddItem(new DataBehaviourLink("Parent"));
    AddItem(new DataArray("Children"));
    AddItem(new DataField("Time"));
    AddItem(new DataObjectLink("Actor"));
    AddItem(new DataBehaviourLink("Behaviour"));
    AddItem(new DataArray("Team"));
    AddItem(new DataField("Task"));
    AddItem(new DataArray("Behaviours"));
}

void SituationForm::Refresh()
{
    SituationModel *S=(SituationModel*)mSourceData;
    ((DataLink*)GetField("Actor"))->Refresh(S->mActor);

    Objects teammembers;

    const Team* team=S->GetTeam();
    if (team)
    {
        for(Team::const_iterator TI=team->begin();TI!=team->end();TI++)
        {
            teammembers.push_back((*TI)->Self());
        }
    }
    ((DataArray*)GetField("Team"))->Refresh(teammembers);
    ((DataField*)GetField("Task"))->SetDataString(S->mTask->ToString());
}

DataTarget* SituationForm::Target()
{
SituationModel *S=(SituationModel*)mSourceData;
    return new DataSituationTarget(S);
}

/*
 * Sector Form
 */

void SectorForm::Generate()
{
    AddItem(new DataArray("ID"));
    AddItem(new DataField("Location"));
    AddItem(new DataField("NodeQ"));
    AddItem(new DataArray("Nexi"));
}

void SectorForm::Refresh()
{
    Sector* S=(Sector*)mSourceData;    

    // exits
    ((DataField*)GetField("ID"))->SetDataInteger(S->ID());
    DataField* LF=((DataField*)GetField("Location"));
    switch(S->location)
    {
        case NORTH:
            LF->SetDataString("North");
        break;
        case SOUTH:
            LF->SetDataString("South");
        break;
        case EAST:
            LF->SetDataString("East");
        break;
        case WEST:
            LF->SetDataString("West");
        break;
        case CENTER:
            LF->SetDataString("Center");
        break;
        case CORNER:
            LF->SetDataString("Corner");
        break;
        default:
            LF->SetDataString("None");
        break;
    }
    ((DataField*)GetField("NodeQ"))->SetDataInteger(S->NodeQ());
    ((DataArray*)GetField("Nexi"))->Refresh(sWorldModel->nexusPoints(S));
}

DataTarget* SectorForm::Target()
{
Sector* S=(Sector*)mSourceData;
    return new DataSectorTarget(S);
}



/*
 * Unknown Form
 */
 
void UnknownForm::Generate()
{
    AddItem(new DataField("Link"));
    mCursor=mItems.begin();
}

void UnknownForm::Refresh()
{
    if (mSourceData)
    {
        ((DataField*)GetField("Link"))->SetDataString("Unknown");
    }
    else
    {
        ((DataField*)GetField("Link"))->SetDataString("Void");
    }
}

DataTarget* UnknownForm::Target()
{
    return 0;
}




/*
 * General form stuff
 */

string DataField::ToText(string tab)
{
    return tab+Type()+"="+Value()+"\n";
}

string DataLink::ToText(string tab)
{
    if (mLink)
    {
        if (!Opened())
        {
            return tab+"*"+Type()+" "+mLink->Title()+"\n";
        }
        else
        {
            return tab+"&"+Type()+" "+mLink->Title()+"\n"+mLink->ToText(tab+" ");
        }
    }
    else
    {
        return tab+"*"+Type()+" void"+"\n";
    }
}

string DataArray::ToText(string tab)
{

    if (!Opened())
    {
        return tab+"+"+Type()+" "+Value()+"\n";
    }
    else
    {
        string out;
        out=tab+"-"+Type()+" "+Value()+"\n";
        for(vector<DataObject*>::iterator I=mItems.begin();I!=mItems.end();I++)
        {
            if ((I==mCursor) && (!(*I)->Opened()))
            {
                out+=">"+(*I)->ToText(tab+" ");
            }
            else
            {
               out+=" "+(*I)->ToText(tab+" ");
            }
        }
        return out;
    }
}


/*
 * Cursor behaviour
 */
 
void DataArray::CursorNext()
{
    if ((mCursor!=mItems.end()) && ((*mCursor)->Opened()))
    {
        (*mCursor)->CursorNext();
    }
    else
    {
        vector<DataObject*>::iterator I=mCursor;
        I++;
        if (I!=mItems.end())
        {
            mCursor=I;
        }
    }
}

void DataArray::CursorPrev()
{
    if ((mCursor!=mItems.end()) && ((*mCursor)->Opened()))
    {
        (*mCursor)->CursorPrev();
    }
    else
    {
        if (mCursor!=mItems.begin())
        {
            mCursor--;
        }
    }
}
void DataArray::CursorOpen()
{
    if (Opened())
    {
        if (mCursor!=mItems.end())
        {
            (*mCursor)->CursorOpen();
        }
    }
    else
    {
        Open();
    }
}

void DataArray::CursorClose()
{
    if ((mCursor!=mItems.end()) && ((*mCursor)->Opened()))
    {
        (*mCursor)->CursorClose();
    }
    else
    {
        Close();
    }
}

void DataLink::CursorOpen()
{
    if (!Opened())
    {
        Open();
    }
    else
    {
        mLink->CursorOpen();
    }
}

void DataLink::CursorClose()
{
    if ((mLink->GetAtCursor()) && (mLink->GetAtCursor()->Opened()))
    {
        mLink->CursorClose();
    }
    else
    {
        Close();
    }
}

void DataLink::CursorNext()
{
    if (Opened())
    {
        mLink->CursorNext();
    }
}

void DataLink::CursorPrev()
{
    if (Opened())
    {
        mLink->CursorPrev();
    }
}

void DataObjectTarget::Draw(ViewerWindow* aWindow,long aColor)
{
    double x,y;
    GetObjectPosition(mObject,x,y);
    if (dynamic_cast<Node*>(mObject))
    {
        aWindow->SetPenColor(0xFFFFFF);
        aWindow->SetFillColor(aColor);
        aWindow->SetFill(true);
        aWindow->Marker(x,y,'~',10);
/*
        // code used for screenshot in workshop presentation: wave search
        BasicActor *A=aWindow->Actor();
        WorldModel* WM=A->model();
        int length=A->Time()-4;
        if (length>0)
        {
            const sectors_t s=WM->findSectors(mObject);
            for(sectors_t::const_iterator SI=s.begin();SI!=s.end();SI++)
            {
                for(objects_t::const_iterator OI=(*SI)->objects.begin();OI!=(*SI)->objects.end();OI++)
                {
                    Node* dest=dynamic_cast<Node*>(*OI);
                    if (dest)
                    {
                        const Path* P=(*SI)->GetInternalPath(mObject,dest);
                        if (P->Length()<=length)
                        {
                            P->Draw(aWindow,0x00FF00);
                        }
                    }
                }
            }
        }
*/
    }
    else if (dynamic_cast<Humanoid*>(mObject))
    {
        aWindow->SetPenColor(aColor);
        aWindow->SetFill(false);
        aWindow->Marker(x,y,'#',10);
    }
    else if (dynamic_cast<Road*>(mObject))
    {
        Path* P=new Path();
        (*P)+=mObject;
        P->Draw(aWindow,aColor);
        delete P;
    }
    else
    {
        aWindow->SetPenColor(0xFFFFFF);
        aWindow->SetFillColor(aColor);
        aWindow->SetFill(true);
        aWindow->Marker(x,y,'=',10);
    }
    BasicAgent *A=dynamic_cast<BasicAgent*>(aWindow->Actor());
    if ((A) && (A->Self()!=mObject))
    {
        Path *P=A->model()->FindPath(A->Self(),mObject,Path::AdvancedTravelFn(A));
        if (P)
        {
            P->Draw(aWindow,0xFF);
            delete P;
        }
    }
}


void DataSituationTarget::Draw(ViewerWindow* aWindow,long aColor)
{

}
void DataBehaviourTarget::Draw(ViewerWindow* aWindow,long aColor)
{

}

void DataSectorTarget::Draw(ViewerWindow* aWindow,long aColor)
{
    aWindow->SetPenColor(aColor);
    Path *P;

    P=new Path();
    (*P)+=mSector->path01;
    P->Draw(aWindow,aColor);
    delete P;

    P=new Path();
    (*P)+=mSector->path13;
    P->Draw(aWindow,aColor);
    delete P;

    P=new Path();
    (*P)+=mSector->path32;
    P->Draw(aWindow,aColor);
    delete P;

    P=new Path();
    (*P)+=mSector->path20;
    P->Draw(aWindow,aColor);
    delete P;

    objects_t nexi=mSector->nexusPoints();
    for(objects_t::iterator NI=nexi.begin();NI!=nexi.end();NI++)
    {
        double x,y;
        GetObjectPosition(*NI,x,y);
        aWindow->SetFill(true);
        aWindow->SetPenColor(aColor);
        aWindow->SetFillColor(aColor);
        aWindow->Marker(x,y,'o',5);
    }
}



