#ifndef DATAFORM_H
#define DATAFORM_H

#include "main.h"


    class DataTarget
    {
    public:
        virtual void Draw(class ViewerWindow* aWindow,long aColor)=0;
    	virtual ~DataTarget()
	{
	}
    };

    class DataObjectTarget : public DataTarget
    {
    protected:
        Rescue::Object* mObject;
    public:
        DataObjectTarget(Rescue::Object* aObject)
        {
            mObject=aObject;
        }
        void Draw(class ViewerWindow* aWindow,long aColor=0x0);
    };

    class DataSectorTarget : public DataTarget
    {
    protected:
        class Sector* mSector;
    public:
        DataSectorTarget(class Sector* aSector)
        {
            mSector=aSector;
        }
        void Draw(class ViewerWindow* aWindow,long aColor=0x0);
    };

    class DataBehaviourTarget : public DataTarget
    {
    protected:
        class Behaviour* mBehaviour;
    public:
        DataBehaviourTarget(class Behaviour* aBehaviour)
        {
            mBehaviour=aBehaviour;
        }
        void Draw(class ViewerWindow* aWindow,long aColor=0x0);
    };

    class DataSituationTarget : public DataTarget
    {
    protected:
        class SituationModel* mSituation;
    public:
        DataSituationTarget(class SituationModel* aSituation)
        {
            mSituation=aSituation;
        }
        void Draw(class ViewerWindow* aWindow,long aColor=0x0);
    };

    
    class DataObject
    {
    public:
        static class WorldModel* sWorldModel;
    protected:
        string mTypeName;

    public:
        DataObject(const string& aName)
        {
            mTypeName=aName;
        }
	virtual ~DataObject()
	{
	}

        string Type()
        {
            return mTypeName;
        }
        virtual string Value()=0;
        virtual string ToText(string tab)=0;
        virtual void CursorNext() {}
        virtual void CursorPrev() {}
        virtual void CursorOpen() {}
        virtual void CursorClose() {}
        virtual bool Opened() { return false; }
        virtual void Open() {}
        virtual void Close() {}
        virtual DataObject* GetAtCursor() { return this; }
    };
    

    class DataField : public DataObject
    {
    protected:
        string mData;
    public:
        DataField(string aName) : DataObject(aName)
        {
            mData=string("<data>");
        }

        
        void SetDataString(const string& aData)
        {
            mData=aData;
        }
        void SetDataFloat(float aData)
        {
            mData=FloatToString(aData);
        }
        void SetDataInteger(long aData)
        {
            mData=IntToString(aData);
        }
        void SetDataID(char aPrefix,unsigned long aData)
        {
        char prefix[2];
            prefix[0]=aPrefix;prefix[1]=0;
            mData=prefix+IntToString(aData);
        }
        void SetDataBool(bool aData)
        {
            if (aData)
            {
                mData="Yes";
            }
            else
            {
                mData="No";
            }
        }
        
        string Value()
        {
            return mData;
        }
        virtual string ToText(string tab);
    };



    class DataArray : public DataObject
    {
    protected:
        bool mOpened;
        vector<DataObject*> mItems;
        vector<DataObject*>::iterator mCursor;
    public:
        DataArray(const string& aName) : DataObject(aName)
        {
            mOpened=false;
            mCursor=mItems.end();
        }
        ~DataArray()
        {
            mItems.clear();
        }

        string Value()
        {
            return IntToString(mItems.size());
        }

        void Open()
        {
            mOpened=true;
            mCursor=mItems.begin();
        }
        void Close()
        {
            mOpened=false;
            mCursor=mItems.end();
        }
        bool Opened()
        {
            return mOpened;
        }
        
        int Items()
        {
            return mItems.size();
        }
        DataObject* Item(int mIndex)
        {
            if ((mIndex>=0) && (mIndex<(int)mItems.size()))
            {
                return mItems[mIndex];
            }
            return 0;
        }
        void Clear()
        {
            mItems.clear();
            mCursor=mItems.end();
        }

        void Refresh(const Rescue::Objects& objs);
        void Refresh(const vector<Rescue::Object*>& objs);
        void Refresh(Rescue::Object** objs);
        void Refresh(const class sectors_t& secs);
    
        void AddItem(DataObject* aItem)
        {
            mItems.push_back(aItem);
            if (mCursor==mItems.end())
            {
                mCursor=mItems.end();
                mCursor--;
            }
        }
        virtual string ToText(string tab);
        DataObject* GetAtCursor();
        virtual void CursorNext();
        virtual void CursorPrev();
        virtual void CursorOpen();
        virtual void CursorClose();
    };
    

    class DataForm : public DataArray
    {
    protected:
        void *mSourceData;
        string mTitle;
        bool mGenerated;
        // make structure for type. this part should be static

        void AddField(DataObject* aData)
        {
            AddItem(aData);
        }
        DataObject* GetField(const string& aName)
        {
            for(vector<DataObject*>::iterator I=mItems.begin();I!=mItems.end();I++)
            {
                if ((*I)->Type()==aName)
                {
                    return *I;
                }
            }
            assert(0);
            return 0;
        }
    public:
        DataForm(const string& aName) : DataArray(aName)
        {
            mSourceData=0;
            mGenerated=false;
            mCursor=mItems.begin();
            mTitle=string("Title");
        }

        void SetTitle(const string& aTitle)
        {
            mTitle=aTitle;
        }
        string Title()
        {
            return mTitle;
        }

        string Value()
        {
            return Title();
        }
                        
        virtual void Generate()=0;
        // fill in data for type
        virtual void Refresh()=0;

        void *Data()
        {
            return mSourceData;
        }

        void SetCursor(DataObject* aCursor);

        virtual DataTarget* Target()=0;
    };

    class DataLink : public DataObject
    {
    protected:
        class DataForm* mLink;
        virtual class DataForm* CreateLink(void* aData)=0;
    public:
        DataLink(const string& aName) : DataObject(aName)
        {
            mLink=0;
        }
        ~DataLink()
        {
            if (mLink)
            {
                delete mLink;
            }
        }
        virtual string Value()
        {
            if (mLink)
            {
                return mLink->Value();
            }
            else
            {
                return string("void");
            }
        }
        class DataForm* Link()
        {
            return mLink;
        }
        void Open()
        {
            if (mLink)
            {
                mLink->Open();
            }
        }
        void Close()
        {
            if (mLink)
            {
                mLink->Close();
            }
        }
        bool Opened()
        {
            if (mLink)
            {
                return mLink->Opened();
            }
            else
            {
                return false;
            }
        }
        void SetTitle(const string& aTitle)
        {
            if (mLink)
            {
                mLink->SetTitle(aTitle);
            }
        }
        string Title()
        {
            if (mLink)
            {
                return mLink->Title();
            }
            else
            {
                return string("Void");
            }
        }

        void Refresh(void* aData);

        virtual string ToText(string tab);

        virtual void CursorNext();
        virtual void CursorPrev();
        virtual void CursorOpen();
        virtual void CursorClose();
        DataObject* GetAtCursor()
        {
            if (mLink)
            {
                return mLink->GetAtCursor();
            }
            else
            {
                return this;
            }
        }
    };

    class DataAgentLink : public DataLink
    {
    protected:
        class DataForm* CreateLink(void* aData);
    public:
        DataAgentLink(const string& aName) : DataLink(aName)
        {
        }
    };

    class DataObjectLink : public DataLink
    {
    protected:
        class DataForm* CreateLink(void* aData);
    public:
        DataObjectLink(const string& aName) : DataLink(aName)
        {
        }
    };
    
    class DataBehaviourLink : public DataLink
    {
    protected:
        class DataForm* CreateLink(void* aData);
    public:
        DataBehaviourLink(const string& aName) : DataLink(aName)
        {
        }
    };

    class DataSituationLink : public DataLink
    {
    protected:
        class DataForm* CreateLink(void* aData);
    public:
        DataSituationLink(const string& aName) : DataLink(aName)
        {
        }
    };

    class DataSectorLink : public DataLink
    {
    protected:
        class DataForm* CreateLink(void* aData);
    public:
        DataSectorLink(const string& aName) : DataLink(aName)
        {
        }
    };
    
class AgentForm : public DataForm
{
protected:
    virtual void Generate();
    virtual void Refresh();
public:
    AgentForm(class BasicActor * aAgent) : DataForm("Agent")
    {
        mSourceData=(void*)aAgent;
    }    
    DataTarget* Target();
};

class RoadForm : public DataForm
{
protected:
    virtual void Generate();
    virtual void Refresh();
public:
    RoadForm(class Rescue::Road* aRoad) : DataForm("Road")
    {
        mSourceData=(void*)aRoad;
    }
    DataTarget* Target();
};

class NodeForm : public DataForm
{
protected:
    DataObject *mID,*mX,*mY,*mNeighbours;

    virtual void Generate();
    virtual void Refresh();
public:
    NodeForm(class Rescue::Node *aNode) : DataForm("Node")
    {
        mSourceData=aNode;
    }
    DataTarget* Target();
};

class BuildingForm : public DataForm
{
protected:
    virtual void Generate();
    virtual void Refresh();
public:
    BuildingForm(class Rescue::Building *aBuilding) : DataForm("Building")
    {
        mSourceData=aBuilding;
    }
    DataTarget* Target();
};

class UnknownForm : public DataForm
{
protected:
    virtual void Generate();
    virtual void Refresh();
public:
    UnknownForm(void* aData) : DataForm("Unknown")
    {
        mSourceData=aData;
    }
    DataTarget* Target();
};

class BehaviourForm : public DataForm
{
protected:
    virtual void Generate();
    virtual void Refresh();
public:
    BehaviourForm(class Behaviour* aBehaviour) : DataForm("Behaviour")
    {
        mSourceData=(void*)aBehaviour;
    }
    DataTarget* Target();
};

class SituationForm : public DataForm
{
protected:
    virtual void Generate();
    virtual void Refresh();
public:
    SituationForm(class SituationModel* aModel) : DataForm("Situation")
    {
        mSourceData=(void*)aModel;
    }
    DataTarget* Target();
};

class SectorForm : public DataForm
{
protected:
    virtual void Generate();
    virtual void Refresh();
public:
    SectorForm(class Sector* aSector) : DataForm("Sector")
    {
        mSourceData=(void*)aSector;
    }
    DataTarget* Target();
};



#endif
