
#ifndef IMAGINARYOBJECT_H
#define IMAGINARYOBJECT_H

#include <adk.hxx>
#include <vector>
#include "main.h"

using namespace Rescue;

class ImaginaryObject
{
protected:
    Object* mObject;
public:
    int mTimeSensed;
    ImaginaryObject(Object* aObject)
    {
        mObject=aObject;
        mTimeSensed=0;
    }
    virtual ~ImaginaryObject()
    {
    }
    Object* ActualObject() { return mObject; }
    S32 ID() { return mObject->id(); }
    virtual class ImaginaryObject* Override()=0;
    virtual float X()=0;
    virtual float Y()=0;
    float DistanceSqrTo(ImaginaryObject* aObj)
    {
        float dx=X()-aObj->X(),dy=Y()-aObj->Y();
        return dx*dx + dy*dy;
    }
    float DistanceTo(ImaginaryObject* aObj)
    {
        return sqrt(DistanceSqrTo(aObj));
    }
};

class ImaginaryNode : public ImaginaryObject
{
protected:
public:
    ImaginaryNode(Node* aObject) : ImaginaryObject(aObject)
    {
    }
    virtual ~ImaginaryNode()
    {
    }
    virtual class ImaginaryObject* Override()
    {
        return new ImaginaryNode(*this);
    }
    float X() { return ((Node*)mObject)->x()/1000.0; }
    float Y() { return ((Node*)mObject)->y()/1000.0; }

};

class ImaginaryRoad : public ImaginaryObject
{
protected:
public:
    float mBlock;
public:
    ImaginaryRoad(Road* aObject) : ImaginaryObject(aObject)
    {
        mBlock=(float)aObject->block()/(float)aObject->width();
    }
    virtual ~ImaginaryRoad()
    {
        
    }
    virtual class ImaginaryObject* Override()
    {
        return new ImaginaryRoad(*this);
    }
    float X()
    {
        Node* h=(Node*)((Road*)mObject)->head();
        Node* t=(Node*)((Road*)mObject)->tail();
        return (h->x()+t->x())/2000.0;
    }
    float Y()
    {
        Node* h=(Node*)((Road*)mObject)->head();
        Node* t=(Node*)((Road*)mObject)->tail();
        return (h->y()+t->y())/2000.0;
    }

};

class ImaginaryBuilding : public ImaginaryObject
{
protected:
public:
    float mFire;
    float mHp;
    float mWater;
public:
    ImaginaryBuilding(Building* aObject) : ImaginaryObject(aObject)
    {
        if ((aObject->fieryness()&4)!=0)
        {
            mFire=-0.1f;
        }
        else if ((aObject->fieryness()&3)!=0)
        {
            // todo include size and make it
            // the general amount of water
            mFire=1.0f;
        }
        else
        {
            mFire=0.0f;
        }
        mHp=1.0f-(aObject->brokenness()/10000.0f);
        mWater=0.0f;
    }
    virtual ~ImaginaryBuilding()
    {
    }
    virtual class ImaginaryObject* Override()
    {
        return new ImaginaryBuilding(*this);
    }
    float X() { return ((Building*)mObject)->x()/1000.0; }
    float Y() { return ((Building*)mObject)->y()/1000.0; }
};


class ImaginaryHuman : public ImaginaryObject
{
protected:
public:
    ImaginaryObject* mPosition;
    float mHp;
    float mBurry;
    float mWater;

public:
    ImaginaryHuman(Humanoid* aObject,ImaginaryObject* aPosition) : ImaginaryObject(aObject)
    {
        mPosition=aPosition;
        mHp=aObject->hp()/10000.0f;
        mBurry=1.0f-(1.0f/(1.0f+(float)aObject->buriedness()));
        FireBrigade* FB=dynamic_cast<FireBrigade*>(aObject);
        if (FB)
        {
            mWater=FB->waterQuantity()/config.max_extinguish_power();
        }
        else
        {
            mWater=0.0f;
        }
    }
    virtual ~ImaginaryHuman()
    {
    }
    bool Dead() { return mHp<=0.0f; }
    bool Burried() { return mBurry>=0.0f; }
            
    virtual class ImaginaryObject* Override()
    {
        return new ImaginaryHuman(*this);
    }
    float X()
    {
        PointObject* P=dynamic_cast<PointObject*>(mPosition);
        if (P)
        {
           return P->x()/1000.0;
        }
        Edge *E=dynamic_cast<Edge*>(mPosition);
        if (E)
        {
            PointObject* P1=(PointObject*)E->head();
            PointObject* P2=(PointObject*)E->tail();
            return (P1->x()+P2->x())/2000.0;
        }
        return -1.0f;
    }
    float Y()
    {
        PointObject* P=dynamic_cast<PointObject*>(mPosition);
        if (P)
        {
           return P->y()/1000.0;
        }
        Edge *E=dynamic_cast<Edge*>(mPosition);
        if (E)
        {
            PointObject* P1=(PointObject*)E->head();
            PointObject* P2=(PointObject*)E->tail();
            return (P1->y()+P2->y())/2000.0;
        }
        return -1.0f;
    }
    
};

class ImaginaryObjects : public vector<ImaginaryObject*>
{
protected:
    typedef map<S32,ImaginaryObject*> map_type;
    typedef vector<ImaginaryObject*> cont_type;
    map_type mMap;
public:    
    typedef vector<ImaginaryObject*>::const_iterator const_iterator;
    const_iterator begin() const { return cont_type::begin(); }
    const_iterator end() const { return cont_type::end(); }
    void Sort(binary_function<ImaginaryObject*, ImaginaryObject*, bool> aFn)
    {
//  doesn't work because of binary function      ::sort(cont_type::begin(), cont_type::end(),aFn);
    }
    ImaginaryObject* GetObject(const Object* aObject) const
    {
        map_type::const_iterator I=mMap.find(aObject->id());
        if (I!=mMap.end()) 
        {
            return (*I).second;
        }
        else
        {
            return 0;
        }
    }
    bool Contains(const Object* aObject) const
    {
        return mMap.find(aObject->id())!=mMap.end();
    }
    void Add(ImaginaryObject* aObject)
    {
        assert(aObject!=NULL);
        push_back(aObject);
        pair<map_type::iterator,bool> Pos=mMap.insert(map_type::value_type(aObject->ID(),aObject));
        Pos.first->second=aObject;
    }
};


#endif
