#include "SituationModel.h"
#include "WorldModel.h"
#include "behaviour/behaviour.h"
#include "behaviour/behaviouridle.h"
#include "Team.h"
#include "Task.h"
#include "Summary.h"
#include <assert.h>
#include "ViewerWindow.h"
#include <algorithm>

SituationModel::~SituationModel()
{
    if (!mParent)
    {
        delete mOriginalObjects;
    }

    for(Behaviours::iterator BI=mBehaviours.begin();BI!=mBehaviours.end();BI++)
    {
        delete *BI;
    }
    mBehaviours.clear();
    
    for(iterator SI=begin();SI!=end();SI++)
    {
        delete *SI;
    }
    clear();

}


ImaginaryObject* SituationModel::Add(Object* aObject)
{
    Road* R;
    Building *B;
    Node* N;
    Humanoid* H;
    ImaginaryObject* o;
    if ((R=dynamic_cast<Road*>(aObject)))
    {
        mOriginalObjects->Add(o=new ImaginaryRoad(R));
    }
    else if ((N=dynamic_cast<Node*>(aObject)))
    {
        mOriginalObjects->Add(o=new ImaginaryNode(N));
    }
    else if ((B=dynamic_cast<Building*>(aObject)))
    {
        mOriginalObjects->Add(o=new ImaginaryBuilding(B));
    }
    else if ((H=dynamic_cast<Humanoid*>(aObject)))
    {
        ImaginaryObject* IP=0;
        Object* P=H->position();
        
        /*
        // no longer needed pbly
        if (!P)
        {
            if ((dynamic_cast<FireBrigade*>(aObject)) || (dynamic_cast<AmbulanceTeam*>(aObject)) || (dynamic_cast<PoliceForce*>(aObject)))
            {
                P=mWorldModel->GetSummary()->GetAgentPosition((Humanoid*)aObject);
            }
        }
        */
        if (P)
        {
            IP=GetObject(P);
            if (!IP)
            {
                IP=Add(P);
            }
        }
        mOriginalObjects->Add(o=new ImaginaryHuman(H,IP));
    }
    else
    {
        o=0;
    }
    if (o)
    {
        o->mTimeSensed=mWorldModel->GetSenseTime(aObject);
    }
    return o;
}


SituationModel* SituationModel::Predict(TeamMember* aActor,Behaviour* aBehaviour)
{
    SituationModel* O;
    // todo: test if aActor has acted since last time and advanace time
        
    O=aBehaviour->Predict(this,aActor);
    if (O)
    {
        // get changed actor object
        ImaginaryHuman* newHuman=O->GetHuman(aActor->Self());
        // make new team instance and assign the behaviour to it
        TeamMember *newActor=new TeamMember(newHuman);
        newActor->Behaviour()=aBehaviour;
        O->mActor=newActor;
        // add this result to my predicted situations
        push_back(O);
    }
    return O;    
}

ImaginaryObject* SituationModel::Override(ImaginaryObject* aOldObject)
{
    ImaginaryObject* O=aOldObject->Override();
    mOverrideObjects.Add(O);
    return O;
}


void SituationModel::DoFindObjects(ImaginaryObjects& oOut,const SelectionFunction& aFn)
{
    if (mParent)
    {
        // look for the overriden version
        for(ImaginaryObjects::const_iterator I=mOverrideObjects.begin();I!=mOverrideObjects.end();I++)
        {
            ImaginaryObject* O=*I;
            if (!oOut.Contains(O->ActualObject()))
            {
                if (aFn.Consider(O))
                {
                    oOut.Add(O);
                }
            }
        }
        // recursively search parent
        mParent->DoFindObjects(oOut,aFn);
        return;
    }
    else
    {
        // no overridden: take original
        for(ImaginaryObjects::const_iterator I=mOriginalObjects->begin();I!=mOriginalObjects->end();I++)
        {
            ImaginaryObject* O=*I;
            if (!oOut.Contains(O->ActualObject()))
            {
                if (aFn.Consider(O))
                {
                    oOut.Add(O);
                }
            }
        }
        return;
    }
}

ImaginaryObjects SituationModel::FindObjects(const SelectionFunction& aFn)
{
    ImaginaryObjects Out;
    DoFindObjects(Out,aFn);
    return Out;
}

class CompareObjectsFn : public binary_function<ImaginaryObject*, ImaginaryObject*, bool>
{
    private:
        map<S32,float> mValues;
    public:
        bool operator()(ImaginaryObject* A,ImaginaryObject* B)
        {
            return mValues[A->ActualObject()->id()]>mValues[B->ActualObject()->id()];
        }
        CompareObjectsFn(const ImaginaryObjects& aObjects,const SelectionFunction& aFunction)
        {
            for(ImaginaryObjects::const_iterator I=aObjects.begin();I!=aObjects.end();I++)
            {
                mValues[(*I)->ActualObject()->id()]=aFunction.Value(*I);
            }
        }
};
    
ImaginaryObjects SituationModel::FindSortedObjects(const SelectionFunction& aFn)
{
    ImaginaryObjects Out;
    DoFindObjects(Out,aFn);
//    Out.Sort(CompareObjectsFn(Out,aFn));
    
    ::sort(Out.vector<ImaginaryObject*>::begin(), Out.vector<ImaginaryObject*>::end(),CompareObjectsFn(Out,aFn));
 
    return Out;
}

class ImaginaryTravelFn : public Rescue::Memory::ObjectPairFn
{
    protected:
        SituationModel* mSituation;
    public:
        virtual bool consider(const Object *to, const Object *from) const
        {
        ImaginaryRoad* R=mSituation->GetRoad(to);
            if (to==from)
            {
                return true;
            }
            if (R)
            {
                if (R->mBlock>0)
                {
                    return false;
                }
            }
            return true;
        }
        virtual double f(const Object *to, const Object *from) const
        {
        ImaginaryRoad* R=mSituation->GetRoad(to);
            if (R)
            {
                return ((Road*)R->ActualObject())->length();
            }
            return 0.0f;
        }
        ImaginaryTravelFn(SituationModel* aSituation)
        {
            mSituation=aSituation;
        }
};
    
class Path* SituationModel::FindPath(ImaginaryHuman* oTraveler,ImaginaryObject* oDestination,bool aTravel)
{
    ImaginaryObject* start=oTraveler->mPosition;
    ImaginaryObject* dest;
    ImaginaryHuman* H=dynamic_cast<ImaginaryHuman*>(oDestination);
    if (H)
    {
        dest=H->mPosition;
    }
    else
    {
        dest=oDestination;
    }
    // only look for fast paths
    if (aTravel)
    {
        return mWorldModel->FindPath(start->ActualObject(),dest->ActualObject(),ImaginaryTravelFn(this),true);
    }
    else
    {
        return mWorldModel->FindPath(start->ActualObject(),dest->ActualObject(),Path::DummyTravelFn(),true);
    }
}

ImaginaryObject* SituationModel::GetObject(const Object* aObject)
{
    if (!aObject)
    { // shortcut to failure :-D
        return 0;
    }
    if (mParent)
    {
        // look for the overriden version
        ImaginaryObject* O=mOverrideObjects.GetObject(aObject);
        if (O)
        {
            return O;
        }
        else
        {
            // recursively search parent
            return mParent->GetObject(aObject);
        }
    }
    else
    {   // no overridden: take original
        return mOriginalObjects->GetObject(aObject);
    }
    
}

TeamMember* SituationModel::GetTeammember(Rescue::Object* aObject)
{
    if (mParent)
    {   
        if (mActor->Self()==aObject)
        {  // this situation contains the acted one
            return mActor;
        }
        //  if not test my predecessor
        return mParent->GetTeammember(aObject);
    }
    else
    {
        for(Team::iterator I=mTeam->begin();I!=mTeam->end();I++)
        {
            if ((*I)->Self()==aObject)
            {
                return *I;
            }
        }
        // that object is not a teammember
        return 0;
    } 
}

TeamMember* SituationModel::NextActor()
{
    if (!mParent)
    {   // no one acted, first teammember starts
        return *mTeam->begin();
    }
    else
    for(Team::iterator I=mTeam->begin();I!=mTeam->end();I++)
    {
        if ((*I)->Self()==mActor->Self()) // compare to see if they have the same origin
        {   // the last actor is found
            I++;
            if (I==mTeam->end())
            {   // last one that acted is last teammember. we're done
                return 0;
            }
            else
            {   // the member after the one that acted last is next
                return *I;
            }            
        }
    }
    // the last actor is not in the team? assertion!
    assert(false);
    return 0;
}

int SituationModel::Actors()
{
    return mTeam->size();
}


void SituationModel::GenerateBehaviour(TeamMember* aActor)
{
    if (aActor->Position())
    {
        mBehaviours=mTask->GenerateBehaviour(aActor,this);
    }
    else
    {
        // agents that have an unknown position are "skipped"
    }   
}

bool SituationModel::InSector(class TeamMember* aAgent,class Sector* aSector)
{
    if (aAgent->Position())
    {
        return mWorldModel->insideSector(aSector,aAgent->Position()->ActualObject());
    }
    else
    {
        return false;
    }
}

void SituationModel::SetTeam(SummaryTeam* aTeam)
{
    mTeam=new Team();
    for(vector<Object*>::iterator TI=aTeam->begin();TI!=aTeam->end();TI++)
    {
        ImaginaryHuman* H=GetHuman(*TI);
        assert(H!=0);
        mTeam->push_back(new TeamMember(H));
    }
}



void SituationModel::Draw(ViewerWindow* aWindow)
{
    // todo, draw sector boundaries
    // todo, draw contents
}


