// $Id: Summary.cpp,v 1.3 2003/07/05 18:44:35 darseq Exp $
#include "Summary.h"
#include "WorldModel.h"
#include "Task.h"
#include "SituationModel.h"
#include "ViewerWindow.h"

bool SummaryTeam::Contains(Object * aAgent) const {
    for (vector < Object * >::const_iterator I = begin(); I != end(); I++) {
        if (*I == aAgent) {
            return true;
        }
    }
    return false;
}
//----------------------------------------------------------------------------

void SummaryTeam::Add(Object * aAgent) {
    push_back(aAgent);
}
//----------------------------------------------------------------------------

Summary::Summary() {
}
//----------------------------------------------------------------------------

Summary::~Summary() {
    for (vector < SummaryTeam * >::iterator TI = mTeams.begin();
         TI != mTeams.end(); TI++) {
        delete *TI;
    }
}
//----------------------------------------------------------------------------

void Summary::Build(class WorldModel * aWorldModel) {
    mWorldModel=aWorldModel;
    int index;

    // initialize sector fire properties
    index = 0; const sectors_t & secs = mWorldModel->getSectors();
    for (sectors_t::const_iterator SI = secs.begin(); SI != secs.end();
        SI++) {
        mSectorIndex[*SI] = index++;
        mFire.push_back(0);
        mFireUpdated.push_back(0);
        mIntel.push_back(0);
        mIntelUpdated.push_back(0);
        mDmg.push_back(0);
        mDmgUpdated.push_back(0);
        mRoads.push_back(0);
        mRoadsUpdated.push_back(0);
    }
    ASSERT(index < MSG_SIZE);

    // find agents 
    const Objects & objs = mWorldModel->objects();

    for (Objects::const_iterator OI = objs.begin(); OI != objs.end(); OI++) {
        MovingObject *M = 0;
        switch ((*OI)->type()) {
        case TYPE_FIRE_BRIGADE:
        case TYPE_POLICE_FORCE:
        case TYPE_AMBULANCE_TEAM:
            M = (MovingObject *) (*OI);
            break;
        }
        if (M) {
            mPositionUpdated[M->id()]=0;
            mAgentPosition[M->id()]=M->position();
        }
    }
    // put agents in teams (NB: do this after because now they're sorted by ID
    // instead of first sensed (which put's this agent first)
    SummaryTeam *fireTeamA=new SummaryTeam(SummaryTeam::cFunctionFire);
    mTeams.push_back(fireTeamA);
    SummaryTeam *fireTeamB=new SummaryTeam(SummaryTeam::cFunctionFire);
    mTeams.push_back(fireTeamB);
    SummaryTeam* ambulanceTeam=new SummaryTeam(SummaryTeam::cFunctionRescue);
    mTeams.push_back(ambulanceTeam);
    SummaryTeam *policeTeam=new SummaryTeam(SummaryTeam::cFunctionHighways);
    mTeams.push_back(policeTeam);
    // TODO: increased numbers because travel team sucks
    int mFirePoliceA = 3,mFirePoliceB=3, mAmbulancePolice = 3;
    bool nextFireTeam=1;
    
    for(map<S32,Object*>::iterator AI=mAgentPosition.begin();AI!=mAgentPosition.end();AI++)
    {
        Humanoid* H=(Humanoid*)mWorldModel->get(AI->first);
        switch (H->type()) {
        case TYPE_FIRE_BRIGADE:
            if (nextFireTeam==1)
            {
                fireTeamA->Add(H);
                nextFireTeam=2;
            }
            else
            {
                fireTeamB->Add(H);
                nextFireTeam=1;
            }
            break;
        case TYPE_POLICE_FORCE:
            if (mFirePoliceA > 0) {
                fireTeamA->Add(H);
                mFirePoliceA--;
            } else if (mFirePoliceB >0) {
                fireTeamB->Add(H);
                mFirePoliceB--;
            } else if (mAmbulancePolice > 0) {
                ambulanceTeam->Add(H);

           mAmbulancePolice--;
            } else {
                policeTeam->Add(H);
            }
            break;
        case TYPE_AMBULANCE_TEAM:
             ambulanceTeam->Add(H);
            break;
        }  
    }
    // make routes
    index = 0;
    for (sectors_t::const_iterator SI = secs.begin(); SI != secs.end();SI++) {
        const objects_t & nexi = mWorldModel->nexusPoints(*SI);
        for (objects_t::const_iterator NI1 = nexi.begin();
             NI1 != nexi.end(); NI1++) {
            for (objects_t::const_iterator NI2 = NI1; NI2 != nexi.end();
                 NI2++) {
                if (NI1 != NI2) {
                    const Path *HW = mWorldModel->GetHighway(*NI1, *NI2);
                    for (vector < const Path * >::iterator HWI =
                         mHighway.begin(); HWI != mHighway.end(); HWI++) {
                        // disallow reverse and duplicates
                        const Path *HW2 = *HWI;
                        if ((HW2->Contains(*NI1)) && (HW2->Contains(*NI2))) {
                            HW = 0;
                            break;
                        }
                    }
                    if (HW) {
                        mHighway.push_back(HW);
                        mHighwayFrom.push_back(*NI1);
                        mHighwayTo.push_back(*NI2);
                        
                        mHighwayBlock.push_back(false);
                        // compute score based on frequency of this part
                        // in the total number of highways
                        float score = 0.0f;
                        float q = 1.0f;
                        for (vector < class Path * >::iterator HWI =
                             mWorldModel->mHighway.begin();
                             HWI != mWorldModel->mHighway.end(); HWI++) {
                            if (*HWI) {
                                if (((*HWI)->Contains(*NI1))
                                    && ((*HWI)->Contains(*NI2))) {
                                    score += 1.0f;
                                }
                                q += 1.0f;
                            }
                        }
                        mHighwayPriority.push_back(score / q);
                    }
                }
            }
        }
    }
}
//----------------------------------------------------------------------------

string Summary::Update() {
    // update sectors
    string ac_fire = "(";
    map < Sector *, int >::iterator SI;
    for (SI = mSectorIndex.begin(); SI != mSectorIndex.end(); SI++)
    {
        Sector *S = (*SI).first; const objects_t & objs = S->objects;
        S32 sense_time = 0; int total_fire = 0; int num_buildings = 0;

        objects_t::const_iterator OI;
        for (OI = objs.begin();OI != objs.end();OI++)
        {
            if ((*OI)->type() == TYPE_BUILDING)
            {
                Building *B = (Building *) mWorldModel->get((*OI)->id());
                if (B)
                {
                    if (((B->fieryness() & 3) != 0) &&
                        ((B->fieryness() & 4) == 0))
                        {
                        total_fire += 1; S32 time;
                        if((time = mWorldModel->GetSenseTime(B)) >
                            sense_time) {
                            sense_time = time;
                        }
                    }
                    ++num_buildings;
                }
            }
        }

        if(sense_time > mFireUpdated[(*SI).second])
        {
            mFireUpdated[(*SI).second] = sense_time;
            float average = 0;
            if (num_buildings)
            {
                average = (float) total_fire / num_buildings;
            }
            mFire[(*SI).second] = average;

            ac_fire += "(" + IntToString(SI->second) + "," +
                IntToString((byte) (average*CHAR_MAX)) + "),";
        }
    }
    ac_fire.resize(ac_fire.length()-1);
    if(ac_fire.length())
    {
        ac_fire += ")";
    }

    string ac_intel = "(";
    // todo: update intel
    ac_intel.resize(ac_intel.length()-1);
    if(ac_intel.length())
    {
        ac_intel += ")";
    }

    string ac_dmg = "(";

    ac_dmg.resize(ac_dmg.length()-1);
    if(ac_dmg.length())
    {
        ac_dmg += ")";
    }

    string ac_roads = "(";

    ac_roads.resize(ac_roads.length()-1);
    if(ac_roads.length())
    {
        ac_roads += ")";
    }
    
    // update positions
    string ac_positions = "(";
    for (map < S32, Object* >::iterator AI = mAgentPosition.begin(); AI != mAgentPosition.end(); AI++)
    {
        Humanoid * H = (Humanoid *) mWorldModel->get((*AI).first);
        S32 sense_time = mWorldModel->GetSenseTime(H);

        if(sense_time > mPositionUpdated[(*AI).first])
        {
            mPositionUpdated[(*AI).first] = sense_time;
            AI->second= H->position();

            // centers use USHRT_MAX to indicate not updated
            // assume position->id never becomes USHRT_MAX
            // assume position->id never becomes NUL byte
            if (AI->second)
            {
               ac_positions += "(" + IntToString(AI->first) + "," +
                  IntToString((word) (AI->second->id())) + "),";
            }
        }
    }
    ac_positions.resize(ac_positions.length()-1);
    if(ac_positions.length()) {ac_positions += ")";}

    // update highways
    // TODO
    return "AC_EXCHANGE\tF="+ac_fire+"\tI="+ac_intel+"\tD="+ac_dmg+"\tR="+ac_roads+"\tP="+ac_positions+"\tH=";
}
//----------------------------------------------------------------------------

SummaryTeam *Summary::MakeTeam(class BasicAgent * aAgent) {
    for (vector < SummaryTeam * >::iterator TI = mTeams.begin();
         TI != mTeams.end(); TI++) {
        if ((*TI)->Contains(aAgent->Self())) {
            return *TI;
        }
    }
    return 0;
}
//----------------------------------------------------------------------------

Task* Summary::MakeTask(class BasicAgent* aAgent,SummaryTeam* aTeam)
{
    Sector* Worst=0;
    float fire,worstFire=0;
    for(map<Sector*,int>::iterator SI=mSectorIndex.begin();SI!=mSectorIndex.end();SI++)
    {
        fire=mFire[(*SI).second];
        if (fire>worstFire)
        {
            Worst=(*SI).first;
            worstFire=fire;
        }
    }

    switch(aTeam->Function())
    {
        case SummaryTeam::cFunctionFire:
        {
            if (Worst)
            {        

                return new FightFireTask(Worst);
            }
            else
            {
                return new PatrolTask();
            }
        }
        break;
        case SummaryTeam::cFunctionRescue:
            if (Worst)
            {

                return new SearchAndRescueTask(Worst);
            }
            else
            {
                return new PatrolTask();
            }
        break;
        
        case SummaryTeam::cFunctionHighways:
            if (Worst)
            {

                return new ImproveTravelTask(Worst);
            }
            else
            {
                return new PatrolTask();
            }
        break;
        default:
            return new PatrolTask();
        break;
    }
    return new PatrolTask();
}
//----------------------------------------------------------------------------

// sectorA : each byte = 1|I|I|I|F|F|F|F
// 1 is always 1 (0 prevention) III = intel (0-7) FFFF = fire (0-15)

// SectorFire lst format: ((SI,F),(SI,F),...)
// SI = sectorIndex, F = fireness
void Summary::crSectorFire(Eltlst * lst,byte* aBuffer) {
    Eltlst::iterator it1; Eltlst * sublst;
    for(it1 = lst->begin();it1 != lst->end();it1++) {
        if((sublst = dynamic_cast<Eltlst *>(*it1))) {
            Eltlst::iterator it2 = sublst->begin();
            int sectorIndex = atoi(((Scalar *) *it2)->c_str());
            assert(sectorIndex<MSG_SIZE-2);
            int sector_fire = atoi(((Scalar *) *(++it2))->c_str());
            byte v=aBuffer[sectorIndex];
            byte ov=v&0xF;
            byte nv=sector_fire>>4;
            if (nv>ov)
            {
                v&=0xF0;
                v|=nv;
            }
            aBuffer[sectorIndex]=v;
        }
    }
}
//----------------------------------------------------------------------------

// SectorIntel lst format: ((SI,I),(SI,I),...)
// SI = sectorIndex, I = intelliness
void Summary::crSectorIntel(Eltlst * lst,byte* aBuffer) {
    Eltlst::iterator it1; Eltlst * sublst;
    for(it1 = lst->begin();it1 != lst->end();it1++) {
        if((sublst = dynamic_cast<Eltlst *>(*it1))) {
            Eltlst::iterator it2 = sublst->begin();
            int sectorIndex = atoi(((Scalar *) *it2)->c_str());
            assert(sectorIndex<MSG_SIZE-2);
            assert(sectorIndex<MSG_SIZE-2);
            int sector_fire = atoi(((Scalar *) *(++it2))->c_str());
            byte v=aBuffer[sectorIndex];
            byte ov=(v>>4)&0x7;
            byte nv=sector_fire>>5;
            if (nv>ov)
            {
                v&=0x70;
                v|=nv<<4;
            }
            aBuffer[sectorIndex]=v;
        }
    }
}
//----------------------------------------------------------------------------

// sectorB : each byte = 1|D|D|D|R|R|R|R
// 1 is always 1 (0 prevention) DDD = damage (0-7) RRRR = roads (0-15)
    
// TODO: make getters for all of these
void SetRoadsInMsg(byte* aBuffer,int aIndex,byte aValue)
{
    byte v=aBuffer[aIndex];
    byte ov=v&0xF;
    byte nv=aValue>>4;
    if (nv>ov)
    {
        v&=0xF0;
        v|=nv;
    }
    aBuffer[aIndex]=v;
}
//----------------------------------------------------------------------------

// SectorRoads lst format: ((SI,R),(SI,R),...)
// SI = sectorIndex, R = road freeness
void Summary::crSectorRoads(Eltlst * lst,byte* aBuffer) {
    Eltlst::iterator it1; Eltlst * sublst;
    for(it1 = lst->begin();it1 != lst->end();it1++) {
        if((sublst = dynamic_cast<Eltlst *>(*it1))) {
            Eltlst::iterator it2 = sublst->begin();
            int sectorIndex = atoi(((Scalar *) *it2)->c_str());
            assert(sectorIndex<MSG_SIZE-2);
            int sector_roads = atoi(((Scalar *) *(++it2))->c_str());
            SetRoadsInMsg(aBuffer,sectorIndex,sector_roads);
        }
    }
}

//----------------------------------------------------------------------------

// SectorDmg lst format: ((SI,I),(SI,I),...)
// SI = sectorDamage, D = damageness

void Summary::crSectorDmg(Eltlst * lst,byte* aBuffer) {
    Eltlst::iterator it1; Eltlst * sublst;
    for(it1 = lst->begin();it1 != lst->end();it1++) {
        if((sublst = dynamic_cast<Eltlst *>(*it1))) {
            Eltlst::iterator it2 = sublst->begin();
            int sectorIndex = atoi(((Scalar *) *it2)->c_str());
            int sector_dmg = atoi(((Scalar *) *(++it2))->c_str());
            assert(sectorIndex<MSG_SIZE-2);
            byte v=aBuffer[sectorIndex];
            byte ov=(v>>4)&0x7;
            byte nv=sector_dmg>>5;
            if (nv>ov)
            {
                v&=0x70;
                v|=nv<<4;
            }
            aBuffer[sectorIndex]=v;
        }
    }
}
//----------------------------------------------------------------------------

S32 GetPositionFromMsg(const byte* aBuffer,int aIndex)
{
    return (aBuffer[aIndex*2]&0x7f)+(((aBuffer[aIndex*2+1])&0x7f)<<7);
}
//----------------------------------------------------------------------------
    
void SetPositionInMsg(byte* aBuffer,int aIndex,S32 aPos)
{
    if ((aBuffer[aIndex*2]==128) && (aBuffer[aIndex*2+1]==128))
    {
        aBuffer[aIndex*2]&=128;
        aBuffer[aIndex*2+1]&=128;

        aBuffer[aIndex*2]|=(aPos)&0x7F;
        aBuffer[aIndex*2+1]|=(aPos>>7)&0x7F;
    }
}
//----------------------------------------------------------------------------

// Position lst format: ((AI,ID),(AI,ID),...)
void Summary::crPosition(Eltlst * lst,byte* aBuffer) {
    Eltlst::iterator it1; Eltlst * sublst;
    for(it1 = lst->begin();it1 != lst->end();it1++) {
        if((sublst = dynamic_cast<Eltlst *>(*it1))) {
            Eltlst::iterator it2 = sublst->begin();
            int agentID = atoi(((Scalar *) *it2)->c_str());
            int positionID = atoi(((Scalar *) *(++it2))->c_str());

            assert(positionID < 8192);
            int index = 0;
            for(map<S32,Object*>::iterator AI=mAgentPosition.begin();AI!=mAgentPosition.end();AI++)
            {
                if (AI->first==agentID)
                {
                    break;
                }
                index++;
            }
            SetPositionInMsg(aBuffer,index,positionID);
        }
    }
}
//----------------------------------------------------------------------------

// Highway lst format: ?
void Summary::crHighway(Eltlst * lst,byte* aBuffer) {

}
//----------------------------------------------------------------------------

void Summary::crSectorA(const byte* aMessage)
{
    assert(aMessage[0]==1);
}
//----------------------------------------------------------------------------

void Summary::crSectorB(const byte* aMessage)
{
    assert(aMessage[0]==2);
}
//----------------------------------------------------------------------------

void Summary::crAgentA(const byte* aMessage)
{
    assert(aMessage[0]==3);
    int index=0;
    for(map<S32,Object*>::iterator AI=mAgentPosition.begin();AI!=mAgentPosition.end();AI++)
    {
        S32 pos=GetPositionFromMsg(aMessage+1,index);
        if (pos)
        {
            SetPositionInMsg(agentA+1,index,pos);
        }
        index++;
    }
}
//----------------------------------------------------------------------------

void Summary::crHighwayA(const byte* aMessage)
{
    assert(aMessage[0]==4);
}
//----------------------------------------------------------------------------

void Summary::arSectorA(const byte* aMessage,S32 aMsgTime)
{
    assert(aMessage[0]==1);
    int p=1;
    for(map< class Sector *, int >::iterator SI=mSectorIndex.begin();SI!=mSectorIndex.end();SI++)
    {
        if (aMsgTime>mFireUpdated[SI->second])
        {
            float fire=((aMessage[p]&0xF))/16.0;
            mFire[SI->second]=fire;
        }
        if (aMsgTime>mIntelUpdated[SI->second])
        {
            float intel=((aMessage[p]>>4)&0x7)/8.0;
            mIntel[SI->second]=intel;
        }    
        p++;
    }
}
//----------------------------------------------------------------------------

void Summary::arSectorB(const byte* aMessage,S32 aMsgTime)
{
    assert(aMessage[0]==2);
    int p=1;
    for(map< class Sector *, int >::iterator SI=mSectorIndex.begin();SI!=mSectorIndex.end();SI++)
    {
        if (aMsgTime>mRoadsUpdated[SI->second])
        {
            float roads=((aMessage[p]&0xF))/16.0;
            mRoads[SI->second]=roads;
        }
        if (aMsgTime>mDmgUpdated[SI->second])
        {
            float dmg=((aMessage[p]>>4)&0x7)/8.0;
            mDmg[SI->second]=dmg;
        }
        p++;
    }
}
//----------------------------------------------------------------------------

void Summary::arAgentA(const byte* aMessage,S32 aMsgTime)
{
    assert(aMessage[0]==3);
    int i=0;
    for(map<S32,Object*>::iterator AI=mAgentPosition.begin();AI!=mAgentPosition.end();AI++)
    {
        if (aMsgTime>mPositionUpdated[AI->first])
        {
            S32 position=GetPositionFromMsg(aMessage+1,i);
            
            AI->second=mWorldModel->get(position);
            mPositionUpdated[AI->first]=aMsgTime;
            
            Humanoid* H=(Humanoid*)mWorldModel->get(AI->first);
            H->setPosition(aMsgTime,AI->second);
        }
        i++;
    }
}
//----------------------------------------------------------------------------

void Summary::arHighwayA(const byte* aMessage,S32 aMsgTime)
{
    assert(aMessage[0]==4);
}
//----------------------------------------------------------------------------

void Summary::centerReset() {

    // MSG_SIZE bytes available
    unsigned int i;
    sectorA[0]=1;
    sectorB[0]=2;
    agentA[0]=3;
    highwayA[0]=4;
    
    for(i=1;i < MSG_SIZE;i++) {
        // UCHAR_MAX indicates no change
        sectorA[i] = 1<<7;
        sectorB[i] = 1<<7;
        highwayA[i] = 1<<7;
        agentA[i] = 1<<7;
    }
    sectorA[i]=0;
    sectorB[i]=0;
    highwayA[i]=0;
    agentA[i]=0;
}

void Summary::centerReceive(string ac_message) {
    Message M(ac_message); Eltlst * lst;
    ASSERT(M.header == "AC_EXCHANGE");

    if((lst = dynamic_cast<Eltlst *>(M.getElt("F")))) {crSectorFire(lst,sectorA+1);}
    if((lst = dynamic_cast<Eltlst *>(M.getElt("I")))) {crSectorIntel(lst,sectorA+1);}
    if((lst = dynamic_cast<Eltlst *>(M.getElt("D")))) {crSectorDmg(lst,sectorB+1);}
    if((lst = dynamic_cast<Eltlst *>(M.getElt("R")))) {crSectorRoads(lst,sectorB+1);}
    if((lst = dynamic_cast<Eltlst *>(M.getElt("P")))) {crPosition(lst,agentA+1);}
    if((lst = dynamic_cast<Eltlst *>(M.getElt("H")))) {crHighway(lst,highwayA+1);}
}

//----------------------------------------------------------------------------

void Summary::Draw(ViewerWindow* aWindow,WorldModel* aWM)
{
    aWindow->SetPenColor(0x701070);
    aWindow->SetFill(false);
    for(map<Sector*,int>::iterator SI=mSectorIndex.begin();SI!=mSectorIndex.end();SI++)
    {
        Sector* S=SI->first;
        double x=S->MiddleX();
        double y=S->MiddleY();
        string s;
        float v;
        long t;
        v=mFire[SI->second];
        t=mFireUpdated[SI->second];
        s+="Fire["+IntToString(t)+"]="+FloatToString(v)+"\n";
        v=mIntel[SI->second];
        t=mIntelUpdated[SI->second];
        s+="Intel["+IntToString(t)+"]="+FloatToString(v)+"\n";
        v=mDmg[SI->second];
        t=mDmgUpdated[SI->second];
        s+="Dmg["+IntToString(t)+"]="+FloatToString(v)+"\n";
        v=mRoads[SI->second];
        t=mRoadsUpdated[SI->second];
        s+="Roads["+IntToString(t)+"]="+FloatToString(v)+"\n";
        aWindow->Text(x,y,s);        
    }
    aWindow->SetPenColor(0xFFFF00);
    aWindow->SetFill(false);
    for(map<S32,Object*>::iterator AI=mAgentPosition.begin();AI!=mAgentPosition.end();AI++)
    {
        Object* A=aWM->get(AI->first);
        Object* P=AI->second;
        if (P)
        {
            S32 x,y;
            GetObjectPosition(P,x,y);
            aWindow->Marker(x,y,'o',8);
            string s="["+IntToString(mPositionUpdated[AI->first])+"]:"+IntToString(A->id());
            aWindow->Text(x,y,s);
        }
    }
    
}


// getters
Object* Summary::GetAgentPosition(Humanoid* aAgent)
{
    return mAgentPosition[aAgent->id()];
}

const Path* Summary::GetHighway(Object* aFrom,Object* aTo)
{   
    for(unsigned int i=0;i<mHighway.size();i++)
    {
        if (((mHighwayFrom[i]==aFrom) && (mHighwayTo[i]==aTo)) ||
            ((mHighwayFrom[i]==aTo) && (mHighwayTo[i]==aFrom)))
        {
            return mHighway[i];
        }
    }
    return NULL;
}

float Summary::GetHighwayPriority(Object* aFrom,Object* aTo)
{
    for(unsigned int i=0;i<mHighway.size();i++)
    {
        if (((mHighwayFrom[i]==aFrom) && (mHighwayTo[i]==aTo)) ||
            ((mHighwayFrom[i]==aTo) && (mHighwayTo[i]==aFrom)))
        {
            return mHighwayPriority[i];
        }
    }
    return 0.0f;
}




