// $Id: Summary.h,v 1.1.1.1 2003/07/04 17:15:03 darseq Exp $
#ifndef SUMMARY_H
#define SUMMARY_H

#include "main.h"
#include "Message.h"

using namespace Rescue;

class SummaryTeam : public  vector<Object*> {
    public:
        enum tTeamFunction
        {
            cFunctionFire,
            cFunctionRescue,
            cFunctionHighways,
            cFunctionMax
        };
    protected:
        tTeamFunction mFunction;
    public:
        SummaryTeam(tTeamFunction aFunction)
        {
            mFunction=aFunction;
        }
        
        bool Contains(Object* aAgent) const;
        void Add(Object* aAgent);
        tTeamFunction Function() const
        {
            return mFunction;
        }
};

typedef unsigned char byte;
typedef unsigned short word;
class Summary {
    private:
        // SectorInf lst format: ((SI,F,X),(SI,F,X),...)
        // SI = sectorIndex, F = fireness, X = something
        void crSectorFire(Eltlst * lst,byte* aBuffer);
        void crSectorIntel(Eltlst * lst,byte* aBuffer);
        void crSectorDmg(Eltlst * lst,byte* aBuffer);
        void crSectorRoads(Eltlst * lst,byte* aBuffer);
        // Position lst format: ((AI,ID),(AI,ID),...)
        void crPosition(Eltlst * lst,byte* aBuffer);
        // Highway lst format: ?
        void crHighway(Eltlst * lst,byte* aBuffer);
    protected:
        // fire percentage for all sectors
        map < class Sector *, int > mSectorIndex;
        vector < float > mFire;
        vector < float > mIntel;
        vector < float > mDmg;
        vector < float > mRoads;
        
        vector < S32 > mFireUpdated;
        vector < S32 > mIntelUpdated;
        vector < S32 > mDmgUpdated;
        vector < S32 > mRoadsUpdated;

        // positions of agents 0 is unknown
        map < S32, Object *  > mAgentPosition;
        map < S32, S32 > mPositionUpdated;


        // highway parts
        vector < const class Path *>mHighway;
        vector < Object*> mHighwayFrom;
        vector < Object*> mHighwayTo;
        
        // blocked bool for all nexus-nexus paths
        vector < bool > mHighwayBlock;
        // float of importance of a highway part
        vector < float >mHighwayPriority;

        // precomputed teams
        vector < SummaryTeam * >mTeams;
        class WorldModel* mWorldModel;
    public:
         Summary();
        ~Summary();
        
        // messages for centers, cc/ca msgs
        // +1: enough for terminating NUL byte
        byte sectorA[MSG_SIZE/sizeof(byte)+1];
        byte sectorB[MSG_SIZE/sizeof(byte)+1];
        byte agentA[MSG_SIZE/sizeof(byte)+1];
        byte highwayA[MSG_SIZE/sizeof(byte)+1];

        // Initialize the structure based on the sectors/highways of
        // the given world model
        void Build(class WorldModel * aWorldModel);

        // send all data using a fixed amount of tells
        // update all data according to the information in the object pool
        void centerReset();
        void centerReceive(string ac_message);

        void crSectorA(const byte* aMessage);
        void crSectorB(const byte* aMessage);
        void crAgentA(const byte* aMessage);
        void crHighwayA(const byte* aMessage);

        void arSectorA(const byte* aMessage,S32 aMsgTime);
        void arSectorB(const byte* aMessage,S32 aMsgTime);
        void arAgentA(const byte* aMessage,S32 aMsgTime);
        void arHighwayA(const byte* aMessage,S32 aMsgTime);

        string Update();

        // compute assumed team for this agent
        SummaryTeam *MakeTeam(class BasicAgent * aAgent);
        // compute assumed task for this agent
        class Task *MakeTask(class BasicAgent * aAgent, SummaryTeam * aTeam);
        
        // getters
        Object* GetAgentPosition(Rescue::Humanoid* aActor);
        
        void Draw(class ViewerWindow*,class WorldModel* aWM);
        float GetHighwayPriority(Object* aFrom,Object* aTo);
        const Path* GetHighway(Object* aFrom,Object* aTo);
};
#endif

