#include "Task.h"
#include "SituationModel.h"
#include "behaviour/behaviour.h"
#include "behaviour/behaviourextinguish.h"
#include "behaviour/behaviourrecon.h"
#include "behaviour/behaviourpatrol.h"
#include "behaviour/behaviourtravel.h"
#include "behaviour/behavioursweeppath.h"
#include "behaviour/behaviourfree.h"
#include "behaviour/behaviourclear.h"
#include "behaviour/behaviourrefillwater.h"
#include "behaviour/behaviourdig.h"
#include "Team.h"
#include "WorldModel.h"
#include <memory>

    class DummyFn : public SelectionFunction
    {
    public:
        bool Consider(ImaginaryObject* aObject) const
        {
            return false;
        }
        float Value(ImaginaryObject* aObject) const
        {
            return 0.0f;
        }
    };

    class BurningBuildingFn : public SelectionFunction
    {
    public:
        bool Consider(ImaginaryObject* aObject) const
        {
            ImaginaryBuilding* B=dynamic_cast<ImaginaryBuilding*>(aObject);
            if (B)
            {       
                return B->mFire>0.0;
            }
            else
            {
                return false;
            }
        }
        float Value(ImaginaryObject* aObject) const
        {
            return 0.0f;
        }
    };

        

Behaviours FightFireTask::GenerateBehaviour(TeamMember* aActor,SituationModel* aModel)
{
    Behaviours Out;
    {
        const ImaginaryObjects& burningbuildings=aModel->FindObjects(BurningBuildingFn());
        if (aActor->IsPoliceForce())
        {
            const Team* mTeam=aModel->GetTeam();
            for(vector<TeamMember*>::const_iterator TI=mTeam->begin();TI!=mTeam->end();TI++)
            {
                assert(aModel->GetHuman((*TI)->Self())!=0);
                
                if ((*TI)->IsFireBrigade()) 
                {
                    if ((!aModel->InSector(*TI,mSector))&& ((*TI)->Position()))
                    {


/*
                        const objects_t& nexi=mSector->nexusPoints();
                        for(objects_t::const_iterator ni=nexi.begin();ni!=nexi.end();ni++)
                        {
                            Path* P=aModel->FindPath((*TI)->Human(),aModel->GetObject(*ni),false);
                            if (P)
                            {
                                Out.Add(new BehaviourSweepPath(P));
                            }
                        }
*/
                        Out.Add(new BehaviourFree((*TI)->Self()));
                    }
                } 
            }
            if (!aModel->InSector(aActor,mSector))
            {
                Out.Add(new BehaviourTravel(mSector));
                const objects_t& nexi=mSector->nexusPoints();
                for(objects_t::const_iterator ni=nexi.begin();ni!=nexi.end();ni++)
                {
                    Path* P=aModel->FindPath(aActor->Human(),aModel->GetObject(*ni),false);
                    if (P)
                    {
                        Out.Add(new BehaviourSweepPath(P));
                    }
                }
            }
            else
            {
                Out.Add(new BehaviourRecon(mSector));
                for(ImaginaryObjects::const_iterator I=burningbuildings.begin();I!=burningbuildings.end();I++)
                {
                    // TODO: from teammembers that are fire agent? or from nexus
                    Path* P=aModel->FindPath(aActor->Human(),*I,false);
                    if (P)
                    {
                        Out.Add(new BehaviourSweepPath(P));
                    }
                }
            }
            // TODO: paths to teammembers from nexi?
        }
        else if (aActor->IsFireBrigade())
        {
            Out.Add(new BehaviourRefillWater(aModel->GetClosestRefuge()));
            if (!aModel->InSector(aActor,mSector))
            {
                Out.Add(new BehaviourTravel(mSector));
            }
            else
            {
                Out.Add(new BehaviourRecon(mSector));

                for(ImaginaryObjects::const_iterator I=burningbuildings.begin();I!=burningbuildings.end();I++)
                {
                    Out.Add(new BehaviourExtinguish((Rescue::Building*)(*I)->ActualObject()));
                }
            }
        }
    }   
    return Out;
}

string FightFireTask::ToString()
{
    return string("Fight fire ")+IntToString(mSector->ID());
}

    class BlockedRoadFn : public SelectionFunction
    {
    protected:
        Sector* mSector;
    public:
        bool Consider(ImaginaryObject* aObject) const
        {
            ImaginaryRoad* R=dynamic_cast<ImaginaryRoad*>(aObject);
            if (R)
            {
                if (R->mBlock>0.0)
                {
                    return mSector->insideSector(R->ActualObject());
                }
            }
            return false;
	    
        }
        float Value(ImaginaryObject* aObject) const
        {
            return 0.0f;
        }
        BlockedRoadFn(Sector* aSector)
        {
            mSector=aSector;
        }
    };
     
Behaviours ImproveTravelTask::GenerateBehaviour(TeamMember* aActor,SituationModel* aModel)
{
    Behaviours Out;

    if (aActor->IsPoliceForce())
    {
        // TODO behaviour that scans for detours
       const sectors_t& secs=aModel->WorldModel()->findSectors(aActor->Position()->ActualObject());
       for(sectors_t::const_iterator SI=secs.begin();SI!=secs.end();SI++)
       {
            const objects_t& nexi=(*SI)->nexusPoints();
            for(objects_t::const_iterator NI=nexi.begin();NI!=nexi.end();NI++)
            {
                objects_t::const_iterator NI2=NI;
                NI2++;
                if (NI2==nexi.end())
                {
                    NI2=nexi.begin();
                }
                Out.Add(new BehaviourClear(*NI,*NI2));
            }
       }
    }
    return Out;
}

string ImproveTravelTask::ToString()
{
    return string("Improve ")+IntToString(mSector->ID());
}


Behaviours ReconTask::GenerateBehaviour(TeamMember* aActor,SituationModel* aModel)
{
    Behaviours Out;
    if (aActor->IsFireBrigade())
    {

    }
    if (aActor->IsPoliceForce())
    {

    }
    if (aActor->IsAmbulanceTeam())
    {

    }
    return Out;
}

string ReconTask::ToString()
{
    return string("Recon ")+IntToString(mSector->ID());
}

    class BuriedHumansFn : public SelectionFunction
    {
    public:
        bool Consider(ImaginaryObject* aObject) const
        {
            ImaginaryHuman* H=dynamic_cast<ImaginaryHuman*>(aObject);
            if (H)
            {
                return H->mBurry>0.0;
            }
            else
            {
                return false;
            }
        }
        float Value(ImaginaryObject* aObject) const
        {
            return 0.0f;
        }
    };



Behaviours SearchAndRescueTask::GenerateBehaviour(TeamMember* aActor,SituationModel* aModel)
{
    Behaviours Out;
    if (aActor->IsFireBrigade())
    {

    }
    if (aActor->IsPoliceForce())
    {

    }
    if (aActor->IsAmbulanceTeam())
    {
        if (!aModel->InSector(aActor,mSector))
        {
            Out.Add(new BehaviourTravel(mSector));
        }
        else
        {
            Out.Add(new BehaviourRecon(mSector));
            const ImaginaryObjects& buriedhumans=aModel->FindObjects(BuriedHumansFn());

            for(ImaginaryObjects::const_iterator I=buriedhumans.begin();I!=buriedhumans.end();I++)
            {
                Out.Add(new BehaviourDig((Rescue::Humanoid*)(*I)->ActualObject()));
            }
        }

    }
    return Out;
}

string SearchAndRescueTask::ToString()
{
    return string("Rescue ")+IntToString(mSector->ID());
}


Behaviours PatrolTask::GenerateBehaviour(class TeamMember* aActor,class SituationModel* aSituation)
{
    Behaviours Out;
    Out.Add(new BehaviourPatrol());
    if (aActor->IsPoliceForce())
    {
        const Team* mTeam=aSituation->GetTeam();
        for(vector<TeamMember*>::const_iterator TI=mTeam->begin();TI!=mTeam->end();TI++)
        {
            assert(aSituation->GetHuman((*TI)->Self())!=0);
                
            if ((*TI)->Position())
            {
                if ((*TI)->IsFireBrigade())
                {
                    Out.Add(new BehaviourFree((*TI)->Self()));
                }
            }
        }
    }
    return Out;
}


string PatrolTask::ToString()
{
    return string("Patrol ");
}

