#include <X11/keysym.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include "ViewerWindow.h"
#include "main.h"
#include "path.h"
#include "BasicActor.h"
#include <adk.hxx>

/*
 * Open and closing
 */
     
void ViewerWindow::Create(int width,int height)
{
    char *icon_name = "window";
    char *default_title="Viewer Window";
    XSizeHints size_hints;
    Window rootwin;

    mWidth=width;mHeight=height;

    mTitle=new char[strlen(default_title)+1];
    strcpy(mTitle,default_title);

    mDisplay = XOpenDisplay(getenv("DISPLAY"));

    if (mDisplay == NULL)
    {
        fprintf(stderr, "Failed to open display\n");
        return;
    }

    mScreen = DefaultScreen(mDisplay);
    mDepth = DefaultDepth(mDisplay, mScreen);
    rootwin = RootWindow(mDisplay, mScreen);
    mWindow = XCreateSimpleWindow(mDisplay, rootwin, 10, 10, mWidth, mHeight, 5,
        BlackPixel(mDisplay, mScreen), mBackgroundColor);

    size_hints.flags = PSize | PMinSize | PMaxSize;
    size_hints.min_width=600;
    size_hints.min_height=400;
    size_hints.max_width=size_hints.max_height=1000;
    
    XSetWindowBackgroundPixmap(mDisplay,mWindow,None);
    
    XSetStandardProperties(mDisplay, mWindow, mTitle, icon_name, None,
        0, 0, &size_hints); 

    XSelectInput(mDisplay, mWindow, ButtonPressMask | KeyPressMask);
    XMapWindow(mDisplay, mWindow);

    
    XGCValues gc_values;

    mGC=XCreateGC(mDisplay,mWindow,0,&gc_values);

    mHandle=0;
    SetFont("8x16",16);

}

void ViewerWindow::Close()
{
    if (mFont) XFreeFont(mDisplay,mFont);
    mFont=0;

    if (mBuffer) XFreePixmap(mDisplay,mBuffer);
    XFreeGC(mDisplay,mGC);
    XCloseDisplay(mDisplay);
    
    mDisplay=0;
    mWindow=0;
    mHandle=0;
}

/*
 * Events
 */
 
void ViewerWindow::BeginFrame()
{
    if (!mClosed)
    {    
        // check for resize

        XWindowAttributes attributes;
        XGetWindowAttributes (mDisplay,mWindow,&attributes);
        if (((int)mWidth!=attributes.width) || ((int)mHeight!=attributes.height))
        {
            mWidth=attributes.width;
            mHeight=attributes.height;
            // if resized, adjust all registered views
            for(vector<View*>::iterator VI=mViews.begin();VI!=mViews.end();VI++)
            {
                (*VI)->Resize(mWidth,mHeight);
            }
            // destroy old buffer and create new one (slow) but automatically resizes and clears
            if (mBuffer)
            {
                XFreePixmap(mDisplay,mBuffer);
                mBuffer=0;
            }
        }
        else
        {
            if (mBuffer)
            {
                // clear clipper
                XSetClipMask(mDisplay,mGC,None);
                // change color
                XGCValues values,oldvalues;
                XGetGCValues(mDisplay,mGC,GCForeground,&oldvalues);
                values.foreground=mBackgroundColor;
                XChangeGC(mDisplay,mGC,GCForeground,&values);
                // clear
                XFillRectangle(mDisplay,mBuffer,mGC,0,0,mWidth,mHeight);
                // restore color
                XChangeGC(mDisplay,mGC,GCForeground,&oldvalues);
                // restore view
                mCurrentView->Select(mDisplay,mGC);
            }
        }
    }
    XEvent xev;
    int num_events;

    // draw directly to window during events
    mHandle=mWindow;
    
//    XFlush(mDisplay);
    num_events = XPending(mDisplay);
    while((num_events != 0))
    {
        num_events--;
        XNextEvent(mDisplay, &xev);
        ProcessEvent(xev);
    }

    // set drawing back to (possibly new) buffer
    if (!mBuffer)
    {
        mBuffer=XCreatePixmap(mDisplay,mWindow,mWidth,mHeight,mDepth);
    }
    mHandle=mBuffer;
    
}

void ViewerWindow::EndFrame()
{
    // clear clipper
    XSetClipMask(mDisplay,mGC,None);
    // now clear the window to blit to
//    XClearWindow(mDisplay,mWindow);
    // and blit    
    XCopyArea(mDisplay,mBuffer,mWindow,mGC,0,0,mWidth,mHeight,0,0);
    // Restore clipper
    mCurrentView->Select(mDisplay,mGC);

/*
    // fill entire window with background
    unsigned long oldcolor;
    XGCValues values;
    XGetGCValues(mDisplay,mGC,GCForeground,&values);
    oldcolor=values.foreground;
    values.foreground=mBackgroundColor;
    XChangeGC(mDisplay,mGC,GCForeground,&values);

    XFillRectangle(mDisplay,mWindow,mGC,0,0,mWidth,mHeight);

    // restore color
    values.foreground=oldcolor;
    XChangeGC(mDisplay,mGC,GCForeground,&values);
    // Restore clipper
    mCurrentView->Select(mDisplay,mGC);
*/


}


void ViewerWindow::ProcessEvent(XEvent report)
{
    KeySym key;

    switch(report.type)
    {
        case KeyPress:
            key = XLookupKeysym(&report.xkey, 0);
            switch(key)
            {
                case XK_KP_Divide:   // '/': prev actor
                {
                    for(vector<BasicActor*>::iterator I=gActors.begin();I!=gActors.end();I++)
                    {
                        if (*I==mActor)
                        {
                            if (I==gActors.begin())
                            {
                                I=gActors.end();
                                I--;
                                mActor=*I;
                            }
                            else
                            {
                                I--;
                                mActor=*I;
                            }
                        }
                    }
                }
                break;
                case XK_KP_Multiply: // '*': next actor
                {
                    for(vector<BasicActor*>::iterator I=gActors.begin();I!=gActors.end();I++)
                    {
                        if (*I==mActor)
                        {
                            I++;
                            if (I==gActors.end())
                            {
                                I=gActors.begin();
                                mActor=*I;
                            }
                            else
                            {
                                mActor=*I;
                            }
                            break;
                        }
                     }
                  }
                break;
                case XK_F1:
                case XK_F2:
                case XK_F3:
                case XK_F4:
                case XK_F5:
                case XK_F6:
                case XK_F7:
                case XK_F8:
                case XK_F9:
                case XK_F10:
                case XK_F11:
                case XK_F12:
                {
                    int k=key;
                    int s=report.xkey.state;
                    if (s&1) k+=12;
                    if (s&8) k+=24;
                    mViewFlags^=1<<(k-XK_F1);   
                }
                break;
                default:
                    Actor()->Key(key);
                break;
            }
            break;

        case ButtonPressMask:
        {

            for(vector<View*>::iterator I=mViews.begin();I!=mViews.end();I++)
            {
                if ((*I)->Inside(report.xbutton.x,report.xbutton.y))
                {
                    printf("Mouse button %d in %s at (%f,%f)\n",
                        report.xbutton.button,
                        (*I)->Name(),
                        (*I)->ConvertViewToX(report.xbutton.x),
                        (*I)->ConvertViewToY(report.xbutton.y));
                }
                else
                {
                    printf("Mouse button %d outside %s\n",
                        report.xbutton.button,
                        (*I)->Name());
                }
            }
            Actor()->MouseClick(report.xbutton.x,report.xbutton.y,report.xbutton.button);
        }
        break;
        case DestroyNotify:
        {
            mClosed=true;
        }
        break;
        default:
            break;
    }
}

/*
 * Changing of window properties
 */
 
void ViewerWindow::SetTitle(const char* aTitle)
{
    delete mTitle;
    mTitle=0;
    if (aTitle)
    {
        mTitle=new char[strlen(aTitle)+1];
        strcpy(mTitle,aTitle);
        XTextProperty Text;
        XGetWMName(mDisplay,mWindow,&Text);
        Text.value=(unsigned char*)mTitle;
        Text.format=8;
        Text.encoding= XA_STRING;
        Text.nitems=strlen(mTitle);
        XSetWMName(mDisplay,mWindow,&Text);
    }
    
}

void ViewerWindow::SetBackgroundColor(unsigned long aRGB)
{
XSetWindowAttributes attributes;
    attributes.background_pixel=aRGB;
    XChangeWindowAttributes(mDisplay,mWindow,CWBackPixel,&attributes);
    mBackgroundColor=aRGB;
}

void View::SetName(const char* aName)
{
    if (mName)
    {
        delete mName;
        mName=0;
    }
    if (aName)
    {
        mName=strdup(aName);
    }
    else
    {
        mName=strdup("<view>");
    }
}

void View::SetView(int aX1,int aY1,int aX2,int aY2)
{
    mViewX1=aX1;
    mViewY1=aY1;
    mViewX2=aX2;
    mViewY2=aY2;
    Recompute();
}

void View::Resize(int aWidth,int aHeight)
{
    mWidth=aWidth;
    mHeight=aHeight;
    Recompute();
};


void View::Recompute()
{
    mClipX1=mViewX1;
    mClipY1=mViewY1;
    if (mViewX2<0)
    {
        mClipX2=1+mWidth+mViewX2;
    }
    else
    {
        mClipX2=mViewX2;
    }
    if (mViewY2<0)
    {
        mClipY2=1+mHeight+mViewY2;
    }
    else
    {
        mClipY2=mViewY2;;
    }

    if (mSquare)
    {
        int bordersize;
        bordersize=(mClipX2-mClipX1)-(mClipY2-mClipY1);
        if (bordersize>0)
        {
            mClipX1+=bordersize/2;
            mClipX2-=bordersize/2;
        }
        else
        {
            mClipY1-=bordersize/2;
            mClipY2+=bordersize/2;
        }
    }
}

int View::ConvertXToView(double aX)
{
    int width=(mClipX2-(mClipX1+1));
    if (!mMirrorX)
    {
        return (int)(mClipX1+((aX+mScrollX)*(mZoom*width)));
    }
    else
    {
        return (int)(mClipX2-((aX+mScrollX)*(mZoom*width)));
    }

}

int View::ConvertYToView(double aY)
{
    int height=(mClipY2-(mClipY1+1));
    if (!mMirrorY)
    {
        return (int)(mClipY1+((aY+mScrollY)*(mZoom*height)));
    }
    else
    {
        return (int)(mClipY2-((aY+mScrollY)*(mZoom*height)));
    }
}

double View::ConvertViewToX(int aX)
{
    int width=(mClipX2-(mClipX1+1));
    if (!mMirrorX)
    {
        return (double)(((aX-mClipX1)/(width*mZoom))-mScrollX);
    }
    else
    {
         return (double)(((mClipX2-aX)/(width*mZoom))-mScrollX);
    }

}


double View::ConvertViewToY(int aY)
{
    int height=(mClipY2-(mClipY1+1));
    if (!mMirrorY)
    {
        return (double)(((aY-mClipY1)/(height*mZoom))-mScrollY);
    }
    else
    {
        return (double)(((mClipY2-aY)/(height*mZoom))-mScrollY);
    }
}



void View::Select(Display *aDisplay,GC aGC)
{
    XRectangle rect;
    rect.x=mClipX1;
    rect.y=mClipY1;
    rect.width=mClipX2-mClipX1;
    rect.height=mClipY2-mClipY1;
    XSetClipRectangles(aDisplay,aGC,0,0,&rect,1,Unsorted);
}

View* ViewerWindow::CreateView(const char* aName)
{
    View* O=new View;
    mViews.push_back(O);
    if (aName)
    {
        O->SetName(aName);
    }
    O->Resize(mWidth,mHeight);
    return O;
}

void ViewerWindow::SelectView(View* aView)
{
    if (aView)
    {
        mCurrentView=aView;
    }
    else
    {
        mCurrentView=mDefaultView;
    }
    mCurrentView->Select(mDisplay,mGC);
}
    
void ViewerWindow::SelectPenColor()
{
    XGCValues values;
    values.foreground=mPenColor;
    XChangeGC(mDisplay,mGC,GCForeground,&values);
}

void ViewerWindow::SelectFillColor()
{
    XGCValues values;
    values.foreground=mFillColor;
    XChangeGC(mDisplay,mGC,GCForeground,&values);
}

void ViewerWindow::SelectFillColorBack()
{
    XGCValues values;
    values.background=mFillColor;
    XChangeGC(mDisplay,mGC,GCBackground,&values);
}


void ViewerWindow::SetFont(const char* aName,int aSize)
{
    if (mFont)
    {
        XFreeFont(mDisplay,mFont);
        mFont=0;
    }
    char **names;
    char mask[1024];
    int returned;

    // find matching name with closest matching size
    sprintf(mask,"*%s*",aName);
    names=XListFonts (mDisplay,mask,100,&returned);
    XFontStruct* font;
    int difference,bestmatch=65535;
    for(int i=0;i<returned;i++)
    {
        font=XLoadQueryFont(mDisplay,names[i]);
        difference=aSize-(font->ascent);//+font->descent);
        if ((difference>=0) && (difference<bestmatch))
        {
            if (mFont)
            {
                XFreeFont(mDisplay,mFont);
            }
            mFont=font;
            bestmatch=difference;
        }
        else
        {
            XFreeFont(mDisplay,font);
        }
    }
    XFreeFontNames(names);
    
    XGCValues values;
    if (mFont)
    {
        values.font=mFont->fid;
        XChangeGC(mDisplay,mGC,GCFont,&values);
    }
    else
    {
        values.font=None;
    }
    
}


void ViewerWindow::Line(double aX1,double aY1,double aX2,double aY2,int aWidth,int aStyle)
{
    // compute pixel coordinates
    int x1,y1,x2,y2;
    x1=mCurrentView->ConvertXToView(aX1);
    y1=mCurrentView->ConvertYToView(aY1);
    x2=mCurrentView->ConvertXToView(aX2);
    y2=mCurrentView->ConvertYToView(aY2);

    SelectPenColor();
    // set line style/width
    XGCValues oldvalues,values;
    XGetGCValues(mDisplay,mGC,GCLineWidth|GCLineStyle,&oldvalues);
    
    values.line_width=aWidth;
    values.line_style=aStyle;
    XChangeGC(mDisplay,mGC,GCLineWidth|GCLineStyle,&values);
     // draw
    XDrawLine(mDisplay,mHandle,mGC,x1,y1,x2,y2);
    // restore line style/width
    XChangeGC(mDisplay,mGC,GCLineWidth|GCLineStyle,&oldvalues);
}

void ViewerWindow::Rectangle(double aX1,double aY1,double aX2,double aY2)
{
int x1,y1,x2,y2;
    x1=mCurrentView->ConvertXToView(aX1);
    y1=mCurrentView->ConvertYToView(aY1);
    x2=mCurrentView->ConvertXToView(aX2);
    y2=mCurrentView->ConvertYToView(aY2);

    SelectPenColor();
    if (mFill)
    {
        SelectFillColor();
        XFillRectangle(mDisplay,mHandle,mGC,x1,y1,x2-x1,y2-y1);
    }
    XDrawRectangle(mDisplay,mHandle,mGC,x1,y1,x2-x1,y2-y1);
}
void ViewerWindow::Plot(double aX,double aY)
{
int x,y;
    x=mCurrentView->ConvertXToView(aX);
    y=mCurrentView->ConvertYToView(aY);

    SelectPenColor();

    XDrawPoint(mDisplay,mHandle,mGC,x,y);
}

void ViewerWindow::Text(double aX,double aY,const string& aText)
{
int x,y;
    if (mFont)
    {
        const char *p;
        int l;
        char buffer[1024];
        p=aText.c_str();
        x=mCurrentView->ConvertXToView(aX);
        y=mCurrentView->ConvertYToView(aY)+mFont->ascent;
        SelectPenColor();
        while (true)
        {
            l=0;
            while ((p[l]!=0) && (p[l]!='\n'))
            {
                l++;
            }
            strncpy(buffer,p,l);
            buffer[l]=0;
            
            if (mFill)
            {
                SelectFillColorBack();
                XDrawImageString(mDisplay,mHandle,mGC,x,y,p,l);
            }
            else
            {
                XDrawString(mDisplay,mHandle,mGC,x,y,p,l);
            }
            y+=mFont->ascent+mFont->descent;
            if (p[l]==0)
                break;
            p+=(l+1);
        }
    }
}


void ViewerWindow::Polygon(int aN,double* aVertices)
{
    XPoint* p;
    p=new XPoint[aN];
    for(int i=0;i<aN;i++)
    {
        p[i].x=mCurrentView->ConvertXToView(aVertices[i*2]);
        p[i].y=mCurrentView->ConvertYToView(aVertices[i*2+1]);
    }
    if (mFill)
    {
        SelectFillColor();
//        XFillPolygon(mDisplay,mHandle,mGC,p,aN,Convex,CoordModeOrigin);
        XFillPolygon(mDisplay,mHandle,mGC,p,aN,Nonconvex,CoordModeOrigin);
    }
    SelectPenColor();
    XDrawLines(mDisplay,mHandle,mGC,p,aN,CoordModeOrigin);
    delete p;
    
}


void ViewerWindow::Marker(double aX,double aY,char aMarker,int aSize)
{
int x,y;
    int s2=2*aSize+1;
    x=mCurrentView->ConvertXToView(aX);
    y=mCurrentView->ConvertYToView(aY);
    switch(aMarker)
    {
        case 'x': // little diagonal cross
        {
            SelectPenColor();
            XDrawLine(mDisplay,mHandle,mGC,x,y,x+aSize,y+aSize);
            XDrawLine(mDisplay,mHandle,mGC,x,y,x+aSize,y-aSize);
            XDrawLine(mDisplay,mHandle,mGC,x,y,x-aSize,y+aSize);
            XDrawLine(mDisplay,mHandle,mGC,x,y,x-aSize,y-aSize);
        }
        break;
        case 'o':
        case 'O': // little circle
        {
            if (mFill)
            {
                SelectFillColor();
                XFillArc(mDisplay,mHandle,mGC,x-aSize,y-aSize,s2,s2,0,64*360);
            }
            SelectPenColor();
            XDrawArc(mDisplay,mHandle,mGC,x-aSize,y-aSize,s2,s2,0,64*360);
        }
        break;
        case '#': // little square
        {
            if (mFill)
            {
                SelectFillColor();
                XFillRectangle(mDisplay,mHandle,mGC,x-aSize,y-aSize,s2,s2);
            }
            SelectPenColor();
            XDrawRectangle(mDisplay,mHandle,mGC,x-aSize,y-aSize,s2,s2);
        }
        break;
        case '.': // largish dot
            SelectPenColor();
            XDrawPoint(mDisplay,mHandle,mGC,x,y);
            XDrawPoint(mDisplay,mHandle,mGC,x-1,y);
            XDrawPoint(mDisplay,mHandle,mGC,x+1,y);
            XDrawPoint(mDisplay,mHandle,mGC,x,y-1);
            XDrawPoint(mDisplay,mHandle,mGC,x,y+1);
        break;
        case '~': // tiny flag
            SelectPenColor();
            XDrawLine(mDisplay,mHandle,mGC,x,y,x,y-aSize);
            SelectFillColor();
            XDrawPoint(mDisplay,mHandle,mGC,x+1,y-aSize);
            XDrawPoint(mDisplay,mHandle,mGC,x+1,y-(aSize-1));
            XDrawPoint(mDisplay,mHandle,mGC,x+1,y-(aSize-2));
            XDrawPoint(mDisplay,mHandle,mGC,x+2,y-(aSize-1));
            XDrawPoint(mDisplay,mHandle,mGC,x+2,y-(aSize-2));
            XDrawPoint(mDisplay,mHandle,mGC,x+3,y-(aSize-1));
        break;
        case '=': // large flag
            SelectPenColor();
            XDrawLine(mDisplay,mHandle,mGC,x,y,x,y-aSize);
            SelectFillColor();
            XDrawPoint(mDisplay,mHandle,mGC,x+1,y-aSize);
            XDrawPoint(mDisplay,mHandle,mGC,x+1,y-(aSize-1));
            XDrawPoint(mDisplay,mHandle,mGC,x+1,y-(aSize-2));

            XDrawPoint(mDisplay,mHandle,mGC,x+2,y-(aSize-1));
            XDrawPoint(mDisplay,mHandle,mGC,x+2,y-(aSize-2));
            XDrawPoint(mDisplay,mHandle,mGC,x+2,y-(aSize-3));

            XDrawPoint(mDisplay,mHandle,mGC,x+3,y-(aSize-1));
            XDrawPoint(mDisplay,mHandle,mGC,x+3,y-(aSize-2));
            XDrawPoint(mDisplay,mHandle,mGC,x+3,y-(aSize-3));

            XDrawPoint(mDisplay,mHandle,mGC,x+4,y-aSize);
            XDrawPoint(mDisplay,mHandle,mGC,x+4,y-(aSize-1));
            XDrawPoint(mDisplay,mHandle,mGC,x+4,y-(aSize-2));

            XDrawPoint(mDisplay,mHandle,mGC,x+5,y-7);
            XDrawPoint(mDisplay,mHandle,mGC,x+5,y-(aSize-1));
            XDrawPoint(mDisplay,mHandle,mGC,x+5,y-(aSize-2));

        break;
        
    }
}

