// $Id: WorldModel.cpp,v 1.1.1.1 2003/07/04 17:15:03 darseq Exp $
// copyright, s.post & cp

#include "WorldModel.h"
#include "ViewerWindow.h"
#include "path.h"
#include "Summary.h"

// find, find_list, find_path,
// update_path,verify_path

struct HeuristicFn : public Memory::DistanceObjectFn {
    double f(const Object *o) const {return 0;}
    HeuristicFn(const Object *o) : Memory::DistanceObjectFn(o) {}
};

//----------------------------------------------------------------------------

struct DistanceNodeFn : public Memory::DistanceObjectFn {
    bool consider(const Object * o) const {return o->type() == TYPE_NODE;}
    DistanceNodeFn(double x,double y) : Memory::DistanceObjectFn(x,y) {}
};

//----------------------------------------------------------------------------

struct DistanceNodeXFn : public Memory::DistanceObjectFn {
    bool consider(const Object * o) const {return o->type() == TYPE_NODE;}
    double f(const Object *o) const {
        double x = o->x(); double y = o->y();
        return (CORRECTION * (x - m_x) * (x - m_x) + (y - m_y) * (y - m_y));
    }
    DistanceNodeXFn(double x,double y) : Memory::DistanceObjectFn(x,y) {}
};

//----------------------------------------------------------------------------

struct DistanceNodeYFn : public Memory::DistanceObjectFn {
    bool consider(const Object * o) const {return o->type() == TYPE_NODE;}
    double f(const Object *o) const {
        double x = o->x(); double y = o->y();
        return ((x - m_x) * (x - m_x) + CORRECTION * (y - m_y) * (y - m_y));
    }
    DistanceNodeYFn(double x,double y) : Memory::DistanceObjectFn(x,y) {}
};
//----------------------------------------------------------------------------

struct BorderTravelFn : public Memory::TraversalObjectPairFn {
    bool consider(const Object * a, const Object * b) const {
        if (((a->type() == TYPE_NODE) && (b->type() == TYPE_ROAD)) ||
            ((a->type() == TYPE_ROAD) && (b->type() == TYPE_NODE))) {
            return Memory::TraversalObjectPairFn::consider(a,b);
        } else {return false;}
    }
};
//----------------------------------------------------------------------------

struct TargetNodeFn : public Memory::IsObjectObjectFn {
    TargetNodeFn(Object* aObj) : Memory::IsObjectObjectFn(aObj) {}
};
//----------------------------------------------------------------------------

WorldModel::WorldModel(int aSectorX,int aSectorY) {
/*
    if (aSectorX<4) aSectorX=4;
    if (aSectorY<4) aSectorY=4;
    if (aSectorX>MAXSECTORS) aSectorX=MAXSECTORS;
    if (aSectorY>MAXSECTORS) aSectorY=MAXSECTORS;
*/
    
    mSectorX=aSectorX;
    mSectorY=aSectorY;

    mSummary=new Summary;    
}
//----------------------------------------------------------------------------

WorldModel::~WorldModel(void)
{
    delete mSummary;
}
//----------------------------------------------------------------------------

void Sector::doCalcDimensions(const objects_t& aBorder) {
    for(objects_t::const_iterator I=aBorder.begin();I!=aBorder.end();I++) {
        S32 x,y;
        GetObjectPosition(*I,x,y);
        if (x<mMinX) mMinX=x;
        if (x>mMaxX) mMaxX=x;
        if (y<mMinY) mMinY=y;
        if (y>mMaxY) mMaxY=y;
    }
}
//----------------------------------------------------------------------------
        
void Sector::calcDimensions() {
    mMinX=MAXLONG;mMaxX=MINLONG;
    mMinY=MAXLONG;mMaxY=MINLONG;
    doCalcDimensions(path01);
    doCalcDimensions(path13);
    doCalcDimensions(path32);
    doCalcDimensions(path20);
}
//----------------------------------------------------------------------------

void WorldModel::analyseMap(void) {
    long timestart,timeend;
    timestart=timer();
    
    bool point_value = false;
    Objects::iterator i = m_objects.begin();
    for(;i != m_objects.end();i++) {
        Node * node = dynamic_cast<Node *>(*i);
        if(node) {
            if(point_value == false) {
                point_value = true;
                x1 = x2 = node->x();
                y1 = y2 = node->y();
                continue;
            }
            if(node->x() < x1) {x1 = node->x();}
            if(node->x() > x2) {x2 = node->x();}
            if(node->y() < y1) {y1 = node->y();}
            if(node->y() > y2) {y2 = node->y();}
        }
    }
/*
    Debug(spam_startup,"min x: %ld, max x: %ld,"
        "min y: %ld, max y: %ld\n",x1,x2,y1,y2);
*/
    sectors[0][0].location = sectors[mSectorY-1][mSectorX-1].location = CORNER;
    sectors[mSectorY-1][0].location = sectors[0][mSectorX-1].location = CORNER;

    Object * n[4]; // nodes
    S32 ivalx = (x2-x1)/(mSectorX-2);
    S32 ivaly = (y2-y1)/(mSectorY-2);

    // assign IDs
    int id=1;
    for(int j=0;j < mSectorY;j++) {
        for(int i=0;i < mSectorX;i++) {
            sectors[j][i].mID=id++;
        }
    }
        
    for(int j=1;j <= mSectorY-2;j++) {
        for(int i=1;i <= mSectorX-2;i++) {
            S32 gx1 = x1+(i-1)*ivalx;
            S32 gy1 = y1+(j-1)*ivaly;
            S32 gx2 = gx1+ivalx; S32 gy2 = gy1+ivaly;
            n[0] = find(DistanceNodeFn(gx1,gy1),FUNCTION_MINIMIZE);
            n[1] = find(DistanceNodeFn(gx2,gy1),FUNCTION_MINIMIZE);
            n[2] = find(DistanceNodeFn(gx1,gy2),FUNCTION_MINIMIZE);
            n[3] = find(DistanceNodeFn(gx2,gy2),FUNCTION_MINIMIZE);

            if(i == 1) {
                if(j == 1) {
                    n[1] = find(DistanceNodeXFn(gx2,gy1),FUNCTION_MINIMIZE);
                    n[2] = find(DistanceNodeYFn(gx1,gy2),FUNCTION_MINIMIZE);
                    sectorPaths(i,j,n); borderPaths(i,j,n,SOUTH);
                } else if(j == mSectorY-2) {
                    n[0] = find(DistanceNodeYFn(gx1,gy1),FUNCTION_MINIMIZE);
                    n[3] = find(DistanceNodeXFn(gx2,gy2),FUNCTION_MINIMIZE);
                    sectorPaths(i,j,n); borderPaths(i,j,n,NORTH);
                } else {
                    n[0] = find(DistanceNodeYFn(gx1,gy1),FUNCTION_MINIMIZE);
                    n[2] = find(DistanceNodeYFn(gx1,gy2),FUNCTION_MINIMIZE);
                    sectorPaths(i,j,n);
                }
                borderPaths(i,j,n,WEST);
            } else if(i == mSectorX-2) {
                if(j == 1) {
                    n[0] = find(DistanceNodeXFn(gx1,gy1),FUNCTION_MINIMIZE);
                    n[3] = find(DistanceNodeYFn(gx2,gy2),FUNCTION_MINIMIZE);
                    sectorPaths(i,j,n); borderPaths(i,j,n,SOUTH);
                } else if(j == mSectorY-2) {
                    n[1] = find(DistanceNodeYFn(gx2,gy1),FUNCTION_MINIMIZE);
                    n[2] = find(DistanceNodeXFn(gx1,gy2),FUNCTION_MINIMIZE);
                    sectorPaths(i,j,n); borderPaths(i,j,n,NORTH);
                } else {
                    n[1] = find(DistanceNodeYFn(gx2,gy1),FUNCTION_MINIMIZE);
                    n[3] = find(DistanceNodeYFn(gx2,gy2),FUNCTION_MINIMIZE);
                    sectorPaths(i,j,n);
                }
                borderPaths(i,j,n,EAST);
            } else if(j == 1) {
                n[0] = find(DistanceNodeXFn(gx1,gy1),FUNCTION_MINIMIZE);
                n[1] = find(DistanceNodeXFn(gx2,gy1),FUNCTION_MINIMIZE);
                sectorPaths(i,j,n); borderPaths(i,j,n,SOUTH);
            } else if(j == mSectorY-2) {
                n[2] = find(DistanceNodeXFn(gx1,gy2),FUNCTION_MINIMIZE);
                n[3] = find(DistanceNodeXFn(gx2,gy2),FUNCTION_MINIMIZE);
                sectorPaths(i,j,n); borderPaths(i,j,n,NORTH);
            } else {sectorPaths(i,j,n);}
/*
            Debug(spam_startup,"(%d,%d), node1:%d, node2:%d, "
                "node3:%d, node4:%d\n",i,j,n[0]->id(),n[1]->id(),
                    n[2]->id(),n[3]->id());
*/
            sectors[j][i].calcDimensions();
        }
    }

    timeend=timer();
    Debug(spam_startup|spam_performance,"WorldModel::Analyze() Made sectors in %ldms\n",timeend-timestart);
    mSectorTime=timeend-timestart;
    timestart=timeend;
    
    Objects::iterator o = m_objects.begin();
    for(;o != m_objects.end();o++) {
        if(dynamic_cast<AmbulanceTeam *>(*o)) {
            a_agents.insert((AmbulanceTeam *) *o);
        } else if(dynamic_cast<FireBrigade *>(*o)) {
            f_agents.insert((FireBrigade *) *o);
        } else if(dynamic_cast<PoliceForce *>(*o)) {
            p_agents.insert((PoliceForce *) *o);
        } else {
            int result = 0; // 0 or -1: try next sector
            int sector_count = 0;
            for(int j=0;result != 1 && j < mSectorY;j++) {
                for(int i=0;result != 1 && i < mSectorX;i++) {
                    if(sectors[j][i].location != CORNER) {
                        result = sectors[j][i].trySector(*o);
                        if(result) {
                            sectors[j][i].objects.push_back(*o);
                            ++sector_count;
                        }
                    }
                }
            }
            //                      __   __
            // a sector may have : |  \_/  |
            // overlapping paths   |__/ \__|
            if(sector_count == 0) { // || sector_count > 4) {
                Debug(spam_startup,"Warning: object %ld assigned to %d "
                    "sectors\n",(*o)->id(),sector_count);
            }
        }
    }

    timeend=timer();
    Debug(spam_startup|spam_performance,"WorldModel::Analyze() Assigned objects to sectors in %ldms\n",
        timeend-timestart);
    mSectorTime+=timeend-timestart;
    mSectorBytes=0;
    timestart=timeend;

    PrecomputeInternalPaths();

    timeend=timer();
    Debug(spam_startup|spam_performance,"WorldModel::Analyze() Made internal paths in %ldms\n",
        timeend-timestart);
    mPathTime=timeend-timestart;
    timestart=timeend;

    PrecomputeHighways();

    timeend=timer();
    Debug(spam_startup|spam_performance,"WorldModel::Analyze() Made highways in %ldms\n",
        timeend-timestart);
    mHighwayTime=timeend-timestart;
    timestart=timeend;


    mSummary->Build(this);
    timeend=timer();
    Debug(spam_startup|spam_performance,"WorldModel::Analyze() Made summary in %ldms\n",
        timeend-timestart);
    mSummaryTime=timeend-timestart;
    timestart=timeend;        

    // find all agents in the map

    const Objects &objs = objects();
    for(Objects::const_iterator it = objs.begin();it != objs.end();it++) {
        Building * building;
        if((dynamic_cast<AmbulanceCenter *>(*it) ||
            dynamic_cast<PoliceOffice *>(*it) ||
            dynamic_cast<FireStation *>(*it)) &&
            (building = dynamic_cast<Building *>(*it))) {
            mCenters.push_back(building);
        }
        Humanoid* human;
        if((dynamic_cast<AmbulanceTeam *>(*it) ||
            dynamic_cast<PoliceForce *>(*it) ||
            dynamic_cast<FireBrigade *>(*it)) &&
            (human = dynamic_cast<Humanoid *>(*it))) {
            mPlatoonAgents.push_back(building);
        }
    }
}
//----------------------------------------------------------------------------

void path2poly(const objects_t * path,int & count,S32* poly_x,S32* poly_y,S32* points) {

    Node * N; // ignore road objects
    for(unsigned int i=0;i < path->size();i++) {
        if((N = dynamic_cast<Node *>((*path)[i]))) {
            if(count == MAXPOINTS) {
                Debug(spam_startup,"Error: maxpoints exceeded\n");
                break;
            }
            poly_x[count] = N->x();
            poly_y[count] = N->y();
            points[count*2+0] = N->x();
            points[count*2+1] = N->y();
            ++count;
        }
    }
}
//----------------------------------------------------------------------------

int Sector::inSector(S32 x,S32 y) const {
    S32 poly_x[MAXPOINTS];
    S32 poly_y[MAXPOINTS];
    S32 points[2*MAXPOINTS];

    int npol = 0;
    if ((x >= mMinX) && (x <= mMaxX) &&
        (y >= mMinY) && (y <= mMaxY)) {
        path2poly(&path01,npol,poly_x,poly_y,points);
        path2poly(&path13,npol,poly_x,poly_y,points);
        path2poly(&path32,npol,poly_x,poly_y,points);
        path2poly(&path20,npol,poly_x,poly_y,points);

        int i,j,c = 0;
        for (i=0,j = npol-1;i < npol;j = i++) {
            if((x == poly_x[i] && y == poly_y[i]) ||
                x == poly_x[j] && y == poly_y[j]) {
                return -1;
            } else if ((((poly_y[i] <= y) && (y < poly_y[j])) ||
                ((poly_y[j] <= y) && (y < poly_y[i]))) &&
                (x < (poly_x[j] - poly_x[i]) * (double)
                (y - poly_y[i]) / (poly_y[j] - poly_y[i]) + poly_x[i])) {
                c = !c;
            }
        }
        return c;
    } else {return 0;}
};
//----------------------------------------------------------------------------

int Sector::trySector(Object* obj) const
{
    int result = 0; // assume object has been found
    Edge * E; PointObject * P; // P is Node or Building
    MovingObject* M;
    if((P = dynamic_cast<PointObject *>(obj))) {
        result = inSector(P->x(),P->y());
    } else if((E = dynamic_cast<Edge *>(obj))) {
        PointObject * P1 = (PointObject *) E->head();
        PointObject * P2 = (PointObject *) E->tail();
        result = inSector(P1->x(),P1->y()) & // no -> && !!
                 inSector(P2->x(),P2->y());
    } else if ((M=dynamic_cast<MovingObject*>(obj))) {
        result=insideSector(M->position());
    } else if(obj && dynamic_cast<World *>(obj)==NULL) {
        Debug(spam_spam,"Unknown object %ld\n",obj->type());
    }
    return result;
}
//----------------------------------------------------------------------------
    
void WorldModel::borderPaths(int i,int j,Object * n[4],location_t location) {
    Sector * border;
    Node * vnod1 = new Node();
    Node * vnod2 = new Node();

    switch(location) {
        case NORTH:
            border = &sectors[j+1][i];
            border->path01 = sectors[j][i].path23;
            vnod1->setX(0,n[2]->x()); vnod1->setY(0,y2);
            vnod2->setX(0,n[3]->x()); vnod2->setY(0,y2);

            border->path13.push_back(vnod2);
            border->path32.push_back(vnod1);
            border->path20.push_back(n[2]);
        break;
        case SOUTH:
            border = &sectors[j-1][i];
            border->path32 = sectors[j][i].path10;
            vnod1->setX(0,n[0]->x()); vnod1->setY(0,y1);
            vnod2->setX(0,n[1]->x()); vnod2->setY(0,y1);

            border->path01.push_back(vnod2);
            border->path13.push_back(n[1]);
            border->path20.push_back(vnod1);
        break;
        case EAST:
            border = &sectors[j][i+1];
            border->path20 = sectors[j][i].path31;
            vnod1->setX(0,x2); vnod1->setY(0,n[3]->y());
            vnod2->setX(0,x2); vnod2->setY(0,n[1]->y());

            border->path01.push_back(vnod2);
            border->path13.push_back(vnod1);
            border->path32.push_back(n[3]);
        break;
        case WEST:
            border = &sectors[j][i-1];
            border->path13 = sectors[j][i].path02;
            vnod1->setX(0,x1); vnod1->setY(0,n[2]->y());
            vnod2->setX(0,x1); vnod2->setY(0,n[0]->y());

            border->path01.push_back(n[0]);
            border->path32.push_back(vnod1);
            border->path20.push_back(vnod2);
        break;
        case CORNER:
        case CENTER:
        break;
    }
    border->location = location;
    border->calcDimensions();
}
//----------------------------------------------------------------------------

void WorldModel::sectorPaths(int i,int j,Object * n[4]) {
    Sector * sector = &sectors[j][i];
    sector->location = CENTER;

    if(i > 1) {
        sector->path20 = sectors[j][i-1].path31;
        sector->path02 = sectors[j][i-1].path13;
    } else {
        sector->path20 = path2objs(find_path(Location(n[2]), // searching from
            TargetNodeFn(n[0]), // conditions for destination
            HeuristicFn(n[0]),  // heuristic to destination
            BorderTravelFn())); // cost function
        sector->path02 = reversePath(&sector->path20,n[2]);
    }

    if(j > 1) {
        sector->path01 = sectors[j-1][i].path23;
        sector->path10 = sectors[j-1][i].path32;
    } else {
        sector->path01 = path2objs(find_path(Location(n[0]), // searching from
            TargetNodeFn(n[1]), // conditions for destination
            HeuristicFn(n[1]),  // heuristic to destination
            BorderTravelFn())); // cost function
        sector->path10 = reversePath(&sector->path01,n[0]);
    }

    sector->path13 = path2objs(find_path(Location(n[1]),TargetNodeFn(n[3]),
        HeuristicFn(n[3]),BorderTravelFn()));
    sector->path31 = reversePath(&sector->path13,n[1]);
    sector->path32 = path2objs(find_path(Location(n[3]),TargetNodeFn(n[2]),
        HeuristicFn(n[2]),BorderTravelFn()));
    sector->path23 = reversePath(&sector->path32,n[3]);
}
//----------------------------------------------------------------------------

objects_t WorldModel::path2objs(Object ** path) {
    Object * O; objects_t objects;
    for(int i=0;(O = path[i]);i++) {objects.push_back(O);}
    return objects;
}
//----------------------------------------------------------------------------

objects_t WorldModel::reversePath(objects_t * path,Object * begin) {
// X -> B
// X Y -> X B
// X Y Z -> Y X B
    objects_t reversed; int pl = path->size()-1;
    while(pl > 0) {reversed.push_back((*path)[--pl]);}
    if(path->size()) {reversed.push_back(begin);}
    return reversed;
}
//----------------------------------------------------------------------------

void WorldModel::draw(ViewerWindow * aWindow) {
    // draw all sectors
    for(int j=0;j < mSectorY;j++) {
        for(int i=0;i < mSectorX;i++) {
            if(sectors[j][i].location != CORNER) {
                int col = (i+j+j)*3160016+6145334;
                pencol = col; // aWindow->SetPenColor(col);
                filcol = col; // aWindow->SetFillColor(col);
                sectors[j][i].Draw(aWindow,aWindow->mViewFlags,0x0);
//*/
/*
                drawPath(aWindow,&sectors[j][i].path01);
                drawPath(aWindow,&sectors[j][i].path13);
                drawPath(aWindow,&sectors[j][i].path32);
                drawPath(aWindow,&sectors[j][i].path20);
*/
            }
        }
    }
    
    // draw all nodes that are unreachable by the precomputed paths
    if (aWindow->mViewFlags&view_patherrors)
    {
        aWindow->SetPenColor(0x6080FF);
        const sectors_t& sectors=getSectors();
        for(sectors_t::const_iterator SI=sectors.begin();SI!=sectors.end();SI++)
        {
            const objects_t& objs=sectorObjects(*SI);
            for(objects_t::const_iterator OI1=objs.begin();OI1!=objs.end();OI1++)
            {
                if ((*OI1)->type()==TYPE_NODE)
                {
                    bool found = true;
                    Node *N1=(Node*)*OI1;
                    Node *N2;
                    for(objects_t::const_iterator OI2=objs.begin();OI2!=objs.end();OI2++)
                    {
                        if ((*OI2)->type()==TYPE_NODE)
                        {
                            N2=(Node*)*OI2;
                            if (N1!=N2)
                            {
                                const Path* P=(*SI)->GetNodePath(N1,N2);
                                if (!P)
                                {
                                    found=false;
                                    break;
                                }
                            }

                        }
                    }
                    if (!found)
                    {
                        double x,y;
                        GetObjectPosition(N1,x,y);
                        aWindow->Marker(x,y,'x',5);
                        break;
                    }
                }
            }
        }
    }
    // mark all objects not in a sector

    if (aWindow->mViewFlags&view_patherrors)
    {
        aWindow->SetPenColor(0xFF8060);
        for(Objects::const_iterator OI=objects().begin();OI!=objects().end();OI++)
        {
            const sectors_t &sec=findSectors(*OI);
            if (sec.size()==0) // || sec.size() > 4)
            {
                double x,y;
                GetObjectPosition(*OI,x,y);
                aWindow->Marker(x,y,'x',5);
            }
        }
    }
    if (aWindow->mViewFlags&view_summary)
    {
        mSummary->Draw(aWindow,this);
    }
}

//----------------------------------------------------------------------------

void drawPath(ViewerWindow * aWindow,const objects_t * path) {
    if(path == NULL) {Debug(spam_startup,"drawPath: Warning, empty path\n");}

    PointObject *P1,*P2; Road * R;
    for(unsigned int i=0;i < path->size();i++) {
        if((R=dynamic_cast<Road *>((*path)[i]))) {
            P1=(PointObject*)R->head();
            P2=(PointObject*)R->tail();
            aWindow->Line(P1->x(),P1->y(),P2->x(),P2->y(),4);
        } else if(dynamic_cast<Node *>((*path)[i]));
        else {
            Debug(spam_startup,"drawPath: Warning invalid object detected\n");
        }
    }
}
//--
//----------------------------------------------------------------------------

void Sector::Draw(ViewerWindow * aWindow,long aViewFlags,long aColor) const 
{
    // draw nexi using nexusPoints();
    if (aViewFlags&view_nexi)
    {
        aWindow->SetPenColor(aColor);
        aWindow->SetFillColor(aColor);
        objects_t nexi=nexusPoints();
        for(objects_t::iterator NI=nexi.begin();NI!=nexi.end();NI++) {
            double x,y;
            GetObjectPosition(*NI,x,y);
            aWindow->Marker(x,y,'O',5);
        }
    }
    if (aViewFlags&view_borders)
    {
        aWindow->SetPenColor(aColor);
        aWindow->SetFillColor(aColor);

/*
        Object * from; Object * to; objects_t objs;
        from = to = NULL;
        if(path10.size())
        from = path10[path10.size()-1];
        if(path01.size())
        to = path01[path01.size()-1];
        if(from && to) {
            objs = getBorder(from,to);
            drawPath(aWindow,&objs);
        }

        from = to = NULL;
        if(path31.size())
            from = path31[path31.size()-1];
        if(path13.size())
            to = path13[path13.size()-1];
        if(from && to) {
            objs = getBorder(from,to);
            drawPath(aWindow,&path13);
        }

        from = to = NULL;
        if(path23.size())
            from = path23[path23.size()-1];
        if(path32.size())
            to = path32[path32.size()-1];
        if(from && to) {
            objs = getBorder(from,to);
            drawPath(aWindow,&objs);
        }

        from = to = NULL;
        if(path02.size())
            from = path02[path02.size()-1];
        if(path20.size())
        to = path20[path20.size()-1];
        if(from && to) {
            objs = getBorder(from,to);
            drawPath(aWindow,&objs);
        }
*/
        drawPath(aWindow,&path01);
        drawPath(aWindow,&path13);
        drawPath(aWindow,&path32);
        drawPath(aWindow,&path20);
        drawPath(aWindow,&path10);
        drawPath(aWindow,&path31);
        drawPath(aWindow,&path23);
        drawPath(aWindow,&path02);
        
    }
    if (aViewFlags&view_sectorcontent)
    {
        long red,green,blue,color;
        long flag=path01.size()^path13.size()^
            path32.size()^path20.size();
        red=green=blue=0x40;
        if (flag&0x1) red+=0x60;
        if (flag&0x2) green+=0x60;
        if (flag&0x4) blue=0x60;
        if (flag&0x8) red+=0x60;
        if (flag&0x10) green+=0x60;
        if (flag&0x20) blue+=0x60;
        
        color=(red)+(green<<8)+(blue<<16);
    
        aWindow->SetPenColor(color);
        aWindow->SetFillColor(color);
        aWindow->SetFill(true);        
        // show buildings
        for(unsigned int i=0;i < objects.size();i++) {
            Building * B;
            if((B = dynamic_cast<Building *>(objects[i]))) {
                aWindow->Marker(B->x(),B->y(),'O',2);
            }
        }
    }
}
// --------------------------------------------------------------------------

void Sector::addNexus(objects_t * nexpoints,const objects_t * path) const {
    if(path->size()) {nexpoints->push_back((*path)[path->size()-1]);}
    else {/*fprintf(stderr,"addNexus: error, zero path length!\n");*/}
}
//----------------------------------------------------------------------------

objects_t WorldModel::nexusPoints(Sector * sector) {
    if(sector == NULL) {fprintf(stderr,"nexusPoints: argument error!\n");}
    return sector->nexusPoints();
}
/*
 * CP: split
 */

objects_t Sector::nexusPoints() const
{
    objects_t nexpoints;
    switch(location) {
        case NORTH:
            addNexus(&nexpoints,&path20);
            addNexus(&nexpoints,&path01);
        break;
        case SOUTH:
            addNexus(&nexpoints,&path32);
            addNexus(&nexpoints,&path13);
        break;
        case EAST:
            addNexus(&nexpoints,&path20);
            addNexus(&nexpoints,&path32);
        break;
        case WEST:
            addNexus(&nexpoints,&path01);
            addNexus(&nexpoints,&path13);
        break;
        case CENTER:
            addNexus(&nexpoints,&path20);
            addNexus(&nexpoints,&path01);
            addNexus(&nexpoints,&path32);
            addNexus(&nexpoints,&path13);
        break;
        case CORNER:
        break;
    }
    return nexpoints;
}
//----------------------------------------------------------------------------

objects_t WorldModel::getBorder(Object * from,Object * to) {
    if(from == NULL || to == NULL) {
        fprintf(stderr,"getBorder: argument error!\n");
    }

    objects_t objects;
    for(int j=0;j < mSectorY;j++) {
        for(int i=0;i < mSectorX;i++) {
            Sector * sector = &sectors[j][i];
            if(sector->location == CENTER) {
                objects_t nexpoints = nexusPoints(sector);
                if(from == nexpoints[0]) {
                    if(to == nexpoints[1]) {objects = sector->path01;}
                    else if(to == nexpoints[2]) {objects = sector->path02;}
                } else if(from == nexpoints[1]) {
                    if(to == nexpoints[0]) {objects = sector->path10;}
                    else if(to == nexpoints[3]) {objects = sector->path13;}
                } else if(from == nexpoints[2]) {
                    if(to == nexpoints[0]) {objects = sector->path20;}
                    else if(to == nexpoints[3]) {objects = sector->path23;}
                } else if(from == nexpoints[3]) {
                    if(to == nexpoints[1]) {objects = sector->path31;}
                    else if(to == nexpoints[2]) {objects = sector->path32;}
                }
            }
        }
    }
    return objects;
}
//----------------------------------------------------------------------------

objects_t WorldModel::sectorObjects(Sector * sector) {
    if(sector == NULL) {fprintf(stderr,"sectorObjects: argument error!\n");}
    return sector->objects;
}
//----------------------------------------------------------------------------

sectors_t WorldModel::findSectors(Object * object) {
    if(object == NULL) {fprintf(stderr,"findSectors: argument error!\n");}

    sectors_t result;
    for(int j=0;j < mSectorY;j++) {
        for(int i=0;i < mSectorX;i++) {
            Sector * sector = &sectors[j][i];
            if(sector->trySector(object)) {
                 if(sector->location != CORNER && sector->Contains(object)) {
                     result.push_back(sector);
                 }
            }
        }
    }
    return result;
}
//----------------------------------------------------------------------------

Humanoid * WorldModel::nearestMember(Hset * agents,S32 & x,S32 & y,
    int members) {
    Humanoid * agent = NULL;
    double min_length = MAXFLOAT;
    Hset::iterator nearest = agents->begin();
    for(Hset::iterator it = nearest;it != agents->end();it++) {
        S32 dx = (*it)->x() - x;
        S32 dy = (*it)->y() - y;
        double length = dx * dx + dy * dy;
        if(length < min_length) {min_length = length; nearest = it;}
    }

    if(nearest != agents->end()) {
        agent = *nearest;
        agents->erase(nearest);
        x = (x * members + agent->x())/(members+1);
        y = (y * members + agent->y())/(members+1);
    } else {
        Debug(spam_startup,"nearestMember: Warning, no members available\n");
    }
    return agent;
}
//----------------------------------------------------------------------------

void WorldModel::computeTeams() {
/*
    Hset::iterator it = a_agents.begin();
    for(;it != a_agents.end();it++) {
        MotionlessObject * M =
            dynamic_cast<MotionlessObject*>((*it)->position());
        ASSERT(M);
    }
*/
/*
    int a_remain = A_AMOUNT % A_TEAMS;
    int f_remain = F_AMOUNT % F_TEAMS;
    int p_remain = P_AMOUNT % P_TEAMS;
    int t_lim = A_TEAMS+F_TEAMS+P_TEAMS;
    while(t_lim--) {
        GroupInfo team; S32 x = x1; S32 y = y1;
        int a_lim = A_AMOUNT/A_TEAMS+(a_remain-- > 0);
        while(a_lim--){team.insert(nearestMember(&a_agents,x,y,team.size()));}
        int f_lim = F_AMOUNT/F_TEAMS+(f_remain-- > 0);
        while(f_lim--){team.insert(nearestMember(&f_agents,x,y,team.size()));}
        int p_lim = P_AMOUNT/P_TEAMS+(p_remain-- > 0);
        while(p_lim--){team.insert(nearestMember(&p_agents,x,y,team.size()));}
        teams.push_back(team);
        Debug(spam_group_information,"Team %d\n",t_lim);
        team.print();
    }
*/
}
//----------------------------------------------------------------------------

/*
 * Pauper's domain: Code below should not be touched by anyone under
 * 25 years old!! ;-)
 */

bool Sector::Contains(const Object* aObject) const
{
    for(objects_t::const_iterator I=objects.begin();I!=objects.end();I++)
    {
        if ((*I)==aObject)
            return true;
    }
    return false;
}

Object* Sector::GetObjectById(S32 aId) const
{
    for(objects_t::const_iterator I=objects.begin();I!=objects.end();I++)
    {
        if ((*I)->id()==aId)
        {
            return *I;
        }
    }
    return 0;
}

sectors_t WorldModel::getSectors()
{
    sectors_t Out;
    int j,i;
    for(j=0;j < mSectorY;j++) {
        for(i=0;i < mSectorX;i++) {
            Out.push_back(&sectors[j][i]);
        }
    }

    return Out;
}

bool WorldModel::insideSector(Sector* aSector,Object* aObject)
{
    assert(aSector);
    return aSector->trySector(aObject);
}

bool WorldModel::IsNexus(Object* object)
{
    for(map<S32,int>::iterator NI=mNexusIndex.begin();NI!=mNexusIndex.end();NI++)
    {
        if (NI->first==object->id())
        {
            return true;
        }
    }
    return false;
}
 

#include "SituationModel.h"
#include "Team.h"
#include "Task.h"
#include "ImaginaryObject.h"

class ClosestRefugeFn : public Rescue::Memory::ObjectFn
{
protected:
    Sector* mSector;
public:
    virtual bool consider(const Object *o) const
    {
        return (dynamic_cast<const Refuge*>(o)!=NULL);
    }
    
    virtual double f(const Object *o) const
    {
        if (mSector)
        {
            long x,y;
            GetObjectPosition(o,x,y);
            x-=mSector->MiddleX();y-=mSector->MiddleY();
        
            return (x*x)+(y*y);
        }
        else
        {
            return 0.0;
        }
    }
    
    ClosestRefugeFn(Sector* aSector)
    {
        mSector=aSector;
    }
};
    void WorldModel::AddPaths(SituationModel* aModel,Object* aDestination)
    {
        const sectors_t& secs=findSectors(aDestination);
        for(sectors_t::const_iterator SI=secs.begin();SI!=secs.end();SI++)
        {
            const objects_t& nexi=(*SI)->nexusPoints();
            for(objects_t::const_iterator NI=nexi.begin();NI!=nexi.end();NI++)
            {
                // paths from all nexi of all sectors destination is in 
                // to the closest node to destination
                // we don't care it's not to destination since destination itself 
                // is added anyway.
                const Path* NP=(*SI)->GetInternalPath(*NI,aDestination);
                const Objects& po=NP->GetObjects();
                for(Objects::const_iterator PI=po.begin();PI!=po.end();PI++)
                {
                    if (!aModel->GetObject(*PI))
                    {
                        aModel->Add(*PI);   
                    }                
                }
            }
        }
    }
SituationModel* WorldModel::ExtractCurrentSituation(BasicAgent * aAgent)
{
    SituationModel* O=new SituationModel;
    // todo, choose sector from task
    SummaryTeam* team=mSummary->MakeTeam(aAgent);
    O->mTask=mSummary->MakeTask(aAgent,team);

    O->mSector=0;
    const sectors_t& secs=findSectors(aAgent->Position());
    for(sectors_t::const_iterator SI=secs.begin();SI!=secs.end();SI++)
    {
        if (*SI==O->mTask->TargetSector())
        {
            O->mSector=*SI;
            break;
        }
    }
    if (!O->mSector)
    {
            // sector can be 0, in that case the situation model just contains the highways
            // and the paths to the team members and the refuge
    }
        
    O->mOriginalObjects=new ImaginaryObjects;
    
    O->mWorldModel=this;
    O->mTime=aAgent->Time();
/** /
    // To test the effect of sectors as focus we add all objects
    for(Objects::const_iterator I=objects().begin();I!=objects().end();I++)
    {
        if ((dynamic_cast<Building*>(*I)) || (dynamic_cast<Road*>(*I)) || (dynamic_cast<Node*>(*I)) || (dynamic_cast<Humanoid*>(*I)) ||(dynamic_cast<Civilian*>(*I)))
        {
            O->Add(*I);
        }
        if (dynamic_cast<Refuge*>(*I))
        {
            O->mClosestRefuge=(Refuge*)*I;
        }
    }   
    O->SetTeam(team);
    return O;
/ **/
    
    if (O->mSector)
    {
        const vector<Object*>& secobj=sectorObjects(O->mSector);
        
        // add all sector objects
        for(vector<Object*>::const_iterator I=secobj.begin();I!=secobj.end();I++)
        {
            O->Add(*I);
        }
        O->Add(O->mClosestRefuge=(Rescue::Refuge*)find(ClosestRefugeFn(O->mSector),FUNCTION_MINIMIZE));
        AddPaths(O,O->mClosestRefuge);
        
        // TODO paths from nexi of sector of refuge to refuge
    }
    else
    {
        O->Add(O->mClosestRefuge=(Rescue::Refuge*)find(ClosestRefugeFn(0),FUNCTION_MINIMIZE));
        AddPaths(O,O->mClosestRefuge);
    }
    // add all highway objects
    for(objects_t::const_iterator I=mHighwayObjects.begin();I!=mHighwayObjects.end();I++)
    {
        if ((O->mSector==NULL) || (!O->mSector->Contains(*I)))
        {
            O->Add(*I);
        }
    }
    /// add all known civilians
    for(Objects::const_iterator OI=objects().begin();OI!=objects().end();OI++)
    {
        Civilian *C=dynamic_cast<Civilian*>(*OI);
        if (C)
        {
            if (C->position())
            {
                O->Add(C);
                AddPaths(O,C);
            }
        }
    }
    for(vector<Object*>::iterator TI=team->begin();TI!=team->end();TI++)
    {
        ImaginaryHuman* IH=(ImaginaryHuman*)O->Add(*TI);
        assert(IH!=0);
        ImaginaryObject* IO=IH->mPosition;
        if (IO)
        {
            AddPaths(O,IO->ActualObject());
        }
    }
        
    O->SetTeam(team);
    
/*    
    switch(aAgent->Type())
    {
        case AGENTTYPE_FIRE_COMPANY:
            O->mTask=new FightFireTask(S);
        break;
        case AGENTTYPE_AMBULANCE_TEAM:
            O->mTask=new SearchAndRescueTask(S);
        break;
        case AGENTTYPE_POLICE_FORCE:
            O->mTask=new ImproveTravelTask(S);
        break;
    }
*/
    return O;
}



    /*
     * Internal paths
     */
     
    const Path* Sector::GetNodePath(const Object* aFrom,const Object* aTo) const
    {
        assert(aFrom->type()==TYPE_NODE);
        assert(aTo->type()==TYPE_NODE);
        assert(Contains(aFrom));
        assert(Contains(aTo));

        int from=mNodeIndex.find(aFrom->id())->second;
        int to=mNodeIndex.find(aTo->id())->second;
        return GetNodePath(from,to);
    }

    const Path* Sector::GetNodePath(int aFrom,int aTo) const
    {
        if ((aFrom>0) && (aFrom<mNodeQ) && (aTo>0) && (aTo<mNodeQ))
        {
            return mInternalPath[aFrom+(mNodeQ*aTo)];
        }
        else
        {
            return 0;
        }
    }

    void Sector::SetNodePath(int aFrom,int aTo,Path* aPath)
    {
        if ((aFrom>0) && (aFrom<mNodeQ) && (aTo>0) && (aTo<mNodeQ))
        {
            Path* P=mInternalPath[aFrom+(mNodeQ*aTo)];
            if (P)
            {
                delete P;
            }
            mInternalPath[aFrom+(mNodeQ*aTo)]=aPath;
        }
    }

    void Sector::SetNodePath(const Object* aFrom,const Object* aTo,Path* aPath)
    {
        assert(aFrom->type()==TYPE_NODE);
        assert(aTo->type()==TYPE_NODE);
        assert(Contains(aFrom));
        assert(Contains(aTo));

        int from=mNodeIndex[aFrom->id()];
        int to=mNodeIndex[aTo->id()];
        SetNodePath(from,to,aPath);
    }




    
    void Sector::OldDoCalcPaths(Node* aStart)
    {
/* */
        // make base path for start
        Path* startpath=new Path();
        (*startpath)+=aStart;
        SetNodePath(aStart,aStart,startpath);
        list<Object*> available;
        available.push_back(aStart);
        map<S32,float> distances;

        Node* best=0;
        do
        {
            best=0;
            const Path *rest,*bestRest=0;
            Node *succ=0;
            Road *road,*bestRoad=0;
            float score,bestScore=MAXFLOAT;

            // N=forall Nodes
            list<Object*>::iterator NI;
            for(list<Object*>::iterator I=available.begin();I!=available.end();I=NI)
            {
                // get successor in advance because of internal erasing
                NI=I;
                NI++;

                bool hit=false;
                Node* node=dynamic_cast<Node*>(*I);
                if ((node))
                {
                    // if there is a path to this node
                    rest=GetNodePath(aStart,node);
                    if (rest)
                    {
                        float restdistance=rest->Distance();
                        // for all roads in the neighbours of N
                        for(Objects::const_iterator NI=node->neighbors().begin();NI!=node->neighbors().end();NI++)
                        {
                            if ((road=dynamic_cast<Road*>(*NI)))
                            {
                                // successor is other side of road
                                if (road->head()==node) succ=(Node*)road->tail();
                                else succ=(Node*)road->head();
                                // if part of this sector
                                if (mNodeIndex[succ->id()])
                                {
                                // if no path from start to this possibility yet
                                   const Path* P=GetNodePath(aStart,succ);
                                   if (!P)
                                   {
                                       score=restdistance+road->length()/1000.0;
                                       hit=true;
                                       if ((!best) || (score<bestScore))
                                       {
                                           bestScore=score;
                                           best=succ;
                                           bestRoad=road;
                                           bestRest=rest;
                                       }

                                    }
                                }
                            }
                        }
                    }
                }
                if (!hit)
                { // no more possibilities. remove this element
                    available.erase(I);
                }
            }
            // store best path
            if (best)
            {
                Path* P=new Path();
                (*P)+=*bestRest;
                (*P)+=bestRoad;
                (*P)+=best;
                SetNodePath(aStart,best,P);
                available.push_back(best);
                available.unique();
            }
            else
            {
                // no expandable nodes, done for start Node
                break;
            }
        } while(best);

    }



void Sector::PrecomputeInternalPaths()
{
    // make node indices
    mNodeQ=1;
    for(objects_t::iterator I=objects.begin();I!=objects.end();I++)
    {
        if ((*I)->type()==TYPE_NODE)
        {
            mNodeIndex[(*I)->id()]=mNodeQ++;
        }
    }

    mInternalPath.resize(mNodeQ*mNodeQ,0);
/*
    for(objects_t::iterator I=objects.begin();I!=objects.end();I++)
    {
        if ((*I)->type()==TYPE_NODE)
        {
            Node* Start=(Node*)*I;
            OldDoCalcPaths(Start);

        }
    }

*/
    for(objects_t::iterator I=objects.begin();I!=objects.end();I++)
    {
        if ((*I)->type()==TYPE_NODE)
        {
            Node* Start=(Node*)*I;
            int startindex=mNodeIndex[Start->id()];
            map<int,float> distances;

            // make base path for start
            Path* startpath=new Path();
            (*startpath)+=Start;
            SetNodePath(startindex,startindex,startpath);
            distances[startindex]=0;

            list<Object*> available;
            available.push_back(Start);

            Node* best=0;
            do
            {
                best=0;
                const Path *rest,*bestRest=0;
                Node *succ=0;
                Road *road,*bestRoad=0;
                float score,bestScore=MAXFLOAT;
                int bestindex;
                // N=forall Nodes
                list<Object*>::iterator NI;
                for(list<Object*>::iterator I=available.begin();I!=available.end();I=NI)
                {
                    // get successor in advance because of internal erasing
                    NI=I;
                    NI++;

                    bool hit=false;
                    Node* node=dynamic_cast<Node*>(*I);
                    if ((node))
                    {
                        // there should be a path to here as it is in available
                        int restindex=mNodeIndex[node->id()];
                        rest=GetNodePath(startindex,restindex);
                        assert(rest!=NULL);
                        float restdistance=distances[restindex];
//                        float restdistance=distances[node->id()];
                        if (restdistance<=bestScore)
                        {
                            // for all roads in the neighbours of N
                            for(Objects::const_iterator NI=node->neighbors().begin();NI!=node->neighbors().end();NI++)
                            {
                                if ((road=dynamic_cast<Road*>(*NI)))
                                {
                                    // successor is other side of road
                                    if (road->head()==node) succ=(Node*)road->tail();
                                    else succ=(Node*)road->head();
                                    // if part of this sector
                                    int succindex=mNodeIndex[succ->id()];
                                    if (succindex)
                                    {
                                    // if no path from start to this possibility yet
                                       const Path* P=GetNodePath(startindex,succindex);
                                       if (!P)
                                       {
                                           score=restdistance+road->length()/1000.0;
                                           hit=true;
                                           if ((!best) || (score<bestScore))
                                           {
                                               bestindex=succindex;
                                               bestScore=score;
                                               best=succ;
                                               bestRoad=road;
                                               bestRest=rest;
                                           }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    if (!hit)
                    { // no more possibilities. remove this element
                        NI=available.erase(I);
                    }
                }
                // store best path
                if (best)
                {
                    Path* P=new Path();
                    (*P)+=*bestRest;
                    (*P)+=bestRoad;
                    (*P)+=best;
                    distances[bestindex]=bestScore;
                    SetNodePath(startindex,bestindex,P);
                    available.push_back(best);
                    available.unique();
//                      distances[mNodeIndex[best->id()]]=bestScore;
//                    distances[best->id()]=bestScore;
                }
                else
                {
                    // no expandable nodes, done for start Node
                    break;
                }
            } while(best);

        }
    }

}
    void Sector::DuplicatePaths(Sector* aOriginal,ObjectPool* aObjectPool)
    {
        mNodeQ=aOriginal->mNodeQ;
        mNodeIndex=aOriginal->mNodeIndex;
        mInternalPath.resize(mNodeQ*mNodeQ,0);
        for(int from=0;from<mNodeQ;from++)
        {
            for(int to=0;to<mNodeQ;to++)
            {
                const Path* P=aOriginal->GetNodePath(from,to);
                if (P)
                {
                    Path* NewP=new Path(*P);
                    NewP->ConvertObjectPool(aObjectPool);
                    SetNodePath(from,to,NewP);
                }
                else
                {
                    SetNodePath(from,to,0);
                }
            }
        }
    }
void WorldModel::PrecomputeInternalPaths()
{
    if (gActors.begin()!=gActors.end())
    {
        /*
         * Duplicate internal paths for all sectors
         */
        Debug(spam_startup,"Converting internal paths\n");
        BasicActor* OA=*gActors.begin();
        WorldModel* OM=OA->model();
        for(int j=0;j < mSectorY;j++)
        {
            for(int i=0;i < mSectorX;i++)
            {
                Sector* sectorfrom=&(OM->sectors[j][i]);
                sectors[j][i].DuplicatePaths(sectorfrom,this);
            }
        }
    }
    else
    {
        mInternalPaths=0;
        /*
         * precompute internal paths for all sectors
         */
        unsigned long totalfound=0,totalsize=0,totalmax=0;
        float totaldistance=0.0;
        for(int j=0;j < mSectorY;j++) {
            for(int i=0;i < mSectorX;i++) {
                sectors[j][i].PrecomputeInternalPaths();
                int found=0;
                unsigned int size=0;
                float distance=0.0;
                const vector<Path*>& paths=sectors[j][i].InternalPaths();
                int nodeq=sectors[j][i].NodeQ();

                for(vector<Path*>::const_iterator PI=paths.begin();PI!=paths.end();PI++)
                {
                    if (*PI)
                    {
                        assert((*PI)->Consistency());
                        found++;
                        mInternalPaths++;
                        size+=sizeof(*PI);
                        distance+=(*PI)->Distance();
                    }
                }
                int maxpaths=(nodeq-1)*(nodeq-1);
                float percentage;
                if (maxpaths>0) percentage=found/(float)maxpaths;
                else percentage=1.0;
                
                Debug(spam_startup,"Sector(%d,%d) - Paths %d of %d found (%3.2f\%), size %d bytes, distance %.0f meters\n",i,j,found,maxpaths,percentage*100.0f,size,distance);
                totalfound+=found;
                totalmax+=maxpaths;
                totalsize+=size;
                totaldistance+=distance;
            }
        }
        Debug(spam_startup,"Done precomputing sector paths: total found %ld of %ld (%3.2f\%)\n size %ld bytes, distance %.0f meters\n",totalfound,totalmax,(totalfound*100.0)/totalmax,totalsize,totaldistance);
        mPathBytes=totalsize;
    }
}

        void WorldModel::SetHighway(Object* aFrom,Object* aTo,Path* aPath)
        {
            int from,to;
            from=mNexusIndex[aFrom->id()];
            to=mNexusIndex[aTo->id()];
            SetHighway(from,to,aPath);
        }

        void WorldModel::SetHighway(int aFrom,int aTo,Path* aPath)
        {
            if ((aFrom>0) && (aFrom<mNexusQ) && (aTo>0) && (aTo<mNexusQ))
            {
                Path* P=mHighway[aFrom+(aTo*mNexusQ)];
                if (P)
                {
                    delete P;
                }
                mHighway[aFrom+(aTo*mNexusQ)]=aPath;
            }
        }

        const Path* WorldModel::GetHighway(Object* aFrom,Object* aTo)
        {
            int from,to;
            from=mNexusIndex[aFrom->id()];
            to=mNexusIndex[aTo->id()];
            return GetHighway(from,to);
        }

        const Path* WorldModel::GetHighway(int aFrom,int aTo)
        {
            
            if ((aFrom>0) && (aFrom<mNexusQ) && (aTo>0) && (aTo<mNexusQ))
            {
                return mHighway[aFrom+(aTo*mNexusQ)];
            }
            return 0;
        }        

    void WorldModel::DoPrecomputeHighways(Object* aStart)
    {
        list<Object*> available;
        available.push_back(aStart);
        Path* startpath=new Path();
        (*startpath)+=aStart;
        SetHighway(aStart,aStart,startpath);

        Object *best=0;
        do
        {
            best=0;
            Object *succ=0;
            const Path *rest=0,*bestRest=0;
            const Path *street=0,*bestStreet=0;
            float score=0,bestScore=MAXFLOAT;
            list<Object*>::iterator CI,NCI;
            for(CI=available.begin();CI!=available.end();CI=NCI)
            {
                NCI=CI;NCI++;
                bool hit=false;
                rest=GetHighway(aStart,*CI);
                if (rest)
                {
                    // get all neighbours of the current and the path pieces to there
                    const sectors_t& secs=findSectors(*CI);
                    objects_t neighbours;
                    map<S32,const Path*> streets;

                    for(sectors_t::const_iterator SI=secs.begin();SI!=secs.end();SI++)
                    {
                        const objects_t& nexi=(*SI)->nexusPoints();
                        for(objects_t::const_iterator NI=nexi.begin();NI!=nexi.end();NI++)
                        {
                            const Path* current=GetHighway(aStart,*NI);
                            if (!current)
                            {
                                const Path* P=(*SI)->GetNodePath(*CI,*NI);
                                assert(P!=0);
                                streets[(*NI)->id()]=P;
                                neighbours.push_back(*NI);
                                hit=true;
                            }
                        }
                    }
                    if (hit)
                    {
                         // check all neighbours for the best total score
                         float restdistance=rest->Distance();
                         for(objects_t::const_iterator NI=neighbours.begin();NI!=neighbours.end();NI++)
                         {
                             succ=*NI;
                             street=streets[succ->id()];
                             score=restdistance+street->Distance();
                             if (score<bestScore)
                             {
                                 bestRest=rest;
                                 best=succ;
                                 bestStreet=street;
                                 bestScore=score;
                             }
                         }
                    }
                    else
                    {
//                        available.erase(CI);
                    }
                }
            }
            if (best)
            {
                Path* P=new Path();
                (*P)+=*bestRest;
                (*P)+=*bestStreet;
                P->Optimize();
                SetHighway(aStart,best,P);
                available.push_back(best);
                available.unique();
            }
        } while(best);
    }
 
        
        void WorldModel::DoPrecomputeHighwaysNew(Object* Start)
        {
            
            // initialize helper structures
            vector<float> distances;
            distances.resize(mNexusQ,MAXFLOAT);
            list<Object*> available;
            available.push_back(Start);

            // make starting path
            Path* startpath=new Path();
            (*startpath)+=Start;
            SetHighway(Start,Start,startpath);
            distances[mNexusIndex[Start->id()]]=0.0;

            Object *best=0;
            do
            {
                best=0;
                Object *succ=0;
                const Path *rest=0,*bestRest=0;
                const Path *street=0,*bestStreet=0;
                float score=0,bestScore=MAXFLOAT;
                
                list<Object*>::iterator CI,NCI;
                for(CI=available.begin();CI!=available.end();CI=NCI)
                {
                    NCI=CI;NCI++;
                    rest=GetHighway(Start,*CI);
                    if (rest)
                    {
                        float restdistance=distances[mNexusIndex[(*CI)->id()]];
                        if (restdistance<bestScore)
                        {
                            bool hit=false;
                            // get all neighbours of the current and the path pieces to there
                            const sectors_t& secs=findSectors(*CI);
                            objects_t neighbours;
                            map<S32,const Path*> streets;

                            for(sectors_t::const_iterator SI=secs.begin();SI!=secs.end();SI++)
                            {
                                const objects_t& nexi=(*SI)->nexusPoints();
                                for(objects_t::const_iterator NI=nexi.begin();NI!=nexi.end();NI++)
                                {
                                    const Path* current=GetHighway(Start,*NI);
                                    if (!current)
                                    {
                                        const Path* P=(*SI)->GetNodePath(*CI,*NI);
                                        if (P)
                                        {
                                        assert(P!=0);
                                        streets[(*NI)->id()]=P;
                                        neighbours.push_back(*NI);
                                        hit=true;
                                        }
                                    }
                                }
                            }
                            if (hit)
                            {
                                 // check all neighbours for the best total score
                                 for(objects_t::const_iterator NI=neighbours.begin();NI!=neighbours.end();NI++)
                                 {
                                     succ=*NI;
                                     street=streets[succ->id()];
                                     score=restdistance+street->Distance();
                                     if (score<bestScore)
                                     {
                                         bestRest=rest;
                                         best=succ;
                                         bestStreet=street;
                                         bestScore=score;
                                     }
                                 }
                            }
                            else
                            {
                                NCI=available.erase(CI);
                            }
                        }
                    }
                }
                if (best)
                {
                    Path* P=new Path();
                    (*P)+=*bestRest;
                    (*P)+=*bestStreet;
//                    P->Optimize();
                    distances[mNexusIndex[best->id()]]=bestScore;
                    SetHighway(Start,best,P);
                    available.push_back(best);
                    available.unique();
                }
            } while(best);
        }                        
void WorldModel::PrecomputeHighways()
{
    if (gActors.begin()!=gActors.end())
    {   // copy from first agent
        BasicActor* OA=*gActors.begin();
        WorldModel* OM=OA->model();
        mNexusQ=OM->mNexusQ;
        mHighway.resize(mNexusQ*mNexusQ);
        Debug(spam_startup,"Converting %d highways\n",mNexusQ*mNexusQ);

        mNexusIndex=OM->mNexusIndex;

        for(int from=0;from<mNexusQ;from++)
        {
            for(int to=0;to<mNexusQ;to++)
            {
                const Path* HW=OM->GetHighway(from,to);
                if (HW)
                {
                    Path* NewHW=new Path(*HW);
                    NewHW->ConvertObjectPool(this);
                    SetHighway(from,to,NewHW);
                }
                else
                {
                    SetHighway(from,to,0);
                }
            }        
        } 
    }
    else
    {   // make own highways
        
        const sectors_t &secs=getSectors();
        int index=1;
        objects_t allnexi;

        // add all nexi to the nexus index
        for(sectors_t::const_iterator SI=secs.begin();SI!=secs.end();SI++)
        {
            const objects_t &nexi=(*SI)->nexusPoints();
            for(objects_t::const_iterator NI=nexi.begin();NI!=nexi.end();NI++)
            {
                if (mNexusIndex.find((*NI)->id())==mNexusIndex.end())
                { // still unique?
                    mNexusIndex[(*NI)->id()]=index++;
                    allnexi.push_back(*NI);
                }
            }
        }
        mHighway.resize(index*index,0);
        mNexusQ=index;
        
        // make highways with each found nexus as starting point
        Debug(spam_startup,"Making highways between %d nexi\n",allnexi.size());
        mHighwayBytes=0;
        mHighways=0;
        for(objects_t::const_iterator NI=allnexi.begin();NI!=allnexi.end();NI++)
        {
            Object* Start=*NI;
            DoPrecomputeHighwaysNew(Start);

            for(objects_t::const_iterator NI2=allnexi.begin();NI2!=allnexi.end();NI2++)
            {
                const Path* HW=GetHighway(*NI,*NI2);
                if (!HW)
                {
                    Debug(spam_startup,"No highway from N%03ld to N%03ld\n",(*NI)->id(),(*NI2)->id());
                }
                else
                {
                    assert(HW->Consistency());
                    mHighwayBytes+=sizeof(*HW);
                    mHighways++;
                }
            }
        }
    }
    // make a container with all objects involved in highways
    for(vector<Path*>::iterator HWI=mHighway.begin();HWI!=mHighway.end();HWI++)
    {
        if (*HWI)
        {
            const Objects& hwo=(*HWI)->GetObjects();
            for(Objects::const_iterator HWOI=hwo.begin();HWOI!=hwo.end();HWOI++)
            {
                bool found=false;
                for(objects_t::iterator HWOI2=mHighwayObjects.begin();HWOI2!=mHighwayObjects.end();HWOI2++)
                {
                    if ((*HWOI2)==(*HWOI))
                    {
                        found=true;
                        break;
                    }
                }
                if (!found)
                {
                    mHighwayObjects.push_back(*HWOI);
                }
            }
        }
    }
        
    Debug(spam_startup,"Finished creating highways....\n");

}

    Node* ClosestNodeTo(Object* aObject,Object*& extraitem,const Memory::ObjectFn& heuristic)
    {
        Node* N=dynamic_cast<Node*>(aObject);
        if (N)
        {
            extraitem=0;
            return N;
        }
                            
        MovingObject* M=dynamic_cast<MovingObject*>(aObject);
        if (M)
        {
            Object* pos=M->position();
            Road* R=dynamic_cast<Road*>(pos);
            if (R)
            {
                extraitem=R;
                if (M->positionExtra()<R->length()/2)
                {
                    return (Node*)R->head();
                }
                else
                {
                    return (Node*)R->tail();
                }
            }
            else
            {
                return ClosestNodeTo(pos,extraitem,heuristic);
            }
        }
        Building* B=dynamic_cast<Building*>(aObject);
        if (B)
        {
            extraitem=B;
            Object* Best=0;
            float score,bestscore=-MAXFLOAT;
            for(Objects::const_iterator EI=B->entrances().begin();EI!=B->entrances().end();EI++)
            {
                if (heuristic.consider(*EI))
                {
                    score=heuristic.f(*EI);
                    if (score>bestscore)
                    {
                        Best=*EI;
                        bestscore=score;
                    }
                }
            }
            return (Node*)Best;
        }
        Road* R=dynamic_cast<Road*>(aObject);
        if (R)
        {
            extraitem=R;
            if (!heuristic.consider(R->tail()))
            {
                if (heuristic.consider(R->head()))
                {
                    return (Node*)R->head();
                }
                else
                {
                    return 0;
                }
            }
            if (!heuristic.consider(R->head()))
            {
                if (heuristic.consider(R->tail()))
                {
                    return (Node*)R->tail();
                }
                else
                {
                    return 0;
                }
            }
            if (heuristic.f(R->head())>heuristic.f(R->tail()))
            {
                return (Node*)R->head();
            }
            else
            {
                return (Node*)R->tail();
            }
        }
        return 0; // unrecognized type
    }

    class InvDistanceFn : public Memory::DistanceObjectFn
    {
    public:
        virtual double f(const Object *o) const
        {
            return -Memory::DistanceObjectFn::f(o);
        }
        InvDistanceFn(const Object *o) : DistanceObjectFn(o)
        {
        } 
   };


    const Path* Sector::GetInternalPath(Object* aStart,Object* aDest) const
    {
        Object *preitem=0,*postitem=0;
        Node *from,*to;
        from=ClosestNodeTo(aStart,preitem,InvDistanceFn(aDest));
        to=ClosestNodeTo(aDest,postitem,InvDistanceFn(aStart));
        return GetNodePath(from,to);
    }

Path* WorldModel::FindFastPath(Object* aStart,Object* aDest)
{
    Path* out=0;
    if (aStart==aDest)
    {
        out=new Path;
        (*out)+=aStart;
        return out;
    }
            
    // find closest node to aStart and start off with how to get there in path
    Node *from,*to;
    Object *preitem=0,*postitem=0;


    
    from=ClosestNodeTo(aStart,preitem,InvDistanceFn(aDest));
    to=ClosestNodeTo(aDest,postitem,InvDistanceFn(aStart));

    const sectors_t& secfrom=findSectors(from);
    const sectors_t& secto=findSectors(to);

    for(sectors_t::const_iterator S1=secfrom.begin();S1!=secfrom.end();S1++)
    {
        for(sectors_t::const_iterator S2=secto.begin();S2!=secto.end();S2++)
        {
            if (*S1==*S2)
            {
                const Path* middle=(*S1)->GetNodePath(from,to);
                if (middle)
                {
                    out=new Path;
                    (*out)+=preitem;
                    (*out)+=*middle;
                    (*out)+=postitem;
                    assert(out->Consistency());
                    return out;
                }
            }
        }
    }
    Object *pre_nexus=0,*post_nexus=0;
    Sector *pre_sector=0,*post_sector=0;

    float score,best;
    best=MAXFLOAT;
    for(sectors_t::const_iterator S1=secfrom.begin();S1!=secfrom.end();S1++)
    {
        const objects_t& nexi=(*S1)->nexusPoints();
        for(objects_t::const_iterator N=nexi.begin();N!=nexi.end();N++)
        {
            score=DistanceBetween(from,*N)+DistanceBetween(*N,to);
            if (score<best)
            {
                best=score;
                pre_nexus=*N;
                pre_sector=*S1;
            }
        }
    }
    best=MAXFLOAT;
    for(sectors_t::const_iterator S2=secto.begin();S2!=secto.end();S2++)
    {
        const objects_t& nexi=(*S2)->nexusPoints();
        for(objects_t::const_iterator N=nexi.begin();N!=nexi.end();N++)
        {
            score=DistanceBetween(from,*N)+DistanceBetween(*N,to);
            if (score<best)
            {
                best=score;
                post_nexus=*N;
                post_sector=*S2;
            }
        }
    }

    if ((pre_nexus) && (post_nexus))
    {
        out=new Path;
        const Path* prepath=pre_sector->GetNodePath(from,pre_nexus);
        const Path* postpath=post_sector->GetNodePath(post_nexus,to);
        const Path* highway=GetHighway(pre_nexus,post_nexus);
        if ((prepath) && (highway) && (postpath))
        {
            (*out)+=preitem;
            (*out)+=*prepath;
            (*out)+=*highway;
            (*out)+=*postpath;
            (*out)+=postitem;
            out->Optimize();
            assert(out->Consistency());
            return out;
        }
    }
    return 0;
}

Path* WorldModel::FindPath(Object* aStart,Object* aDest,const Memory::Memory::ObjectPairFn& travelfn,bool aFast)
{
    // to test results of precomputed paths, insert slow path here:
//    return Path::Create(aStart,aDest,this,travelfn);
    
    Path* P;
    P=FindFastPath(aStart,aDest);
    if (P)
    {
        if (!P->Check(travelfn))
        {
            delete P;
            P=0;
        }
    }
    if ((!P) && (!aFast))
    {
        P= Path::Create(aStart,aDest,this,travelfn);
    }
    assert((!P) || (P->Consistency()));
    return P;

}

Path* WorldModel::FindPath(Object* aStart,const Memory::ObjectFn& aTargetFn,const Memory::Memory::ObjectPairFn& travelfn,bool aFast)
{
    Object** target=find_list(aTargetFn);
    Path* P;
    if (target)
    {
        // use fast path
        int t;
        for(t=0;target[t];t++)
        {
/** /
//          to test results of precomputed paths, insert slow path here:
            P=Path::Create(aStart,target[t],this,travelfn);
            if (P)
            {
                return P;
            }
/ **/
            P=FindFastPath(aStart,target[t]);
            if (P)
            {
                if (P->Check(travelfn))
                {
                    assert(P->Consistency());
                    return P;
                }
                else
                {
                    delete P;
                }
            }
/**/
        }
        if (!aFast)
        {
            // use A*
            for(t=0;target[t];t++)
            {
                P=Path::Create(aStart,target[t],this,travelfn);
                if (P)
                {
                    assert(P->Consistency());
                    return P;
                }
            }
        }
    }
    return 0;
}


bool UnitTest()
{
    return true;
}

/*
const Path* Sector::GetInternalPath(Object* aFrom,Object* aTo)
{
    if ((aFrom->type()==TYPE_NODE) && (aTo->type()==TYPE_NODE))
    {
        return GetNodePath(aFrom,aTo);
    }

    Building *B;
    MovingObject *M;
    Road* R;

    // get closest node to aFrom
    if ((M=dynamic_cast<MovingObject*>(aFrom)))
    {
        return GetInternalPath(M->position(),aTo);
    }
    else if ((B=dynamic_cast<Building*>(aFrom)))
    {
        float score,bestscore=MAXFLOAT;
        const Path* best=0,*test;
        for(Objects::const_iterator E=B->entrances().begin();E!=B->entrances().end();E++)
        {
            test=GetInternalPath(*E,aTo);

            if (test)
            {
                score=test->Distance();
                if (score<bestscore);
                {
                    bestscore=score;
                    best=test;
                }
            }
        }
        return test;
    }
    else if ((R=dynamic_cast<Road*>(aFrom)))
    {
        const Path *hpath,*tpath;
        hpath=GetInternalPath(R->head(),aTo);
        tpath=GetInternalPath(R->tail(),aTo);
        // todo, should add road to front
        if ((hpath) && (tpath) && (hpath->Distance()<tpath->Distance()))
        {
            return hpath;
        }
        else
        {
            return tpath;
        }
    }

    // get closest node to aTo
    if ((M=dynamic_cast<MovingObject*>(aTo)))
    {
        return GetInternalPath(aFrom,M->position());
    }

    else if ((B=dynamic_cast<Building*>(aTo)))
    {
        float score,bestscore=MAXFLOAT;
        const Path* best=0,*test;
        for(Objects::const_iterator E=B->entrances().begin();E!=B->entrances().end();E++)
        {
            test=GetInternalPath(aFrom,*E);

            if (test)
            {
                score=test->Distance();
                if (score<bestscore);
                {
                    bestscore=score;
                    best=test;
                }
            }
        }
        return test;
    }

    else if ((R=dynamic_cast<Road*>(aTo)))
    {
        const Path *hpath,*tpath;
        hpath=GetInternalPath(aFrom,R->head());
        tpath=GetInternalPath(aFrom,R->tail());
        // todo, should add road to front
        if ((hpath) && (tpath) && (hpath->Distance()<tpath->Distance()))
        {
            return hpath;
        }
        else
        {
            return tpath;
        }
    }
    return 0;
}
*/


