// $Id: WorldModel.h,v 1.1.1.1 2003/07/04 17:15:03 darseq Exp $
// copyright, s.post & cp
#ifndef _WorldModel_h
#define _WorldModel_h
#include <adk.hxx>
#include <vector>
#include <map>

#include "GroupInfo.h"
#include "Team.h"
#include "Task.h"

using namespace Rescue;
// Location, ObjectFn,
// path_cost, path_destination, path_subpath

#define MAXSECTORS 10 // <- commandline options checked against this?
//#define SECTORX 6 // including borders (>= 4)
//#define SECTORY 5 // including borders (>= 4)
#define CORRECTION 0.1 // correction for outer points on the map
#define MAXPOINTS 500 // max amount of points of sector polygon
#define A_AMOUNT 5  // howmany ambulances there are in the map
#define F_AMOUNT 10 // howmany ambulances there are in the map
#define P_AMOUNT 10 // howmany ambulances there are in the map
#define A_TEAMS 1 // amount of ambulance teams
#define F_TEAMS 2 // amount of firebrigade teams
#define P_TEAMS 1 // amount of police teams
#define P_PER_T 2 // howmuch police members A & F teams get.

typedef vector<Object *> objects_t;
typedef enum {NORTH,SOUTH,EAST,WEST,CENTER,CORNER} location_t;

class Sector {
protected:
    void addNexus(objects_t * nexpoints,const objects_t * path) const;

    friend class WorldModel;
    friend class SectorForm;
    // CPs path sjit
    typedef map<S32,int> tNodeIndexMap;
    tNodeIndexMap mNodeIndex; // maps nodes to indexes
    int mNodeQ;
    // [nodes*nodes] array with precomputed paths from nodeindex to nodeindex
    vector<class Path*> mInternalPath;
    const class Path* GetNodePath(const Object* aFrom,const Object* aTo) const;
    const class Path* GetNodePath(int aFrom,int aTo) const;
    void SetNodePath(const Object* aFrom,const Object* aTo,class Path* aPath);
    void SetNodePath(int aFrom,int aTo,class Path* aPath);

    int NodeQ() { return mNodeQ; }
    vector<class Path*> InternalPaths() { return mInternalPath; }
    void OldDoCalcPaths(Node* aStart);
    void DuplicatePaths(class Sector* aOriginal,ObjectPool* aObjectPool);
    void PrecomputeInternalPaths();

    S32 mMinX,mMaxX,mMinY,mMaxY;
    void calcDimensions();
    void doCalcDimensions(const objects_t& aBorder);
    int mID;
    bool Contains(const Object* aObject) const;
    int inSector(S32 x,S32 y) const;
    int trySector(Object* aObject) const;
//    const class Path* GetInternalPath(Object* aFrom,Object* aTo);
public:
    // sector boundaries
    // 2---3
    // |   |
    // 0---1
    objects_t path01; objects_t path10;
    objects_t path13; objects_t path31;
    objects_t path32; objects_t path23;
    objects_t path20; objects_t path02;

    int ID() { return mID; }
    location_t location;
    objects_t objects;
    objects_t nexusPoints() const;

    Object* GetObjectById(S32 aId) const;
    bool UnitTest();

    S32 MiddleX() { return (mMinX+mMaxX)/2; }
    S32 MiddleY() { return (mMinY+mMaxY)/2; }
    const Path* GetInternalPath(Object* aStart,Object* aDest) const;
    bool insideSector(Object* aObject) const { return trySector(aObject)!=0; }
    Sector() {mNodeQ=0;}
    ~Sector() {mInternalPath.clear();}
    string ToString() const { return string("Sector"); }

    void Draw(class ViewerWindow * aWindow,long aViewFlags,long aColor) const;

};

class sectors_t : public vector<Sector *> {};

class WorldModel : public Memory {
friend class Summary;
public:
    S32 x1,y1,x2,y2; // map boundaries

    WorldModel(int aSectorX,int aSectorY);
    ~WorldModel(void);
    void analyseMap(void);
    void draw(class ViewerWindow * aWindow);

    // Important getters
    objects_t nexusPoints(Sector * sector);
    objects_t getBorder(Object * from,Object * to);
    objects_t sectorObjects(Sector * sector);
    sectors_t findSectors(Object * object);
    bool insideSector(Sector* aSector,Object* aObject);
    bool IsNexus(Object* object);
    
    sectors_t getSectors();
    int mSectorX,mSectorY;

    const objects_t Centers() { return mCenters; }
    
    // vars below should become private later...
    Hset commlist; // may become a list or queue
    Hset a_agents; // used for team composition
    Hset f_agents; // used for team composition
    Hset p_agents; // used for team composition
private:
    int pencol,filcol;

    Sector sectors[MAXSECTORS][MAXSECTORS];

    void borderPaths(int i,int j,Object * n[4],location_t location);
    void sectorPaths(int i,int j,Object * nodes[4]);
    objects_t reversePath(objects_t * path,Object * begin);
    objects_t path2objs(Object ** path);

    vector<GroupInfo> teams;
    Humanoid * nearestMember(Hset * agents,S32 & x,S32 & y,int members);
    void computeTeams();

    objects_t mCenters;
    objects_t mPlatoonAgents;

    Path* FindFastPath(Object* aStart,Object* aDest);
protected:
    /*
     * Highways and internal paths
     */
    void PrecomputeInternalPaths();
    void DoPrecomputeHighways(Object* aStart);
    void DoPrecomputeHighwaysNew(Object* aStart);
    void PrecomputeHighways();

    objects_t mHighwayObjects;
    
    map<S32,int> mNexusIndex;
    vector<class Path*> mHighway;
    int mNexusQ;
    void SetHighway(Object* aFrom,Object* aTo,Path* aPath);
    const Path* GetHighway(Object* aFrom,Object* aTo);
    void SetHighway(int aFrom,int aTo,Path* aPath);
    const Path* GetHighway(int aFrom,int aTo);

    const class Path* GetHighway(Sector* aFrom,Sector* aTo);

    map<S32,int> mSenseTimes;
    class Summary* mSummary;
public:
    Path * FindPath(Object* aStart,Object* aDest,
        const Rescue::Memory::ObjectPairFn& travelfn,bool aFast=false);
    Path * FindPath(Object* aStart,const Rescue::Memory::ObjectFn& aDestFn,
        const Rescue::Memory::ObjectPairFn& travelfn,bool aFast=false);
    Team * getTeam(Humanoid * aTeamMember);
    Task * getTask(Team * aTeam);

    void SetSenseTime(Object* aObject,int aTime) { mSenseTimes[aObject->id()]=aTime; }
    int GetSenseTime(Object* aObject) { return mSenseTimes[aObject->id()]; }

    /*
     * Worldmodel -> situationmodel
     */
protected:
    void AddPaths(SituationModel* aModel,Object* aDestination);
public:
    class SituationModel* ExtractCurrentSituation(class BasicAgent * aAgent);
    class Summary *GetSummary() { return mSummary; }
    // results
public:
    long mSectorBytes,mPathBytes,mHighwayBytes;
    long mInternalPaths,mHighways;
    long mSectorTime,mPathTime,mHighwayTime,mSummaryTime;
    int GetSectorX() { return mSectorX; }
    int GetSectorY() { return mSectorY; }  
};

#endif
