
#ifndef BEHAVIOUR_H
#define BEHAVIOUR_H

#include <adk.hxx>
#include <assert.h>
#include <vector>
#include <memory>
#include <string>

class Behaviour;

#include "../BasicAgent.h"
#include "../path.h"

  class Behaviour
  {
  friend class BehaviourForm;
  protected:
    /*
     * Functions for aiding of computing attainabillity score
     */
    // score is 1.0 if the exec is in operating condition 0.0 otherwise
    float GeneralOperationalAttainabilityScore(class TeamMember* aActor);
    // return a score that gives the probabillaty that exec will arrive at dest

    virtual float Priority(class SituationModel* aModel,class TeamMember* aActor)=0;
    virtual float Attainability(class SituationModel* aModel,class TeamMember* aActor)=0;

    Rescue::Object* mTarget;
    Rescue::Object* mPosition; 
    // prediction helpers
    // Situationmodel is not copied, just changed
    // returns aSituation in case of success, NULL in case of failure

    // move predictors, they check the path to the target and set mPosition to the target in case of success
    // situation is updated by setting the new position of aActor to the expect point that is reached
    // Move assumes success, travel actually checks the path and the length
    SituationModel* PredictMove(class SituationModel* aSituation,class TeamMember* aActor,Rescue::Object* aTarget);
    SituationModel* PredictTravel(class SituationModel* aSituation,class TeamMember* aActor,Rescue::Object* aTarget);
   public:
        Behaviour()
        {
            mTarget=0;
            mPosition=0;
         }
        virtual ~Behaviour()
        {

        }

        double Heuristic(class SituationModel* aModel,class TeamMember* aActor)
        {
            return Priority(aModel,aActor)*Attainability(aModel,aActor);
        }
        virtual class SituationModel* Predict(class SituationModel* aSituation,class TeamMember* aActor);
        virtual void Execute(class BasicAgent * aActor)=0;
                
        virtual string ToString()=0;

        virtual void Draw(class ViewerWindow* aWindow);
        
    };

    class Behaviours : public vector<Behaviour* >
    {
    public:
        void Add(Behaviour* aBehaviour)
        {
            push_back(aBehaviour);
        }
    };

#endif
