#include "behaviourdig.h"
#include "../AmbulanceAgent.h"
#include "../Team.h"
#include "../SituationModel.h"
#include "../WorldModel.h"

class DigPositionFn : public SelectionFunction
{
protected:
    ImaginaryHuman* target_human;
public:
    virtual bool Consider(ImaginaryObject * o) const
    {
        if (o==target_human->mPosition)
        {
            return true;
        }
        Node* N=dynamic_cast<Node*>(o->ActualObject());
        if (N)
        {
            for(Objects::const_iterator NI=N->neighbors().begin();NI!=N->neighbors().end();NI++)
            {
                if (target_human->mPosition->ActualObject()==*NI)
                {
                    return true;
                }
            }
        }
        return false;
    }
    virtual float Value(ImaginaryObject * o) const
    {
        ImaginaryBuilding* B=dynamic_cast<ImaginaryBuilding*>(o);
        if (B)
        {
            if (B->mFire!=0.0f)
            {
                return 0.1f;
            }
            if (B->mHp<1.0f)
            {
                return 0.5f;
            }
            else return 1.0f;
        }
        else
        {
            return 1.0f;
        }
    }
    DigPositionFn(ImaginaryHuman* target) : SelectionFunction()
    {
        target_human=target;
    }
    
};



float BehaviourDig::Priority(SituationModel* aModel,TeamMember* aActor)
{
    // TODO calc once for optimization and update on changes in neighbours and fieryness
    ImaginaryHuman* H=aModel->GetHuman(mTarget);
    if ((H) && (!H->Dead()))
    {
        return 0.16666f+(H->mBurry/3.0f)+(0.5f-(H->mHp/2.0));
    }
    else
    {
        return 0.0f;
    }
}

float BehaviourDig::Attainability(SituationModel* aModel,TeamMember* aActor)
{
    ImaginaryHuman* H=aModel->GetHuman(mTarget);
    if ((H) && (H->mBurry>0))
    {
        return GeneralOperationalAttainabilityScore(aActor);
    }
    return 0.0f;
}

SituationModel* BehaviourDig::Predict(SituationModel* aSituation,TeamMember* aActor)
{
    SituationModel* out=Behaviour::Predict(aSituation,aActor);
    const ImaginaryObjects& dig_positions=aSituation->FindSortedObjects(DigPositionFn(aSituation->GetHuman(mTarget)));
    if (dig_positions.Contains(aActor->Position()->ActualObject()))
    {
        ImaginaryHuman* H=(ImaginaryHuman*)out->Override(out->GetHuman(mTarget));
        H->mBurry=0;
        mPosition=0;
        return out;
    }
    else
    {
         for(ImaginaryObjects::const_iterator PI=dig_positions.begin();PI!=dig_positions.end();PI++)
        {
            SituationModel* out2=PredictTravel(out,aActor,(*PI)->ActualObject());
            if (out2)
            {
                return out2;
            }
        }
        return 0;
    }
}



void BehaviourDig::Execute(BasicAgent* aExec)
{
    Debug(spam_behaviour_execution,"Executing %s by %s %04ld\n",ToString().c_str(),aExec->agentName(),aExec->ID());
    WorldModel* WM=aExec->model();

    if (mPosition)
    {
        Path* P=WM->FindPath(aExec->Self(),mPosition,Path::AdvancedTravelFn(aExec));
        if (P)
        {
            if (P->Length()==1)
            {
                AmbulanceAgent* AA=dynamic_cast<AmbulanceAgent*>(aExec);
                if (AA)
                {
                    AA->Dig(dynamic_cast<Humanoid*>(mTarget));
                }
            }
            else
            {
                aExec->Move(P);
            }
        }
    }
    else
    {
         AmbulanceAgent* AA=dynamic_cast<AmbulanceAgent*>(aExec);
         if (AA)
         {
             AA->Dig(dynamic_cast<Humanoid*>(mTarget));
         }
    }
}

string BehaviourDig::ToString()
{
    return string("Dig out "+IntToString(((Humanoid*)mTarget)->id()));
}


