#include "behaviourpatrol.h"
#include "../Team.h"
#include "../SituationModel.h"
#include "../WorldModel.h"



    class UnknownNexiFn : public SelectionFunction
    {
    protected:
        WorldModel* mWM;
    public:
        bool Consider(ImaginaryObject* aObject) const
        {
            ImaginaryNode* N=dynamic_cast<ImaginaryNode*>(aObject);
            if ((N) && (mWM->IsNexus(N->ActualObject())))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        float Value(ImaginaryObject* aObject) const
        {
            return 1.0f/(1.0f+log(1.0f+(float)aObject->mTimeSensed));
        }
        UnknownNexiFn(WorldModel* aWM)
        {
            mWM=aWM;
        }

    };



float BehaviourPatrol::Priority(SituationModel* aModel,TeamMember* aActor)
{
    return 0.1;
}

float BehaviourPatrol::Attainability(SituationModel* aModel,TeamMember* aActor)
{
    return 1.0f;
}

SituationModel* BehaviourPatrol::Predict(SituationModel* aSituation,TeamMember* aActor)
{
    SituationModel* out=Behaviour::Predict(aSituation,aActor);
    const ImaginaryObjects& crossroads=out->FindSortedObjects(UnknownNexiFn(aSituation->WorldModel()));
    for(ImaginaryObjects::const_iterator CI=crossroads.begin();CI!=crossroads.end();CI++)
    {
        SituationModel* out2=PredictTravel(out,aActor,(*CI)->ActualObject());
        if (out2)
        {
            return out2;
        }
    }
    return 0;
}


void BehaviourPatrol::Execute(BasicAgent* aExec)
{
    Debug(spam_behaviour_execution,"Executing %s by %s %04ld\n",ToString().c_str(),aExec->agentName(),aExec->ID());
    WorldModel* WM=aExec->model();

    if (mPosition)
    {
        Path* P=WM->FindPath(aExec->Self(),mPosition,Path::AdvancedTravelFn(aExec));
        if (P)
        {
            aExec->Move(P);
        }
    }
    else
    {

    }
}


string BehaviourPatrol::ToString()
{
  return string("Patrol");
}
