#include "behaviourrecon.h"
#include "../Team.h"
#include "../SituationModel.h"
#include "../WorldModel.h"


    class UnknownCrossroadFn : public SelectionFunction
    {
    protected:
        Sector* mSector;
    public:
        bool Consider(ImaginaryObject* aObject) const
        {
            ImaginaryNode* N=dynamic_cast<ImaginaryNode*>(aObject);
            if (N)
            {
                Node* node=(Node*)aObject->ActualObject();
                int n=0;
                if (mSector->insideSector(node))
                {
                    for(Objects::const_iterator I=node->neighbors().begin();I!=node->neighbors().end();I++)
                    {
                        if (dynamic_cast<Road*>(*I))
                        {
                            n++;
                        }
                    }
                    return n>=3;
                }
            }
            return false;
        }

        float Value(ImaginaryObject* aObject) const
        {
            return 1.0f/(1.0f+log(1.0f+(float)aObject->mTimeSensed));
        }
        
        UnknownCrossroadFn(Sector* aSector) : SelectionFunction()
        {
            mSector=aSector;
        }
    };



float BehaviourRecon::Priority(SituationModel* aModel,TeamMember* aActor)
{
    return 0.1;
}

float BehaviourRecon::Attainability(SituationModel* aModel,TeamMember* aActor)
{
    return 1.0f;
}

SituationModel* BehaviourRecon::Predict(SituationModel* aSituation,TeamMember* aActor)
{
    SituationModel* out=Behaviour::Predict(aSituation,aActor);
    const ImaginaryObjects& crossroads=out->FindSortedObjects(UnknownCrossroadFn(mSector));
    for(ImaginaryObjects::const_iterator CI=crossroads.begin();CI!=crossroads.end();CI++)
    {
        SituationModel* out2=PredictTravel(out,aActor,(*CI)->ActualObject());
        if (out2)
        {
            return out2;
        }
    }
    return 0;
}


void BehaviourRecon::Execute(BasicAgent* aExec)
{
    Debug(spam_behaviour_execution,"Executing %s by %s %04ld\n",ToString().c_str(),aExec->agentName(),aExec->ID());
    WorldModel* WM=aExec->model();

    if (mPosition)
    {
        Path* P=WM->FindPath(aExec->Self(),mPosition,Path::AdvancedTravelFn(aExec));
        if (P)
        {
            aExec->Move(P);
        }
    }
    else
    {

    }
}


string BehaviourRecon::ToString()
{
  return string("Recon ")+IntToString(mSector->ID());
}
