#include "behaviourrefillwater.h"
#include "../FireAgent.h"
#include "../Team.h"
#include "../SituationModel.h"
#include "../WorldModel.h"

float BehaviourRefillWater::Priority(SituationModel* aModel,TeamMember* aActor)
{
    if (aActor->IsFireBrigade())
    {
        float score=aActor->Human()->mWater;
        if (score>0.5) return 0.0f;
        else return 1.0f-score;
    }
    else
    {
        return 0.0;
    }
}


float BehaviourRefillWater::Attainability(SituationModel* aModel,TeamMember* aActor)
{
    if (aActor->IsFireBrigade())
    {
        if (aActor->Human()->mWater>=1.0f) // full
            return 0.0f;
        else
            return GeneralOperationalAttainabilityScore(aActor);
    }
    else
    {
        return 0.0f;
    }
}


SituationModel* BehaviourRefillWater::Predict(SituationModel* aSituation,TeamMember* aActor)
{
    SituationModel* out=Behaviour::Predict(aSituation,aActor);

    if (aActor->Position()->ActualObject()==mTarget)
    {
        ImaginaryHuman* H=(ImaginaryHuman*)out->Override(aActor->Human());
        H->mWater+=config.tank_quantity_maximum()/(float)config.tank_quantity_recovery_on_refuge();
        mPosition=0; // queue me to wait
        return out;
    }
    else
    {
        return PredictTravel(out,aActor,mTarget);
    }
}


void BehaviourRefillWater::Execute(BasicAgent* aExec)
{
    Debug(spam_behaviour_execution,"Executing %s by %s %04ld\n",ToString().c_str(),aExec->agentName(),aExec->ID());
    WorldModel* WM=aExec->model();
    if (mPosition)
    {
        Path* P=WM->FindPath(aExec->Self(),mPosition,Path::AdvancedTravelFn(aExec));
        if (P)
        {
            aExec->Move(P);
        }
    }
    else
    {
        aExec->Idle();
    }
}

string BehaviourRefillWater::ToString()
{
  return string("Refill at "+IntToString(((Refuge*)mTarget)->id()));
}
