#include "logger.h"


Logger::Logger()
{
    m_logfile=0;
    m_logtime=0;
}

Logger::~Logger()
{
    if (m_logfile)
    {
        fclose(m_logfile);
    }
}

void Logger::Open(char* filename)
{
    if(filename)
    {
    // Open log file
        m_logfile = fopen(filename, "wb");
        if(m_logfile == 0)
        {
            fprintf(stderr, "error: can not open file '%s'\n", filename);
        }
        else
        {
            // write header
            const char header[] = "RoboCup-Rescue Prototype Log 00";
            fwrite(header, sizeof(header[0]), strlen(header)+1, m_logfile);
            fflush(m_logfile);
        }
    }
    else
    {   // turn off logging for filename = NULL
        if (m_logfile)
        {
            fclose(m_logfile);
            m_logfile=0;
       }
    }
}

void Logger::Log(long time,ObjectPool* memory)
{
Output logBuffer;
    if (m_logfile)
    {
        if (m_logtime==0)
        {
            // write setup block

            logBuffer.clear();
            logBuffer.put(KS_CONNECT_OK);
            Output::Cursor base = logBuffer.put(~(S32)0);
            memory->output(logBuffer,0);
            logBuffer.setSize(base);
            logBuffer.put(HEADER_NULL);

            logBuffer.log(m_logfile);
            m_logtime++;
            // make up for lost time at first cycle
            while(m_logtime<time)
            {
                logBuffer.clear();
                logBuffer.put(KS_UPDATE);
                Output::Cursor base = logBuffer.put(~(S32)0);
                logBuffer.put(m_logtime);
                memory->output(logBuffer, m_logtime);
                logBuffer.setSize(base);
                logBuffer.put(HEADER_NULL);

                logBuffer.log(m_logfile);
                m_logtime++;
            }
        }
        else
        {
            // write all cycles till specified time with update block
            while(m_logtime<time)
            {
                Output logBuffer;
                logBuffer.clear();
                logBuffer.put(KS_UPDATE);
                Output::Cursor base = logBuffer.put(~(S32)0);
                logBuffer.put(m_logtime);
                memory->output(logBuffer, m_logtime);
                logBuffer.setSize(base);
                logBuffer.put(HEADER_NULL);

                logBuffer.log(m_logfile);
                m_logtime++;
            }
        }
    }
}

