// $Id: main.cpp,v 1.1.1.1 2003/07/04 17:15:03 darseq Exp $
#include "main.h"
#include <getopt.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/timeb.h>
#include <exception>
#include "PoliceAgent.h"
#include "FireAgent.h"
#include "AmbulanceAgent.h"
#include "BasicCenter.h"
#include "BasicAgent.h"
#include "ViewerWindow.h"

#ifdef TESTNR
//#include "TestAgent1.h"
//#include xstr(xconcat(TestAgent,TESTNR))
#endif

using namespace Rescue;
using namespace std;

Config config;
FILE* g_analysis_file=0;
ViewerWindow* gViewerWindow=0;
vector<BasicActor*> gActors;
int gViewFlags;
int gSectorX=6;
int gSectorY=5;

#ifndef NDEBUG
    int g_spamflag=0;
    FILE* debug_file=stderr;

    void Debug(int flag,const char* format,...)
    {
        if (flag&g_spamflag)
        {

        va_list ap;
            va_start(ap,format);
            vfprintf(debug_file,format,ap);
            va_end(ap);
        }
    }
#else
    int g_spamflag=0;
    inline void Debug(int flag,const char* format,...)
    {
    }
#endif


char ObjectTypeToChar(Rescue::TypeId type)
{
    switch(type)
    {
        case  TYPE_RIVER_NODE:
        return 'w';
        case  TYPE_RIVER:
        return 'W';
        case  TYPE_BUILDING:
        return 'B';
        case  TYPE_NODE:
        return 'r';
        case  TYPE_REFUGE:
        return 'E';
        case  TYPE_ROAD:
        return 'R';
        case  TYPE_WORLD:
        return 'X';
        case  TYPE_FIRE_BRIGADE:
        return 'f';
        case  TYPE_POLICE_FORCE:
        return 'p';
        case  TYPE_AMBULANCE_TEAM:
        return 'a';
        case  TYPE_CIVILIAN:
        return 'c';
        case  TYPE_CAR:
        return 'C';
        case  TYPE_FIRE_STATION:
        return 'F';
        case  TYPE_POLICE_OFFICE:
        return 'P';
        case  TYPE_AMBULANCE_CENTER:
        return 'A';
        default:
        return '?';
    }
}


    void GetObjectPosition(const Rescue::Object* obj,double& x,double& y)
    {
        long px,py;
        GetObjectPosition(obj,px,py);
        x=(double)px;
        y=(double)py;
    }

    void GetObjectPosition(const Rescue::Object* obj,long& x,long& y)
    {
        float slide=0.5;
        const MovingObject* M;
        const PointObject *P1=0,*P2=0;
        if ((M=dynamic_cast<const MovingObject*>(obj)))
        {
            obj=M->position();
            slide=0.5;
            // todo: use position extra
        }
        const Edge* E;
        if ((E=dynamic_cast<const Edge*>(obj)))
        {
            P1=(const PointObject*)E->head();
            P2=(const PointObject*)E->tail();
        }
        else if ((P1=dynamic_cast<const PointObject*>(obj)))
        {
            P2=P1;
        }
        if ((P1)  && (P2))
        {
            x=(S32)(slide*P1->x()+(1.0f-slide)*P2->x());
            y=(S32)(slide*P1->y()+(1.0f-slide)*P2->y());
        }
        else
        {
            x=y=0;
        }
    }

    double DistanceBetween(const Rescue::Object* A,const Rescue::Object* B)
    {
        double x1,y1;
        GetObjectPosition(A,x1,y1);
        double x2,y2;
        GetObjectPosition(B,x2,y2);
        return sqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2));
    }


string IntToString(int value)
{
    char buffer[11];
    snprintf(buffer,11,"%d",value);
//    char *S=itoa(value,10,buffer);
    return string(buffer);
}


string FloatToString(float value)
{
    char buffer[31];
    snprintf(buffer,31,"%.2f",value);
    return string(buffer);
}


unsigned long timer()
{
    timeb timestruct;
    ftime(&timestruct);
    return ((timestruct.time&0xFFFFFF)*1000)+timestruct.millitm;
}
    
int main(int argc, char *argv[]) {
    char *host = "localhost"; int port;
    int ambulancenumber = 25,policenumber = 25,firenumber = 25,startcenters=7;
    int testnumber = 0;
    int resultperiod=0;
    int maxcycles=300;
    
    char c;
    bool logging=false;
    while((c = getopt(argc, argv, "c:h:a:p:f:t:s:d:w:v:x:y:lr:e:")) != EOF)
    {
        switch(c)
        {
            case 'c':
            {
                config.readConfigFile(optarg);
            }
            break;
            case 'h': 
            {
                host = optarg; 
            }
            break;
            case 'a': 
            {
                ambulancenumber = atoi(optarg);
            }
            break;
            case 'p': 
            {
                   policenumber = atoi(optarg);
            }
            break;
            case 'f':  
            {
                    firenumber = atoi(optarg);
            }
            break;
            case 't':
            {
                testnumber = atoi(optarg);
            }
            break;
            case 's':  
            {
                startcenters = atoi(optarg);
                if ((strchr(optarg,'p')) || (strchr(optarg,'P')))
                {
                    startcenters|=1;
                }
                if ((strchr(optarg,'f')) || (strchr(optarg,'F')))
                {
                    startcenters|=2;
                }
                if ((strchr(optarg,'a')) || (strchr(optarg,'A')))
                {
                    startcenters|=4;
                }
            
            }
            break;
            case 'd':
            {
                    g_spamflag = atoi(optarg);
            }
            break;
            case 'w':
                    // option removed
            break;
            case 'v':
                gViewFlags= atoi(optarg);
            break;
            case 'x':
                gSectorX=atoi(optarg);
            break;
            case 'y':
                gSectorY=atoi(optarg);
            break;
            case 'l':
                logging=true;
            break;
            case 'e':
                maxcycles = atoi(optarg);
            break;
            case 'r':
                resultperiod=atoi(optarg);
            break;
        }
    }

    port = config.port();
    Controller::setup_connection(host, port);
    BasicAgent *A;
    BasicCenter* C;

    if (startcenters)
    {
        if (startcenters&1)
        {
           Debug(spam_startup,"Connect: PoliceCenter\n");
           C= new BasicCenter( AGENTTYPE_POLICE_OFFICE );
           if (C->connect())
           {
             gActors.push_back(C);
           }
        }
        if (startcenters&2)
        {
           Debug(spam_startup,"Connect: FireCenter\n");
           C= new BasicCenter( AGENTTYPE_FIRE_STATION );
           if (C->connect())
           {
             gActors.push_back(C);
           }
        }
        if (startcenters&4)
        {
           Debug(spam_startup,"Connect: AmbulanceCenter\n");
           C= new BasicCenter( AGENTTYPE_AMBULANCE_CENTER );
           if (C->connect())
           {
             gActors.push_back(C);
           }
        }
    }
    int i;
    char logfilename[100];
    
    for(i=1;i<=firenumber;i++)
    {
        Debug(spam_startup,"Connect: Firebrigade %d/%d\n",i,firenumber);
        A = new FireAgent();
        if (logging)
        {
            sprintf(logfilename,"FireAgent%02d.log",i);
            A->EnableLogging(logfilename);
        }
        
        if (!A->connect())
        {
            Debug(spam_startup,"Connect: Reached maximum number of Firebrigade connects\n");
            break;
        }
        else
        {
            gActors.push_back(A);
        }
    }

    for(i=1;i<=ambulancenumber;i++)
    {
        Debug(spam_startup,"Connect: Ambulance %d/%d\n",i,ambulancenumber);
        A = new AmbulanceAgent();

        if (logging)
        {
            sprintf(logfilename,"Ambulancegent%02d.log",i);
            A->EnableLogging(logfilename);
        }

        if (!A->connect())
        {
            Debug(spam_startup,"Connect: Reached maximum number of Ambulance connects\n");
            break;
        }
        else
        {
            gActors.push_back(A);
        }
    }

    for(i=1;i<=policenumber;i++)
    {
        Debug(spam_startup,"Connect: Police %d/%d\n",i,policenumber);
        A = new PoliceAgent();
        if (logging)
        {
            sprintf(logfilename,"PoliceAgent%02d.log",i);
            A->EnableLogging(logfilename);
        }

        
        if (!A->connect())
        {
            Debug(spam_startup,"Connect: Reached maximum number of Police connects\n");
            break;
        }
        else
        {
            gActors.push_back(A);
        }
    }

#ifdef TESTNR
    for(i=1;i<=testnumber;i++) {
        Debug(spam_startup,"Connect:TESTNR:%d TestAgent %d/%d\n",
            TESTNR,i,testnumber);
        A = new TestAgent();
        if (!A->connect())
        {
            Debug(spam_startup,"Connect: Reached maximum number of Test connects\n");
            break;
        }
        else
        {
            gActors.push_back(A);
        }
    }                 
#endif
    if ((gActors.begin()!=gActors.end()) && (gActors[0]) && (gViewFlags))
    {
        gViewerWindow=new ViewerWindow;
        gViewerWindow->mViewFlags=gViewFlags;
        gViewerWindow->Create(600,400);
        gViewerWindow->SetActor(gActors[0]);
    }
    
    Controller::start();
    unsigned long lastredraw=0,thistime;
    int lastresulttime=-1;
    bool ended=false;
    try
    {
        while(!ended)                               
        {

            if (Controller::loop())
            {
                thistime=timer();

                if (thistime>lastredraw+200)
                {
                    lastredraw=thistime;
                    if ((gViewerWindow) && (gViewerWindow->Actor()) && (gViewerWindow->Actor()->Initialized()))
                    {
                        gViewerWindow->BeginFrame();
                        gViewerWindow->Actor()->Draw(gViewerWindow);
                        gViewerWindow->EndFrame();
                    }
                }
            }
            int agenttime=(*gActors.begin())->Time();
            if ((resultperiod>0) && (agenttime>lastresulttime) && ((agenttime%resultperiod)==0))
            {
                lastresulttime=agenttime;
                char filename[100];
                snprintf(filename,100,"Results_%d_%ld",agenttime,time(0));
                FILE *results=fopen(filename,"w+");
                for(vector<BasicActor*>::iterator AI=gActors.begin();AI!=gActors.end();AI++)
                {
                    BasicAgent* BA=dynamic_cast<BasicAgent*>(*AI);
                    if (BA)
                    {
                        BA->PrintResults(results);
                    }
                    // TODO print results for centers?
                }
                fclose(results);
            }            
            ended=((*gActors.begin())->Time()>=maxcycles) ;//|| ((gViewerWindow) && (gViewerWindow->Closed()));
        }
    }
    catch(exception E)
    {
        fprintf(stderr,"Caught exception: %s\n",E.what());
    }
    catch(...)
    {
        fprintf(stderr,"Caught unknown exception\n");
    }
        
    Controller::end();
                
    if ((gViewerWindow) && (!gViewerWindow->Closed()))
    {
        gViewerWindow->Close();
        delete gViewerWindow;
    }
}
