
#include "path.h"
#include "main.h"
#include "BasicAgent.h"
#include "ViewerWindow.h"
#include "WorldModel.h"

using namespace Rescue;


Path::AdvancedTravelFn::AdvancedTravelFn(BasicAgent* traveler) : TraversalObjectPairFn()
{
    m_traveler=traveler->Self();
    if (traveler->model())
    {
       m_vehicles=traveler->model()->find_list(Memory::IsAgentObjectFn(),Memory::FUNCTION_MINIMIZE);
     }
    else
    {
        m_vehicles=0;
    }
    
}

Path::AdvancedTravelFn::AdvancedTravelFn(Object* traveler,Memory* memory) : TraversalObjectPairFn()
{
    m_traveler=traveler;
    if (memory)
    {
        m_vehicles=memory->find_list(Memory::IsAgentObjectFn(),Memory::FUNCTION_MINIMIZE);
    }
    else
    {
        m_vehicles=0;
    }
}

Path::AdvancedTravelFn::~AdvancedTravelFn()
{
    if (m_vehicles)
    {
        delete m_vehicles;
    }
}

      
bool Path::AdvancedTravelFn::consider(const Object *to, const Object *from) const
{
    if (m_vehicles)
    {

        if (to->type() == TYPE_ROAD)
        {
            Road *r = (Road *) to;
            // compute free lines
            int lines = (from == r->head() ? r->linesToTail() : r->linesToHead());
            lines-= (int)floor((0.5 * (r->linesToHead() + r->linesToTail()) *
                               (r->block() / (double) r->width())) + 0.5);
            if (lines<=0) // no lines available: road blocked
            {
                return false;
            }

//            lines*=2; // head and tail hack since can't tell on what side the vehicles are
            for(int i=0;m_vehicles[i];i++)
            {
                MovingObject* m=(MovingObject*)m_vehicles[i];
                if (m!=m_traveler)
                {
                    if (m->position()==to) // TODO check only direction of m
                    {
                        lines--;
                    }
                }
            }

            if (lines<=0)    // more vehicles on road than available lines: traffic jam
            {
                return false;
            }
            else
            {
                return true;
            }
        }
        else
        {
            for(int i=0;m_vehicles[i];i++)
            {
                MovingObject* m=(MovingObject*)m_vehicles[i];
                if (m!=m_traveler)
                {
                    if (m->position()==to) // TODO check only direction of m
                    {
                        return false;
                    }
                }
            }
            return true;
        }
    }
    else
    {
        return Memory::TraversalObjectPairFn::consider(to,from);
    } 
}

/*
 * Various operators to construct pathss
 */


Path& Path::operator+=(const Path& aRest)
{
    const_iterator i=aRest.begin();
    if (aRest.begin()!=aRest.end())
    {
        if (End()==*i)
        {
            i++;
        }
        for(;i!=aRest.end();i++)
        {
            push_back(*i);
        }
    }
    return *this;
}

Path& Path::operator+=(const list<Object*>& aObjects)
{
    if (aObjects.begin()!=aObjects.end())
    {
        list<Object*>::const_iterator i=aObjects.begin();
        if (End()==*i)
        {
            i++;
        }
        for(;i!=aObjects.end();i++)
        {
            push_back(*i);
        }
    }
    return *this;
}

Path& Path::operator+=(const vector<Object*>& aObjects)
{
    if (aObjects.begin()!=aObjects.end())
    {
        vector<Object*>::const_iterator i=aObjects.begin();
        if (End()==*i)
        {
            i++;
        }

        for(;i!=aObjects.end();i++)
        {
            push_back(*i);
        }
    }
    return *this;
}

Path& Path::operator+=(Object* aObject)
{
    if (End()!=aObject)
    {
        if (aObject)
        {
            push_back(aObject);
        }
    }
    return *this;
}
                  
Path* Path::Create(BasicAgent* exec,const Memory::ObjectFn& destfn, const Memory::ObjectPairFn& travelfn)
{
    // Find all objects for which fn decides it's a good destination
    Object** objs;
    objs=exec->model()->find_list(destfn,Memory::FUNCTION_MINIMIZE);
    
    // TODO: resort possible targets for best target according to the f function of the targetFn
    
    // scan all possible targets for available paths
    int i;
    if ((!objs[0]))
    {
        Debug(spam_path,"Path::Create( ) -  No destinations found\n");
        delete objs;
        return 0;
    }
    else
    {
        for(i=0;(objs[i]);i++)
        {
            if (objs[i]==exec->Position())
            {
                Path* out=new Path;
                (*out)+=objs[i];
                return out;
            }
        }
        for(i=0;(objs[i]);i++)
        {
            Path* out=new Path();
            {
                if (out->Find(exec->Location(),objs[i],exec->model(),travelfn))
                {
                    delete objs;
                    return out;
                }

            }
        }
        Debug(spam_path,"Path::Compute( ) - No path found\n");
        delete objs;
        return 0;
    }
}

Path* Path::Create(Object* aFrom,Object* aTo,Memory* memory,const Memory::ObjectPairFn& travelfn)
{
    Path* Out=new Path();
    Memory::Location L(0,0);
    
    MovingObject* M;
    M=dynamic_cast<MovingObject*>(aFrom);
    if (M)
    {
        L=Memory::Location(M->position(),M->positionExtra());
    }
    else
    {
        L=Memory::Location(aFrom,0);
    }
    M=dynamic_cast<MovingObject*>(aTo);
    if (M)
    {
        aTo=M->position();
    }

    if (Out->Find(L,aTo,memory,travelfn))
    {
        return Out;
    }
    else
    {
        delete Out;
        return 0;
    }
}

bool Path::Find(const Memory::Location& start,Object* dest,Memory* memory,const Memory::ObjectPairFn& travelfn)
{
    Road* R=dynamic_cast<Road*>(start.m_o);
    Memory::Location newstart(0,0);
    if (R)
    {
        if (start.m_dist<=R->length()/2)
        {
            newstart=Memory::Location(start.m_o,0);
        }
        else
        {
            newstart=Memory::Location(start.m_o,R->length());
        }
    }
    else
    {
        newstart=start;
    }
    Object **p;
    p=memory->find_path(newstart,
                        Memory::IsObjectObjectFn(dest),
                        Memory::DistanceObjectFn(dest),
                        travelfn,
                        true);
    if (p)
    {
        clear();
        push_back(newstart.m_o);
        Debug(spam_path,"Path::Compute( ) \n",dest->id());
        // puth path in container
        for(int path_length=0;p[path_length];path_length++)
        {
            push_back(p[path_length]);
        }
        DebugPrint();
        delete p;
        return true;
    }
    else
    {
        return false;
    }
}

void Path::Optimize()
{
    iterator PI,I1,I2;
    
    for(PI=begin(),I1=begin();I1!=end();PI++,I1++)
    {
        
        for((I2=I1)++;I2!=end();I2++)
        {
            if (*I1==*I2) 
            {
                I1=I2;
            }
        }
        *PI=*I1;
    }
    if (PI!=end())
    {
        erase(PI,end());
    }
}



string Path::ToString() const
{
    string out;
    out+="[ ";
    for(const_iterator I=begin();I!=end();I++)
    {
        out+=ObjectTypeToChar((*I)->type())+IntToString((*I)->id())+" ";
        
    }
    out+="]";
    return out;
}

void Path::DebugPrint() const
{
    const string& S=ToString();
    Debug(spam_path,"%s\n",S.c_str());
}


Object* Path::End()  const
{
    if (begin()!=end())
    {
        return *rbegin();
    }
    else
    {
        return 0;
    }
}

Object* Path::Start() const
{
    return *begin();
}


float Path::Distance()  const
{
float distance=0.0f;
Road* R;
    for(const_iterator I=begin();I!=end();I++)
    {
        R=dynamic_cast<Road*>(*I);
        if (R)
        {
            // TODO map measurements
            distance+=(long)R->length()/1000.0f;
        }
    }
    return distance;
}

Objects Path::GetObjects() const
{
    Objects out;
    for(const_iterator i=begin();i!=end();i++)
    {
        out.push_back(*i);
    }
    return out;
}

Object** Path::GetArray() const
{
    Object** O=new Object*[size()+2];
    int n=0;
    for(const_iterator i=begin();i!=end();i++)
    {
        O[n++]=*i;
    }
    O[n]=0;
    return O;
}
  
bool Path::Check(const Memory::ObjectPairFn& fn) const
{
    const_iterator ni;
    for(const_iterator i=begin();i!=end();i++)
    {
        ni=i;ni++;
        if (ni!=end())
        {
            if (!fn.consider(*i,*ni))
            {
                return false;
            }
        }
    }

    return true;
}

bool Path::Consistency() const
{
    const_iterator ni;
    for(const_iterator i=begin();i!=end();i++)
    {
        ni=i;ni++;
        if (ni!=end())
        {
            Building* B=dynamic_cast<Building*>(*i);
            Road* R=dynamic_cast<Road*>(*i);
            Node* N=dynamic_cast<Node*>(*i);
            if (R)
            {
                if ((R->tail()!=*ni) && (R->head()!=*ni))
                {
                    DebugPrint();
                    return false;
                }
            }
            if (N)
            {
                bool found=false;
                for(Objects::const_iterator NI=N->neighbors().begin();NI!=N->neighbors().end();NI++)
                {
                    if (*NI==*ni)
                    {
                        found=true;
                        break;
                    }
                }
                if (!found)
                {
                    DebugPrint();
                    return false;
                }
            }
            if (B)
            {
                bool found=false;
                for(Objects::const_iterator EI=B->entrances().begin();EI!=B->entrances().end();EI++)
                {
                    if (*EI==*ni)
                    {
                        found=true;
                        break;
                    }
                }
                if (!found)
                {
                    DebugPrint();
                    return false;
                }
            }                

        }
    }
    return true;
}

 
bool Path::Contains(const Object* obj) const
{

    for(const_iterator i=begin();i!=end();i++)
    {
        if (*i==obj)
        {
            return true;
        }
    }
    return false;
}

Object* Path::GetReachablePosition()
{
unsigned long limit =  20 * 1000000 / 3600;
unsigned long distance;
Object* out;
    for(iterator I=begin();I!=end();I++)
    {
        out=*I;
        Road* R=dynamic_cast<Road*>(out);
        if (R)
        {
            distance+=R->length();
            if (distance>=limit)
            {
                break;
            }
        }

    }
    return out;
}


void Path::ConvertObjectPool(ObjectPool* aTo)
{
    for(iterator I=begin();I!=end();I++)
    {
        *I=aTo->get((*I)->id());
    }
}


Path* Path::CreateReversePath() const
{
    Path* Out=new Path;
    for(const_reverse_iterator RI=rbegin();RI!=rend();RI++)
    {
        (*Out)+=*RI;
    }
    return Out;
}

void Path::Draw(ViewerWindow* aWindow,long aColor) const
{
    if (!aWindow)
        return;
    Object* O;
    Road* R;
    Building* B;
    PointObject *P1,*P2;
    Memory::TraversalObjectPairFn fn;
    
    for(const_iterator i=begin();i!=end();i++)
    {
        O=*i;
        if ((R=dynamic_cast<Road*>(O)))
        {
            if ((fn.consider(R,R->head())) && (fn.consider(R,R->tail())))
            {
                aWindow->SetPenColor(aColor);
            }
            else
            {
                aWindow->SetPenColor(0xFFFFFF^aColor);
            }

            P1=(PointObject*)R->head();
            P2=(PointObject*)R->tail();
            aWindow->Line(P1->x(),P1->y(),P2->x(),P2->y(),3);
        }
        else if ((B=dynamic_cast<Building*>(O)))
        {
            if (B->brokenness()>0)
            {
                aWindow->SetPenColor(0xFF0000);
            }
            else
            {
                aWindow->SetPenColor(0x00FF00);
            }
            P2=(PointObject*)B;
            Objects entr=B->entrances();
            for(Objects::const_iterator I=entr.begin();I!=entr.end();I++)
            {
                if ((P1=dynamic_cast<PointObject*>(*I)))
                {
                    aWindow->Line(P1->x(),P1->y(),P2->x(),P2->y(),2);
                }
            }
        }
    }
}


