#ifndef PATH_H
#define PATH_H

#include <adk.hxx>
#include <exception>
#include <string>
#include <vector>
using namespace std;
using namespace Rescue;

    
class Path : protected vector<Object*>
{
public:
     class AdvancedTravelFn : public Memory::TraversalObjectPairFn
     {
     protected:
         Object* m_traveler;
         Object** m_vehicles;
     public:
         AdvancedTravelFn(class BasicAgent* traveler);
         AdvancedTravelFn(Object* traveler,Memory* memory);
         virtual ~AdvancedTravelFn();
         virtual bool consider(const Object *a, const Object *b) const;
     };

     class DummyTravelFn : public Memory::TraversalObjectPairFn
     {
           virtual bool consider(const Object *to, const Object *from) const
           {
             return true;
           }
     };

friend class BasicAgent;
protected:
    bool Find(
        const Memory::Location& start,
        Object* dest,
        Memory* memory,
        const Memory::ObjectPairFn& travelfn);
public:
    class ExceptionIndexOutOfBounds/* : public exception */
    {

    };
    Path()
    {
    }
    Path& operator+=(const Path& aRest);
    Path& operator+=(const list<Object*>& aObjects);
    Path& operator+=(const vector<Object*>& aObjects);
    Path& operator+=(Object* aObject);

    ~Path()
    {
    }
    // Create the path from the memory, starting at start and going to a non specific destination
    // determined by fn and a traversalfunction travelfn that defaults to a function that checks for blocked roads and traffic jams
    static Path* Create(BasicAgent* exec,
        const Memory::ObjectFn& destfn,
        const Memory::ObjectPairFn& travelfn);
    static Path* Create(Object* aFrom,Object* aTo,Memory* memory,const Memory::ObjectPairFn& travelfn);
    // gives the number of elements in the path
    int Length() const
    {
        return size();
    }
    // get an element at index from the path. throws an exception if the index is out of bounds
    Objects GetObjects() const;
    Object** GetArray() const;
    // type casts the path to a boolean. true if the path is existent and valid, false otherwise

    // Make a string describing this path
    string ToString() const;
    // print a string representation of the path
    void DebugPrint() const;
    // determine if an agent at the specified position is considered to have reached the end of the path
    bool Done(const Object* position);
    // give last element in path
    Object* End() const;
    // give first element in path
    Object* Start() const;
    // Return length of path in meters
    float Distance() const;
    // Return if the path contains a certain object
    bool Contains(const Object* obj) const;
    // return true if the traversal of the path is considered valid by the travel function
    bool Check(const Memory::ObjectPairFn& fn) const;
    // returns true if the sequention is an executable path
    bool Consistency() const;
    // create a new path that is correct for object pool aTo by mapping the current IDs to those in aTo
    void ConvertObjectPool(ObjectPool* aTo);
    // Create a reversed path to travel from End() to Begin()
    Path* CreateReversePath() const;
    // Get the maximum reachable object in one move by an agent
    // assumed 20km/h as defined in the traffic simulator's Constant.java (whytf is this no in the config!!)
    Object* GetReachablePosition();
          
    // remove doubles and overlaps in path in N! time
    void Optimize();

    void Draw(class ViewerWindow* aWindow,long aColor=0x00FF00) const;
};

#endif
